import java.util.StringTokenizer;
import java.util.NoSuchElementException;

/**
 *
 * This class is used to build a complex number along with elementary functions
 * that are applied to a complex number. The implementation allows one to call
 * the functions without actually instantiating a complex number object.
 *
 * To start we assume a representation of the form
 *
 *   a + i b
 *
 * where a and b are both real values and if b is not equal to zero the number
 * is imaginary.
 *
 * NOTE: In the entire object the variable names 'a' and 'b' are used as in the
 * definitions above. Note that these are global definitions for the object.
 * So, in modifying this object, realize that if 'num' is a ComplexNumber object
 * say from the declaration
 *
 *    ComplexNumber num = new ComplexNumber(a, b);
 *
 * then num.a refers to the real part and num.b refers to the complex part of
 * the complex number.
 *
 */

public class ComplexNumber {
  public double a;
  public double b;

  // CONSTRUCTORS FOR THE CLASS

  /**
   * Creates a complex number with the newA + newB i form.
   */
  public ComplexNumber(double newA, double newB) {
    a = newA;
    b = newB;
  }

  /**
   * Constructs a copy of the given complex number.
   */
  public ComplexNumber(ComplexNumber num) {
    a = num.a;
    b = num.b;
  }

  // THE FOLLOWING METHODS DEFINE THE STANDARD ARITHMETIC OPERATIONS

  /** 
   * Adds two complex numbers and returns the resulting complex number.
   */
  public static ComplexNumber add(ComplexNumber cn1, ComplexNumber cn2) {
    return new ComplexNumber(cn1.a + cn2.a, cn1.b + cn2.b);
  }

  /** 
   * Subtracts two complex numbers and returns the resulting complex number.
   */
  public static ComplexNumber subtract(ComplexNumber cn1, ComplexNumber cn2) {
    return new ComplexNumber(cn1.a - cn2.a, cn1.b - cn2.b);
  }

  /** 
   * Multiplies two complex numbers and returns the resulting complex number.
   */
  public static ComplexNumber multiply(ComplexNumber cn1, ComplexNumber cn2) {
    double aVal = cn1.a * cn2.a - cn1.b * cn2.b;
    double bVal = cn1.a * cn2.b + cn1.b * cn2.a;
    return new ComplexNumber(aVal, bVal);
  }

  /** 
   * Divides two complex numbers and returns the resulting complex number.
   */
  public static ComplexNumber divide(ComplexNumber cn1, ComplexNumber cn2) {
    ComplexNumber result; 
    result = multiply(cn1, cn2.conjugate());
    double divisor = Math.pow(cn2.abs(),2.0);
    return new ComplexNumber(result.a /divisor, result.b /divisor);
  }

  private boolean zeroTest() {
    if(a == 0.0 && b == 0.0) return true;
    return false;
  }

  // REAL AND IMAGINARY PARTS AND CONJUGATION AND SUCH

  /**
   * Negates the complex number.
   */
  public void negate() {
    a = -a;
    b = -b;
  }

  /**
   * Negates the complex number.
   */
  public ComplexNumber negativeOf() {
    return new ComplexNumber(-a,-b);
  }
  public static ComplexNumber negativeOf(ComplexNumber z) {
    return z.negativeOf();
  }

  /**
   * multiplies the complex number by i.
   */
  public ComplexNumber iZ() {
    return new ComplexNumber(-b,a);
  }
  public static ComplexNumber iZ(ComplexNumber z) {
    return z.iZ();
  }

  /**
   * Returns the real part of the complex number.
   */
  public double Re(){
    return a;
  }

  /**
   * Returns the real part of the complex number.
   */
  public static double Re(ComplexNumber z) {
    return z.Re();
  }

  /**
   * Tests to see if the number is real.
   */
  public boolean isReal() {
    if(b == 0.0) {
      return true;
    } else {
      return false;
    }
  }
  public static boolean isReal(ComplexNumber cn) {
    return cn.isReal();
  }

  /**
   * Returns the imaginary part of the complex number.
   *
   * @returns  a ComplexNumber.
   */
  public double Im() {
    return b;
  }

  /**
   * Returns the imaginary part of the complex number.
   */
  public static double Im(ComplexNumber z) {
    return z.Im();
  }

  /**
   * Tests to see if the number is imaginary.
   *
   */
  public boolean isImaginary() {
    if(a == 0.0) {
      return true;
    } else {
      return false;
    }
  }
  public static boolean isImaginary(ComplexNumber cn) {
    return cn.isImaginary();
  }

  /**
   * Returns the conjugate of the complex number.
   */
  public ComplexNumber conjugate() {
    return new ComplexNumber(a,-b);
  }

  /**
   * Returns the conjugate of the complex number.
   */
  public static ComplexNumber conjugate(ComplexNumber z) {
    return z.conjugate();
  }

  /**
   * Returns the magnitude of the complex number.
   *
   * @returns  a double.
   */
  public double abs() {
    return Math.sqrt(a*a+b*b);
  }

  /**
   * Returns the magnitude of the complex number.
   */
  public static double abs(ComplexNumber z) {
    return z.abs();
  }

  /**
   * Returns the power of the complex number.
   */
  public ComplexNumber pow(double p) {
    double rval, realPart;
    double ival, imagPart;
    rval = Math.pow(abs(),p);
    ival = Math.atan(Im()/Re()) * p;
    if(ival != 0.0) {
      realPart = Math.exp(rval) * Math.cos(ival);
      imagPart = Math.exp(rval) * Math.sin(ival);
      return new ComplexNumber(realPart, imagPart);
    } else {
      return new ComplexNumber(Math.pow(Re(),p), 0.0);
    }
  }

  /**
   * Returns the power of the complex number.
   */
  public static ComplexNumber pow(ComplexNumber z, double p) {
    return z.pow(p);
  }

  /**
   * Returns the square root of the complex number.
   */
  public ComplexNumber sqrt() {
    return pow(0.5);
  }

  /**
   * Returns the power of the complex number.
   */
  public static ComplexNumber sqrt(ComplexNumber z) {
    return z.sqrt();
  }

  // COMPLEX VERSIONS OF SOME OF THE ELEMENTARY FUNCTIONS

  /**
   * Returns the exp of the ComplexNumber.
   */
  public ComplexNumber exp() {
    double rval, realPart;
    double ival, imagPart;
    rval = Re();
    ival = Im();
    if(ival != 0.0) {
      realPart = Math.exp(rval) * Math.cos(ival);
      imagPart = Math.exp(rval) * Math.sin(ival);
      return new ComplexNumber(realPart, imagPart);
    } else {
      return new ComplexNumber(Math.exp(Re()), 0.0);
    }
  }

  /**
   * Returns the exp of the ComplexNumber.
   */
  public static ComplexNumber exp(ComplexNumber z) {
    return z.exp();
  }

  /**
   * Returns the log of the ComplexNumber.
   */
  public ComplexNumber log() {
    double rval;
    double ival;
    rval = abs();
    ival = Math.atan(Im()/Re());
    if(ival != 0.0) {
      return new ComplexNumber(Math.log(rval), ival);
    } else {
      return new ComplexNumber(Math.log(Re()), 0.0);
    }
  }

  /**
   * Returns the log of the ComplexNumber.
   */
  public static ComplexNumber log(ComplexNumber z) {
    return z.log();
  }

  /**
   * Returns the sine of the ComplexNumber.
   */
  public ComplexNumber sin() {
    double rval, realPart;
    double ival, imagPart;
    rval = Re();
    ival = Im();
    if(ival != 0.0) {
      realPart = 0.5 * Math.sin(rval) * ( Math.exp(ival) + Math.exp(-ival) );
      imagPart = 0.5 * Math.cos(rval) * ( Math.exp(ival) - Math.exp(-ival) );
      return new ComplexNumber(realPart, imagPart);
    } else {
      return new ComplexNumber(Math.sin(a), 0.0);
    }
  }

  /**
   * Returns the sine of the ComplexNumber.
   */
  public static ComplexNumber sin(ComplexNumber z) {
    return z.sin();
  }

  /**
   * Returns the cosine of the ComplexNumber. 
   */
  public ComplexNumber cos() {
    double rval, realPart;
    double ival, imagPart;
    rval = Re();
    ival = Im();
    if(ival != 0.0) {
      realPart = 0.5 * Math.cos(rval) * ( Math.exp(-ival) - Math.exp(ival) );
      imagPart = 0.5 * Math.sin(rval) * ( Math.exp(-ival) + Math.exp(ival) );
      return new ComplexNumber(realPart, imagPart);
    } else {
      return new ComplexNumber(Math.cos(a), 0.0);
    }
  }

  /**
   * Returns the cosine of the ComplexNumber.
   */
  public static ComplexNumber cos(ComplexNumber z) {
    return z.cos();
  }

  /**
   * Returns the tangent of the ComplexNumber.
   */
  public ComplexNumber tan() {
    return divide(sin(),cos());
  }

  /**
   * Returns the tangent of the ComplexNumber.
   */
  public static ComplexNumber tan(ComplexNumber z) {
    return z.tan();
  }

  /**
   * Returns the secant of the ComplexNumber.
   */
  public ComplexNumber sec() {
    return divide(new ComplexNumber(1.0,0.0),cos());
  }

  /**
   * Returns the secant of the ComplexNumber.
   */
  public static ComplexNumber sec(ComplexNumber z) {
    return z.sec();
  }

  /**
   * Returns the cosecant of the ComplexNumber.
   *
   * @returns  a ComplexNumber.
   */
  public ComplexNumber csc() {
    return divide(new ComplexNumber(1.0,0.0),sin());
  }

  /**
   * Returns the cosecant of the ComplexNumber.
   *
   * @param z  a ComplexNumber object.
   * @returns  a ComplexNumber.
   */
  public static ComplexNumber csc(ComplexNumber z) {
    return z.csc();
  }

  /**
   * Returns the cotangent of the ComplexNumber.
   *
   * @returns  a ComplexNumber.
   */
  public ComplexNumber cot() {
    ComplexNumber den = sin();
    return divide(cos(),sin());
  }

  /**
   * Returns the cotangent of the ComplexNumber.
   *
   * @param z  a ComplexNumber object.
   * @returns  a ComplexNumber.
   */
  public static ComplexNumber cot(ComplexNumber z) {
    return z.cot();
  }

  /**
   * Returns the arcsine of the ComplexNumber.
   */
  public ComplexNumber asin() {
    if(Im() == 0.0) return new ComplexNumber(Math.asin(a), 0.0);
    ComplexNumber z1 = iZ(); 
    ComplexNumber z2
        = sqrt(subtract(new ComplexNumber(1.0,0.0), multiply(this,this))); 
    return negativeOf(iZ(log(add(z1,z2))));
  }
  public static ComplexNumber asin(ComplexNumber z) {
    return z.asin();
  }

  /**
   * Returns the arccosine of the ComplexNumber.
   */
  public ComplexNumber acos() {
    if(Im() == 0.0) return new ComplexNumber(Math.acos(a), 0.0);
    ComplexNumber z1 = iZ(); 
    ComplexNumber z2
        = sqrt(subtract(new ComplexNumber(1.0,0.0), multiply(this,this))); 
    return negativeOf(iZ(log(add(z1,iZ(z2)))));
  }
  public static ComplexNumber acos(ComplexNumber z) {
    return z.acos();
  }

  /**
   * Returns the arctangentof the ComplexNumber.
   */
  public ComplexNumber atan() {
    if(Im() == 0.0) return new ComplexNumber(Math.atan(a), 0.0);
    ComplexNumber z1 = subtract(new ComplexNumber(0.0, 1.0),this);
    ComplexNumber z2 = subtract(new ComplexNumber(0.0, 1.0),this);
    return divide(iZ(log(divide(z1,z2))),new ComplexNumber(2.0,0.0));
  }
  public static ComplexNumber atan(ComplexNumber z) {
    return z.atan();
  }

  /**
   * Returns the sinh of the ComplexNumber.
   */
  public ComplexNumber sinh() {
    ComplexNumber z = negativeOf();
    ComplexNumber two = new ComplexNumber(2.0,0.0);
    ComplexNumber result = divide(subtract(exp(),exp(z)),two);
    return new ComplexNumber(result.Re(),result.Im());
  }
  public static ComplexNumber sinh(ComplexNumber z) {
    return z.sinh();
  }

  /**
   * Returns the cosh of the ComplexNumber.
   */
  public ComplexNumber cosh() {
    ComplexNumber z = negativeOf();
    ComplexNumber two = new ComplexNumber(2.0,0.0);
    ComplexNumber result = divide(add(exp(),exp(z)),two);
    return new ComplexNumber(result.Re(),result.Im());
  }
  public static ComplexNumber cosh(ComplexNumber z) {
    return z.cosh();
  }

  /**
   * Returns the tanh of the ComplexNumber.
   */
  public ComplexNumber tanh() {
    ComplexNumber result = divide(sinh(),cosh());
    return new ComplexNumber(result.Re(),result.Im());
  }
  public static ComplexNumber tanh(ComplexNumber z) {
    return z.tanh();
  }

  /**
   * Returns the sech of the ComplexNumber.
   */
  public ComplexNumber sech() {
    ComplexNumber result = divide(new ComplexNumber(1.0,0.0),cosh());
    return new ComplexNumber(result.Re(),result.Im());
  }
  public static ComplexNumber sech(ComplexNumber z) {
    return z.sech();
  }

  /**
   * Returns the csch of the ComplexNumber.
   */
  public ComplexNumber csch() {
    ComplexNumber result = divide(new ComplexNumber(1.0,0.0),sinh());
    return new ComplexNumber(result.Re(),result.Im());
  }
  public static ComplexNumber csch(ComplexNumber z) {
    return z.csch();
  }

  /**
   * Returns the tanh of the ComplexNumber.
   */
  public ComplexNumber coth() {
    ComplexNumber result = divide(cosh(),sinh());
    return new ComplexNumber(result.Re(),result.Im());
  }
  public static ComplexNumber coth(ComplexNumber z) {
    return z.coth();
  }

  // STRING RELATED METHODS FOR THE OBJECT

  /**
   * Parses the string and returns a complex number.
   * This function will create valid complex numbers for the 
   * following types of examples:
   * <ul><i>
   * <li>5.3
   * <li>10 + 5i
   * <li>5.3 - 1.1i
   * <li>-7.3 + 2i
   * <li>5-3i
   * </ul>
   *
   * @param value  the string that contains the value.
   * @returns  a new ComplexNumber representing the strings value.
   * @exception  NumberFormatException  if the string doesn't contain
   *             a complex number.
   */
  public static ComplexNumber valueOf(String value)
     throws NumberFormatException {
    boolean inNumber = false;
    int start = 0,
        count = 0;
    double a = 0,
           b = 0;
    int length = value.length();
    char c, sign = ' ';

    for (int i = 0; i < length; i++)
    {
      c = value.charAt(i);

      if ((c == ' ' || c == '\t') && inNumber)
      {
        a = Double.valueOf(value.substring(start, i)).doubleValue();
        inNumber = false;
        count++;
      }
      if ((c == '+' || c == '-') && inNumber)
      {
        a = Double.valueOf(value.substring(start, i)).doubleValue();
        sign = c;
        inNumber = false;
        count++;
      }
      else if (c == 'i' && inNumber)
      {
        b = Double.valueOf(value.substring(start, i)).doubleValue();

        // Check the sign.  Only need to change it if it is negative
        if (sign == '-') b = -b;
        else if (sign != '+') throw new NumberFormatException();

        inNumber = false;
        count++;
      }
      else if ((c == '+' || c == '-') && count == 1) sign = c;
      else if (!inNumber && c != ' ' && c != '\t')
      {
        start = i;
        inNumber = true;
      }
    }

    if (count == 0) a = Double.valueOf(value).doubleValue();
    if (count > 2) throw new NumberFormatException();

    return new ComplexNumber(a,b);
  }

  /**
   * Returns the String representation of the complex number.
   */
  public String toString() {
    if(b != 0.0) {
      if(a!= 0.0) {
        if(b<0) {
          return new Double(a).toString() + new Double(b).toString() + "i";
        } else {
          return new Double(a).toString() + "+" + new Double(b).toString() + "i";
        }
      } else {
        return new Double(b).toString() + "i";
      }
    } else {
      return new Double(a).toString();
    }
  }

  // MAIN METHOD FOR TESTING

  // this routine is here only to test the functionality of the model tool
  // it self - the following will input a specific type of model and will base
  // the results on this model. the following code will just set up and
  // evaluate the model and then output results.
  public static void main(String args[]) {
    ComplexNumber cn1 = new ComplexNumber(1.0,1.0);
    ComplexNumber cn2 = new ComplexNumber(0.0,1.0);
    ComplexNumber cn3 = new ComplexNumber(1.0,-1.0);

    System.out.println("\n****** Test #1 Started **************\n");
    System.out.println("cn1 =         " + cn1.toString());
    System.out.println("cn1 real?     "+new Boolean(cn1.isReal()).toString());
    System.out.println("cn1 imag?     "+new Boolean(cn1.isImaginary()).toString());
    System.out.println("cn2 =         " + cn2.toString());
    System.out.println("cn2 real?     "+new Boolean(cn2.isReal()).toString());
    System.out.println("cn2 imag?     "+new Boolean(cn2.isImaginary()).toString());
    System.out.println("cn3 =         " + cn3.toString());
    System.out.println("cn3 real?     "+new Boolean(cn3.isReal()).toString());
    System.out.println("cn3 imag?     "+new Boolean(cn3.isImaginary()).toString());
    System.out.println("cn1 + cn2 =   " + add(cn1,cn2).toString());
    System.out.println("cn1 - cn2 =   " + subtract(cn1,cn2).toString());
    System.out.println("cn1 * cn2 =   " + multiply(cn1,cn2).toString());
    System.out.println("cn1 / cn2 =   " + divide(cn1,cn2).toString());
    System.out.println("cn1 / cn3 =   " + divide(cn1,cn3).toString());
    System.out.println("cn1 =         " + cn1.toString());
    System.out.println("conj(cn1) =   " + cn1.conjugate());
    System.out.println("abs(cn1) =    " + cn1.abs());
    System.out.println("Re(cn1) =     " + cn1.Re());
    System.out.println("Im(cn1) =     " + cn1.Im());
    cn1.negate();
    System.out.println("- cn1 =       " + cn1.toString());
    ComplexNumber cn4 = new ComplexNumber(cn1);
    System.out.println("copy of cn1 = " + cn4.toString());
    cn4 = cn1.sin();
    System.out.println("sin(cn1) =    " + cn4.toString());
    cn4 = cn1.cos();
    System.out.println("cos(cn1) =    " + cn4.toString());
    cn4 = cn2.sin();
    System.out.println("sin(cn2) =    " + cn4.toString());
    cn4 = cn2.cos();
    System.out.println("cos(cn2) =    " + cn4.toString());
    ComplexNumber cn5 = new ComplexNumber(3.14159, 0.0);
    cn4 = cn5.sin();
    System.out.println("sin(cn5) =   " + cn4.toString());
    System.out.println("sin(cn5) =   " + cn2.sin(cn5));
    cn4 = cn5.cos();
    System.out.println("cos(cn5) =   " + cn4.toString());
    System.out.println("cos(cn5) =   " + cn2.cos(cn5));
    cn4 = cn5.tan();
    System.out.println("tan(cn5) =   " + cn4.toString());
    System.out.println("tan(cn5) =   " + cn2.tan(cn5));
    cn4 = cn5.cot();
    System.out.println("cot(cn5) =   " + cn4.toString());
    System.out.println("cot(cn5) =   " + cn2.cot(cn5));
    cn4 = cn5.sec();
    System.out.println("sec(cn5) =   " + cn4.toString());
    System.out.println("sec(cn5) =   " + cn2.sec(cn5));
    cn4 = cn5.csc();
    System.out.println("csc(cn5) =   " + cn4.toString());
    System.out.println("csc(cn5) =   " + cn2.csc(cn5));
    cn4 = cn5.exp();
    System.out.println("exp(cn5) =   " + cn4.toString());
    System.out.println("exp(cn5) =   " + cn2.exp(cn5));
    cn4 = cn5.log();
    System.out.println("log(cn5) =   " + cn4.toString());
    System.out.println("log(cn5) =   " + cn2.log(cn5));
    cn4 = cn5.sqrt();
    System.out.println("sqrt(cn5) =   " + cn4.toString());
    System.out.println("sqrt(cn5) =   " + cn2.sqrt(cn5));
    cn4 = cn5.pow(0.5);
    System.out.println("pow(cn5) =   " + cn4.toString());
    System.out.println("pow(cn5) =   " + cn2.pow(cn5,0.5));
    cn4 = cn5.pow(1.5);
    System.out.println("pow(cn5) =   " + cn4.toString());
    System.out.println("pow(cn5) =   " + cn2.pow(cn5,1.5));
    ComplexNumber zero = new ComplexNumber(0.0, 0.0);
    cn4 = zero.cot();
    System.out.println("cot(0.0) =   " + cn4.toString());
    System.out.println("cot(0.0) =   " + cn2.cot(cn4));
    System.out.println("log(0.0) =   " + cn4.toString());
    System.out.println("log(0.0) =   " + cn2.log(cn4));
    System.out.println("pow(-1.0) =   " + cn4.toString());
    System.out.println("pow(-1.0) =   "
                 + cn2.pow(new ComplexNumber(-1.,0.),0.5));
    try {
      System.out.println(ComplexNumber.valueOf("3.0-2.1i"));
    } catch (NumberFormatException nfe) {
      System.out.println("Number Format Exceotion -1- Encountered");
    }
    try {
      System.out.println(ComplexNumber.valueOf("3.0a-2.1i"));
    } catch (NumberFormatException nfe) {
      System.out.println("Number Format Exception -2- Encountered");
    }
    try {
      System.out.println(ComplexNumber.valueOf("3.0-2.1"));
    } catch (NumberFormatException nfe) {
      System.out.println("Number Format Exception -3- Encountered");
    }
    try {
      System.out.println(ComplexNumber.valueOf("3.0-2.1i"));
    } catch (NumberFormatException nfe) {
      System.out.println("Number Format Exceotion -4- Encountered");
    }
    System.out.println("\nValues of the cot function:\n");
    double dx = 0.1;
    double x = -1.0;
    for(int i=0; i<21; i++) {
      ComplexNumber cn = new ComplexNumber(x,0.0);
      System.out.println("for x = " + x + " cot = " + cn.cot());
      x = x + dx;
    }
    System.out.println("for x = " + 0.0 + " cot = " + ComplexNumber.cot(new ComplexNumber(0.0,0.0)));
    System.out.println("\nValues of the cosh function:\n");
    dx = 0.1;
    x = -1.0;
    for(int i=0; i<21; i++) {
      ComplexNumber cn = new ComplexNumber(x,0.0);
      System.out.println("for x = " + x + " cosh = " + cn.cosh());
      x = x + dx;
    }
    System.out.println("\nValues of the sinh function:\n");
    dx = 0.1;
    x = -1.0;
    for(int i=0; i<21; i++) {
      ComplexNumber cn = new ComplexNumber(x,0.0);
      System.out.println("for x = " + x + " sinh = " + cn.sinh());
      x = x + dx;
    }
    System.out.println("\nValues of the tanh function:\n");
    dx = 0.1;
    x = -1.0;
    for(int i=0; i<21; i++) {
      ComplexNumber cn = new ComplexNumber(x,0.0);
      System.out.println("for x = " + x + " tanh = " + cn.tanh());
      x = x + dx;
    }
    System.out.println("\nValues of the sech function:\n");
    dx = 0.5;
    x = -5.0;
    for(int i=0; i<21; i++) {
      ComplexNumber cn = new ComplexNumber(x,0.0);
      System.out.println("for x = " + x + " sech = " + cn.sech());
      x = x + dx;
    }
    System.out.println("\n****** Test #1 Completed **************\n");

  }

}
