/*
 *
 *  Object Name: MapDisplay
 *  Purpose:     This is a stripped down version of MapEditor
 *  Author:      Joe Koebbe
 *
 */
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.util.Vector;

public class MapDisplay extends Panel implements ActionListener {
  public MapDisplay(Frame f, int w, int h, int xp, int yp) {
    this(f, w, h, xp, yp, null, -1, -1, -1);
  }
  public MapDisplay(Frame f, int w, int h, int xp, int yp,
                    Double [][] d, int nx, int ny, int nz) {
    containingFrame = f;
    widthInput = w;
    heightInput = h;
    xPosition = xp;
    yPosition = yp;
    data = d;
    nmapx = nx;
    nmapy = ny;
    nmapz = nz;
    blockSize();
    init();
  }
  public void init() {

    runtime = Runtime.getRuntime();
    toolkit = Toolkit.getDefaultToolkit();

    screensize = toolkit.getScreenSize();
    screensize.width = 98 * screensize.width / 100;
    screensize.height = 98 * screensize.height / 100;
    screenresolution = toolkit.getScreenResolution();

    setLayout(new GridLayout(0,1));
    setBackground(panelColor);
      mapCanvas = new ImageCanvas(mapCanvasWidth,mapCanvasHeight);
      setMapCanvasSize(nmapx, nmapy);
      mapCanvasPanel = new Panel();
      mapCanvasPanel.setLayout(new GridLayout(1,0));
      mapCanvasPanel.setBackground(Color.white);
        mapCanvasPanel.add(mapCanvas);
        mapCanvasPanel.setSize(mapCanvasPanelWidth,mapCanvasPanelHeight);
    add(mapCanvasPanel);

    // set the size of the panel for the display
    setSize(mapCanvasWidth, mapCanvasHeight);

    // make sure the calling frame is large enough - note that calling the
    // pack routine is necessary before the insets can be retrieved
    containingFrame.pack();
    Insets insets = containingFrame.getInsets();
    containingFrame.setSize(mapCanvasWidth+insets.left+insets.right,
                            mapCanvasHeight+insets.top+insets.bottom);

    // set the location of the frame
    Point p = this.getLocation();
    p.translate(xPosition, yPosition);
    setLocation(p);

  }

  public void paint(Graphics g) {
    showMap();
  }

  public void blockSize() {
    if(nmapx >= nmapy) {
      dx = widthInput / nmapx;
      dy = dx;
      dxBlock = dx;
      dyBlock = dy;
    } else {
      dy = heightInput / nmapy;
      dx = dy;
      dxBlock = dx;
      dyBlock = dy;
    }
  }

  public void setMapCanvasSize(int nx, int ny) {
    mapCanvasWidth = dx * nx;
    mapCanvasHeight = dy * ny;
    mapCanvas.setSize(mapCanvasWidth,mapCanvasHeight);
  }

  public void setData(Double [][] d, int nx, int ny, int nz) {
    data = d;
    nmapx = nx;
    nmapy = ny;
    nmapz = nz;
    blockSize();
    showMap();
  }

  public void showMap() {
    int xval, yval;
    int i, j;
    int j2, k2, ind;
    float blockvalue, rval, gval, bval;
    Color color;

    // check to see if there is any data
    if(data == null) return;

    // set the canvas size
    setMapCanvasSize(nmapx, nmapy);

    computeMinAndMaxValues();

    // build the off screen image to draw to for the computations
    offScreenImage = null;
    offScreenImage = createImage(mapCanvasWidth,mapCanvasHeight);
    offScreenImageGraphics = offScreenImage.getGraphics();

    // set the current color for the graphics context
    color = offScreenImageGraphics.getColor();

    // clear the whole deal first
    offScreenImageGraphics.setColor(panelColor);
    offScreenImageGraphics.fillRect(0, 0, mapCanvasWidth, mapCanvasHeight);

    // compute the third dimension shift
    k2 = ( sliceIndex - 1 ) * nmapx * nmapy;

    // set the colors of the blocks in the map
    for (j=1; j<=nmapy; j++) {
      j2 = ( j - 1 ) * nmapx + k2;
      // get the cell boundary pixels in the second coordinate
      yval = mapCanvasHeight - j * dy + 1; 
      for (i=1; i<=nmapx; i++) {
        ind = i + j2 - 1;
        // get the cell boundary pixels in the first coordinate
        xval = ( i - 1 ) * dx + 1;

        // get the value in the map block
        blockvalue = data[ind][0].floatValue();
        if(minimumValue != maximumValue) {
          rval = gval = bval = (float) (1.0 - (( blockvalue - minimumValue )
                        / ( maximumValue - minimumValue)));
        } else {
          rval = gval = bval = (float) averageValue;
        }
        if(rval < (float) 0.0) rval = (float) 0.0;
        if(gval < (float) 0.0) gval = (float) 0.0;
        if(bval < (float) 0.0) bval = (float) 0.0;
        if(rval > (float) 1.0) rval = (float) 1.0;
        if(gval > (float) 1.0) gval = (float) 1.0;
        if(bval > (float) 1.0) bval = (float) 1.0;
        
        offScreenImageGraphics.setColor(new Color(rval,gval,bval));
        offScreenImageGraphics.fillRect(xval, yval, dx, dy);
      }
    }

    // draw the boundaries
    offScreenImageGraphics.setColor(foreGroundColor);
    offScreenImageGraphics.drawLine(0, 0, mapCanvasWidth-1, 0);
    offScreenImageGraphics.drawLine(mapCanvasWidth-1, 0,
		                    mapCanvasWidth-1, mapCanvasHeight-1);
    offScreenImageGraphics.drawLine(0, mapCanvasHeight-1,
                                    mapCanvasWidth-1, mapCanvasHeight-1);
    offScreenImageGraphics.drawLine(0, 0, 0, mapCanvasHeight-1);

    // check to see if the grid is toggled on
    if (gridOn) {
      offScreenImageGraphics.setColor(foreGroundColor);

      // draw the horizontal grid lines
      j = 1;
      yval = 0;
      while (j<nmapy) {
        yval = yval + dy;
        offScreenImageGraphics.drawLine(0,yval, mapCanvasWidth,yval);
        j++;
      }
  
      // draw the horizontal grid lines
      i = 1;
      xval = 0;
      while (i<nmapx) {
        xval = xval + dx;
        offScreenImageGraphics.drawLine(xval,0,xval,mapCanvasHeight);
        i++;
      }

    }

    // reset the color for the context to the original
    offScreenImageGraphics.setColor(color);

    // get rid of the graphics context for now
    offScreenImageGraphics.dispose();

    mapCanvas.redraw(0,0,offScreenImage);

  }

  public void nextSlice() {
    sliceIndex = sliceIndex + 1;
    if(sliceIndex > nmapz) sliceIndex = 1;
    showMap();
  }

  public void nextComponent() {
    componentIndex = componentIndex + 1;
    if(componentIndex > 8) componentIndex = 0;
    showMap();
  }

  public void previousSlice() {
    sliceIndex = sliceIndex + 1;
    if(sliceIndex < 1) sliceIndex = nmapz;
    showMap();
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *  method for computing the minimum and maximum values in the current map *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  public void computeMinAndMaxValues() {
    if(data == null) return;
    int n = data.length;
    if(data[0][0] == null) return;
    double dval = data[0][0].doubleValue();
    minimumValue = dval;
    maximumValue = minimumValue;
    for(int i=1; i<n; i++) {
      if(data[0][0] != null) {
        dval = data[i][0].doubleValue();
        if(minimumValue > dval) minimumValue = dval;
        if(maximumValue < dval) maximumValue = dval;
      }
    }
    // compute the average value for some uses
    averageValue = 0.5 * ( maximumValue + minimumValue );
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   * methods for getting the values assigned to cells in the map             *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  public Dimension getPreferredSize() {
    return new Dimension(WIDTH, HEIGHT);
  }

  public Dimension getMinimumSize() {
    return new Dimension(minimumWidth,minimumHeight);
  }

  public Dimension getMaximumSize() {
    return new Dimension(WIDTH, HEIGHT);
  }

  public void actionPerformed(ActionEvent e) {
    Object source = e.getSource();
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   * main() - called when the application is run.                            *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  public static void main(String args[]) {
    Frame f = new Frame("Map Display Frame");
    int nx = 10;
    int ny = 10;
    int nz = 1;
    int nn = nx * ny * nz;
    Double [][] d = new Double[nn][9];
    for(int i=0; i<nn; i++) d[i][0] = new Double(i);
    MapDisplay mapDisplay = new MapDisplay(f, 200, 200, 50, 50, d, nx, ny, nz);
    f.add("Center", mapDisplay);
    for(int i=0; i<nn; i++) d[i][0] = new Double(nn-i);
    mapDisplay.setData(d, nx, ny, nz);
    f.show();
  }

  // the basic panel variables for the pieces
  private Frame containingFrame;
  private Panel mainPanel;

  // the map canvas area
  private Panel mapCanvasPanel;
  private int mapCanvasPanelWidth;
  private int mapCanvasPanelHeight;
  private ImageCanvas mapCanvas;
  private int mapCanvasWidth;
  private int mapCanvasHeight;
  private int mapCanvasX;
  private int mapCanvasY;

  private double minimumValue = 0.0;
  private double maximumValue = 1.0;
  private double averageValue;
  private int xBlock = 0;
  private int yBlock = 0;
  private int sliceIndex = 1;
  private int componentIndex = 0;

  // variables for the off screen image being created
  private Image offScreenImage;
  private Graphics offScreenImageGraphics;

  // map grid variables
  private int nmapx;
  private int nmapy;
  private int nmapz;
  private int ntot;
  private int dx, dxBlock;
  private int dy, dyBlock;

  // map values in a linear array
  private Double [][] data = null;
  public boolean gridOn = true;

  // color map
  private Vector colorMap;

  // file stuff
  // the runtime object for the application
  Runtime runtime;
  Toolkit toolkit;

  // screen size variables
  private Dimension screensize;
  private int screenresolution;

  // color variables
  private Color backGroundColor = Color.white;
  private Color foreGroundColor = Color.black;
  private Color panelColor = new Color(196,196,132);
  private Color labelColor = new Color(110,132,255);

  // frame size
  private static int WIDTH = 500;
  private static int HEIGHT = 500;
  private static int minimumWidth = 500;
  private static int minimumHeight = 500;
  private int widthInput = -1;
  private int heightInput = -1;
  private int xPosition = 0;
  private int yPosition = 0;

}
