/*
 *
 *  Object Name: MapEditor
 *  Purpose:     This is a map editor object which can be used in any
 *               application.
 *  Author:      Joe Koebbe
 *
 *
 *  Note that this object does not build a control panel. Instead, it provides
 *  methods, variables, and other tools to control what goes on in the editor.
 *  The developer can build an interface and attach a control panel that uses
 *  the methods and such.
 *
 */
import java.applet.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.io.*;
import java.util.Vector;
import java.util.StringTokenizer;

public class MapEditor extends Panel implements ActionListener, MouseListener,
  MouseMotionListener {

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *  Constructors for the class                                             *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  // constructor with all arguments
  public MapEditor(Frame f) {
    this(f, -1, -1, 0, 0, null);
  }

  // constructor with all arguments
  public MapEditor(Frame f, int w, int h, int xp, int yp, boolean [] cm) {

    // set the calling frame for the panel
    containingFrame = f;
    widthInput = w;
    heightInput = h;
    xPosition = xp;
    yPosition = yp;
    controlMask = cm;

    // this is it
    init();

  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *  applet methods                                                         *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  public void init() {
    int width, height;

    // wait for some memory if necessary
    runtime = Runtime.getRuntime();
    toolkit = Toolkit.getDefaultToolkit();

    // get the OS file separator
    fileSeparator = File.separator;

    // allow the object to get events without a listener if necessary
    enableEvents(AWTEvent.WINDOW_EVENT_MASK);

    // start by getting the screen size and resolution - these are used to set
    // the size of the pieces of the GUI.
    screensize = toolkit.getScreenSize();
    screensize.width = 98 * screensize.width / 100;
    screensize.height = 98 * screensize.height / 100;
    screenresolution = toolkit.getScreenResolution();

    if(controlMask == null) {
      controlMask = new boolean [ncontrols];
      for(int i=0; i<ncontrols; i++) controlMask[i] = true;
    } else {
      if(controlMask.length < ncontrols) {
        boolean [] cm = new boolean [ncontrols];
        for(int i=0; i<controlMask.length; i++) cm[i] = controlMask[i];
        for(int i=controlMask.length; i<ncontrols; i++) cm[i] = true;
        controlMask = cm;
      }
    }

    // build the drawing area
    buildMapCanvas();

    // build the domain interface area
    buildMapDomainControl();

    // build the controls for the editor
    buildMapEditorControls();

    // build the default generator regions
    buildMapDefaultGeneration();

    // build the message panel
    buildMessagePanel();
 
    // set up the main panel
    mainPanel = new Panel();
    mainPanel.setBackground(panelColor);
    mainPanel.setLayout(new BorderLayout());

      // redo on the set up
      displayPanel = new Panel();
      displayPanel.setLayout(new BorderLayout());
        domainPanel = new Panel();
        domainPanel.setLayout(new GridLayout(0,1));
        domainPanel.add(mapDomainPanel);
      displayPanel.add("Center", mapCanvasPanel);
      displayPanel.add("East", mapEditorControls);
      displayPanel.add("South", domainPanel);

      manipulationPanel = new Panel();
      manipulationPanel.setLayout(new GridLayout(0,1));
      otherPanel = new Panel();
      otherPanel.setLayout(new GridLayout(0,1));
        otherPanel.add(defaultGenerationPanel);
        otherPanel.add(messagePanel);
      manipulationPanel.add(otherPanel);
   
    mainPanel.add("Center", displayPanel);
    mainPanel.add("South", manipulationPanel);

    // set the layout manager
    this.setLayout(new GridLayout(0,1));

    // add the whole deal to the applet
    this.add(mainPanel);

    // compute the default block size, the dimensions, and initialize the array
    // of values
    defaultBlockSize();
    mapDimensions();
    buildMapArray();

    // initialize the values in the map
    initMapValues();

    // set the size for the frame and all
    if(widthInput == -1) {
      width = WIDTH;
    } else {
      width = widthInput;
      // the next couple of lines guarantee that the map editor does not take
      // more than one half the screen
      int scrw = screensize.width / 2;
      if(scrw < WIDTH) width = scrw;
    }
    if(heightInput == -1) {
      height = heightInput;
    } else {
      height = HEIGHT;
      if(screensize.height < HEIGHT) height = screensize.height;
    }
    if(width < minimumWidth) width = minimumWidth;
    if(height < minimumHeight) height = minimumHeight;
    setSize(width, height);

    // make sure the calling frame is large enough
    containingFrame.setSize(width, height);

    // set the location of the frame
    Point p = this.getLocation();
    p.translate(xPosition, yPosition);
    setLocation(p);

  }

  public void start() {
  }

  public void stop() {
  }

  public void finalize() {
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *  paint method                                                           *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  public void paint(Graphics g) {
    showMap();
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *  methods for setting block sizes and such in the map                    *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  public void defaultBlockSize() {

    // see if the aspect ratio is 1
    if(nmapx == nmapy) {
      // do some screen resolution work
      if(0<nmapx && nmapx<=50) dx = 8;
      if(50<nmapx && nmapx<=75) dx = 7;
      if(75<nmapx && nmapx<=100) dx = 6;
      if(100<nmapx && nmapx<=125) dx = 4;
      if(125<nmapx && nmapx<=150) dx = 3;
      if(150<nmapx && nmapx<=175) dx = 2;
      if(175<nmapx) dx = 1;
      dy = dx;
      // set the block map sizes
      dxBlock = dx;
      dyBlock = dy;
      return;
    }

    // figure out which dimension is smallest - and preserve the aspect ratio
    if(nmapx > nmapy) {

      // do some screen resolution work
      if(0<nmapx && nmapx<=50) dx = 8;
      if(50<nmapx && nmapx<=75) dx = 7;
      if(75<nmapx && nmapx<=100) dx = 6;
      if(100<nmapx && nmapx<=125) dx = 4;
      if(125<nmapx && nmapx<=150) dx = 3;
      if(150<nmapx && nmapx<=175) dx = 2;
      if(175<nmapx) dx = 1;

      // use a 1:1 aspect ratio
      dy = dx;

      // set the block map sizes
      dxBlock = dx;
      dyBlock = dy;

    } else {

      if(0<nmapy && nmapy<=50) dy = 8;
      if(50<nmapy && nmapy<=75) dy = 7;
      if(75<nmapy && nmapy<=100) dy = 6;
      if(100<nmapy && nmapy<=125) dy = 4;
      if(125<nmapy && nmapy<=150) dy = 3;
      if(150<nmapy && nmapy<=175) dy = 2;
      if(175<nmapy) dy = 1;

      // use a 1:1 aspect ratio
      dx = dy;

      // set the block map sizes
      dxBlock = dx;
      dyBlock = dy;

    }

  }

  public int getXBlockSize() {
    return dx;
  }

  public int getYBlockSize() {
    return dy;
  }

  public void setBlockSize(int dxval, int dyval) {
    dx = dxval;
    dy = dyval;
    dyBlock = dy;
    dyBlock = dy;
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *  map methods for computing the map, zooming in and out, etc.            *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  public void buildMapArray() {

    // set the size of the array of values
    ntot = nmapx * nmapy * nmapz;
    mapValues = new Double [ntot][9];
    displayMapValues = mapValues;
    dispnmapx = nmapx;
    dispnmapy = nmapy;
    // set up the temporary storage array for use in various spots
    tempMapValues = new Double [ntot][9];
    
  }

  public void mapDimensions() {
    // set the lengths in pixels
    xlength = dispnmapx * dx;
    ylength = dispnmapy * dy;
  }

  public void setMapCanvasSize(int nx, int ny) {
    // compute the number of pixels to be used
    mapCanvasWidth = dx * nx + 2 * xoffset;
    mapCanvasHeight = dy * ny + 2 * yoffset;
    // set the size of the canvas
    mapCanvas.setSize(mapCanvasWidth,mapCanvasHeight);
  }

  public void initMapValues() {
    int i,j,k,k2,j2,ind;
    for (k=1; k<=nmapz; k++) {
      k2 = ( k - 1 ) * nmapx * nmapy;
      for (j=1; j<=nmapy; j++) {
        j2 = ( j - 1 ) * nmapx + k2;
        for (i=1; i<=nmapx; i++) {
          ind = i + j2 - 1;
          for(int l=0; l<9; l++) mapValues[ind][l] = new Double(0.0);
          mapValues[ind][0] = new Double(1.0);
          mapValues[ind][4] = new Double(1.0);
          mapValues[ind][8] = new Double(1.0);
        }
      }
    }
    subregionValues = mapValues;
    subnmapx = nmapx;
    subnmapy = nmapy;
    displayMapValues = mapValues;
    dispnmapx = nmapx;
    dispnmapy = nmapy;
    mapLoaded = true;
  }

  public void refineGrid() {

    // set up the temporary values for the refinement
    int tnmapx = 2 * nmapx;
    int idble = 2;
    if(nmapx == 1) {
      tnmapx = nmapx;
      idble = 1;
    }
    int tnmapy = 2 * nmapy;
    int jdble = 2;
    if(nmapy == 1) {
      tnmapy = nmapy;
      jdble = 1;
    }
    int tnmapz = 2 * nmapz;
    int kdble = 2;
    if(nmapz == 1) {
      tnmapz = nmapz;
      kdble = 1;
    }
    int tntot = tnmapx * tnmapy * tnmapz;
    Double [][] tmapValues = new Double[tntot][9];

    // set up the new values
    for(int k=1; k<=nmapz; k++) {
      int k2 = ( k - 1 ) * nmapx * nmapy;
      for(int kk=0; kk<kdble; kk++) {
        int k22 = ( kdble * ( k - 1 ) + kk ) * tnmapx * tnmapy;
        for (int j=1; j<=nmapy; j++) {
          int j2 = ( j - 1 ) * nmapx + k2;
          for(int jj=0; jj<jdble; jj++) {
            int j22 = ( jdble * ( j - 1 ) + jj ) * tnmapx + k22;
            for (int i=1; i<=nmapx; i++) {
              for(int ii=0; ii<idble; ii++) {
                int ind = i + j2 - 1;
                int iind = idble * ( i - 1 ) + ii + j22;
                for(int l=0; l<9; l++) tmapValues[iind][l] = mapValues[ind][l];
              }
            }
          }
        }
      }
    }

    // set the current values
    nmapx = tnmapx;
    nmapy = tnmapy;
    nmapz = tnmapz;
    ntot = tntot;
    mapValues = new Double[ntot][9];
    for(int l=0; l<9; l++) {
      for(int i=0; i<ntot; i++) {
          mapValues[i][l] = new Double(tmapValues[i][l].toString());
      }
    }
    displayMapValues = mapValues;
    dispnmapx = nmapx;
    dispnmapy = nmapy;

    // set the temporary map and redraw things
    setTempMapValues();
    defaultBlockSize();
    redrawMap();

    // keep track of the fact the grid is refined
    gridRefinementLevel++;
    gridRefined = true;

    // set the saved value to false since there is a new map
    mapModified = true;
    newMap = true;
    mapLoaded = true;

  }

  public void unrefineGrid() {

    if(!gridRefined) {
      printMessage("Grid has not been refined");
      return;
    }

    // set up the temporary values for the refinement
    int tnmapx = nmapx / 2;
    int idble = 2;
    if(nmapx == 1) {
      tnmapx = nmapx;
      idble = 1;
    }
    int tnmapy = nmapy / 2;
    int jdble = 2;
    if(nmapy == 1) {
      tnmapy = nmapy;
      jdble = 1;
    }
    int tnmapz = nmapz / 2;
    int kdble = 2;
    if(nmapz == 1) {
      tnmapz = nmapz;
      kdble = 1;
    }
    int tntot = tnmapx * tnmapy * tnmapz;
    Double [][] tmapValues = new Double[tntot][9];

    // set up the new values
    for(int k=1; k<=tnmapz; k++) {
      int k2 = ( k - 1 ) * tnmapx * tnmapy;
      int k22 = kdble * ( k - 1 ) * nmapx * nmapy;
      for (int j=1; j<=tnmapy; j++) {
        int j2 = ( j - 1 ) * tnmapx + k2;
        int j22 = jdble * ( j - 1 ) * nmapx + k22;
        for (int i=1; i<=tnmapx; i++) {
          int ind = i + j2 - 1;
          int iind = idble * ( i - 1 ) + j22;
            for(int l=0; l<9; l++) tmapValues[ind][l] = mapValues[iind][l];
        }
      }
    }

    // set the current values
    nmapx = tnmapx;
    nmapy = tnmapy;
    nmapz = tnmapz;
    ntot = tntot;
    mapValues = new Double[ntot][9];
    for(int l=0; l<9; l++) {
      for(int i=0; i<ntot; i++) {
          mapValues[i][l] = new Double(tmapValues[i][l].toString());
      }
    }
    displayMapValues = mapValues;
    dispnmapx = nmapx;
    dispnmapy = nmapy;

    // set the temporary map and redraw things
    setTempMapValues();
    defaultBlockSize();
    redrawMap();

    // keep track of the fact the grid is refined
    gridRefinementLevel--;
    if(gridRefinementLevel <= 0 ) gridRefined = false;

    // set the saved value to false since there is a new map
    mapModified = true;
    newMap = true;
    mapLoaded = true;

  }

  public void zoomUp() {
    if(dx <= 127 && dy <= 127) {
      setTempMapValues();
      dx = 2 * dx;
      dy = 2 * dy;
      if(!blockMap) togglePixelVsBlock();
      getTempMapValues();
      redrawMap();
      printMessage("Map Zoomed Up");
    }
  }

  public void zoomDown() {
    if(dx > 5 && dy > 5) {
      setTempMapValues();
      dx = dx / 2;
      dy = dy / 2;
      getTempMapValues();
      redrawMap();
      printMessage("Map Zoomed Down");
    }
  }

  public void defaultSize() {
    setTempMapValues();
    defaultBlockSize();
    getTempMapValues();
    redrawMap();
    printMessage("Map Drawn At Default Size");
  }

  public void grabSubregion() {
    // check to see if the user is setting values - must turn off the mouse
    // events associated with that process
    if(doSetValues) toggleSetMapValues();
    if(!mapShowing) redrawMap();
    // now check for the grab action
    if(grabSome) {
      grabSome = false;
      mapEditorGrabSomeButton.setLabel("Grab A Subregion");
      // set up the rectangle for the subregion and the dimensions
      if(subX2 < subX1) {
        int tx = subX2;
        subX2 = subX1;
        subX1 = tx;
      }
      if(subY2 < subY1) {
        int ty = subY2;
        subY2 = subY1;
        subY1 = ty;
      }
      subxlength = subX2 - subX1;
      subylength = subY2 - subY1;
      // compute the pairs of indices associated with the corners of the
      // subregion
      subindx1 = (int) subX1 / dx + 1;
      subindy1 = (int) subY1 / dy + 1;
      subindx2 = (int) subX2 / dx + 1;
      subindy2 = (int) subY2 / dy + 1;
      // compute the total number of values in the subregion
      subnmapx = subindx2 - subindx1 + 1;
      subnmapy = subindy2 - subindy1 + 1;
      subntot = subnmapx * subnmapy;
      // compute the map dimensions
      subxlength = subnmapx * dx;
      subylength = subnmapy * dy;
      // update the subregion to hold the map values
      setSubregionValues();
      // show the map
      redrawSubregionMap();
    } else {
      // set the parameter and text in the button
      grabSome = true;
      mapEditorGrabSomeButton.setLabel("Stop the Grab");
    }
  }

  public void setSubregionValues() {
    // grab the array of values in the subregion
    subregionValues = new Double [subntot][9];
    int indy1 = nmapy - subindy1 + 1;
    int indy2 = nmapy - subindy2 + 1;
    // the indices are flipped so that the order is changed
    for(int j=indy2; j<=indy1; j++) {
      int j2 = ( j - indy2 ) * subnmapx;
      int j22 = ( j - 1 ) * nmapx;
      for(int i=subindx1; i<=subindx2; i++) {
        int indi = i+j22-1; 
        int indo = i-subindx1+j2; 
        for(int l=0; l<9; l++) {
          subregionValues[indo][l] = mapValues[indi][l]; 
        }
      }
    }
  }

  public void updateSubregion() {
    // grab the array of values in the subregion
    for(int j=subindy1; j<=subindy2; j++) {
      int j2 = ( j - subindy1 ) * subnmapx;
      int j22 = ( j - 1 ) * nmapx;
      for(int i=subindx1; i<=subindx2; i++) {
        int indi = i+j22-1; 
        int indo = i-subindx1+j2; 
        for(int l=0; l<9; l++) {
          subregionValues[indo][l] = mapValues[indi][l]; 
        }
      }
    }
  }

  public void showSubregion() {
    if(subregionShowing) return;
    if(subregionValues != null) {
      updateSubregion();
      redrawSubregionMap();
      mapShowing = false;
      subregionShowing = true;
    }
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *  methods for loading, reloading, and saving a map to a file             *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  public boolean getMapFileInfo(String loadOrSave) {
    String s1=null;
    String tmpDirString;
    String tmpLFString;
    String tmpSFString;
    File tmpFile;

    // build and show the file dialog on the system
    FileDialog fileDialog = new FileDialog(containingFrame,
                                           loadOrSave+" Map File");
    fileDialog.setDirectory(directoryString);

    // show the file dialog frame
    fileDialog.show();

    // upon return, start by getting the file and set the selection in the
    // type menu.
    s1 = fileDialog.getFile();

    // get the directory string and set the file name if the user has chosen
    // a file
    if(s1 != null) {
      tmpDirString = fileDialog.getDirectory();
      if(loadOrSave == "Load") {
        tmpLFString= tmpDirString + s1;
      } else {
        tmpSFString = tmpDirString + s1;
      }
    } else {
      printMessage("No File Chosen");
      return false;
    }

    // test a load file for existence and readability if this test works, then
    // the file is read to be accessed in other locations in the code
    if(loadOrSave == "Load") {
      tmpFile = new File(tmpDirString, s1);
      if(tmpFile.isFile() && tmpFile.canRead()) {
        directoryString = tmpDirString;
        loadFileString = directoryString + s1;
        loadFile = new File(directoryString, s1);
        return true;
      } else {
        printMessage("Load File Either Does Not Exist Or Is Not Readable");
        if(!tmpFile.isFile()) printMessage("File Chosen Does Not Exist");
        if(!tmpFile.canRead()) printMessage("File Is Not Readable");
        return false;
      }
    }

    // test a save file for existence and writeability if this test works, then
    // the file is read to be accessed in other locations in the code
    if(loadOrSave == "Save") {
      tmpFile = new File(tmpDirString, s1);

      // now see if the file exists - if so, reset and overwrite, if not,
      // touch the file to get it there
      if(!tmpFile.isFile()) {
        try {
          Runtime.getRuntime().exec("touch " + tmpFile.toString());
          message = "file: " + tmpFile.toString() + " touched";
          printMessage(message);
        } catch (IOException ioe) {
          message = "IO Exception" + "file: " + tmpFile.toString()
                          + " cannot be touched";
          printMessage(message);
          return false;
        }
      }

      // check to see if the file is writeable - if it is, set the file and
      // directory names and if not return a false status
      if(tmpFile.canWrite()) {
        directoryString = tmpDirString;
        saveFileString = directoryString + s1;
        saveFile = new File(directoryString, s1);
        return true;
      } else {
        printMessage("Save File Is Not Writeable");
        return false;
      }

    }

    // just in case, hit a return here
    printMessage("Unknown File Information Problem");
    return false;

  }

  public void loadNewMap() {
    boolean loadtest;
    // if the information about the file is ok load the file - otherwise, dump
    // out
    if(getMapFileInfo("Load")) {
      loadtest = loadMap();
      defaultBlockSize();
      if(loadtest) redrawMap();
    } else {
      printMessage("New Map Cannot Be Loaded");
    }
  }

  public void reloadMap() {
    boolean loadtest;

    // check to see if a file has been loaded - if not go to loadNewMap
    if(!mapLoaded) {
      printMessage("No Map To Reload - Load A New Map");
      loadNewMap();
    } else {
      loadtest = loadMap();
      if(loadtest) redrawMap();
    }

  }

  public boolean loadMap() {
    FileReader fr;
    LineNumberReader lnr;
    String line;
    int nx, ny, nz;

    // build a file reader
    if(loadFile == null) return false;
    try{fr = new FileReader(loadFile);
      // set up the input stream
      try {lnr = new LineNumberReader(fr);
        // read in the dimensions
        // get the number of values in the first dimension
        line = lnr.readLine();
        if(line != null) {
          nx = getFirstInteger(line);
        } else {
          printMessage("File Format Error In loadMap() - reading nx");
          return false;
        }
        if(nx < 1) {
          printMessage("nx < 0 - check the data set");
          return false;
        }
        // get the number of values in the second dimension
        line = lnr.readLine();
        if(line != null) {
          ny = getFirstInteger(line);
        } else {
          printMessage("File Format Error In loadMap() - reading ny");
          return false;
        }
        if(ny < 1) {
          printMessage("ny < 0 - check the data set");
          return false;
        }
        // get the number of values in the third dimension
        line = lnr.readLine();
        if(line != null) {
          nz = getFirstInteger(line);
        } else {
          printMessage("File Format Error In loadMap() - reading nz");
          return false;
        }
        if(nz < 1) {
          printMessage("nz < 0 - check the data set");
          return false;
        }
        // compute the map dimensions and such
        nmapx = nx;
        nmapy = ny;
        nmapz = nz;
        // set up the array for the data being read
        buildMapArray();
        // loop over the components
        for(int ll=0; ll<9; ll++) {
          componentIndex = ll;
          // set the load index location to 0
          nextMapInputIndex = 0;
          // loop through until enough numbers are retrieved
          while(nextMapInputIndex<ntot) {
            line = lnr.readLine();
            if(line != null) {
              getNumbers(line);
            } else {
              printMessage("File Format Error In loadMap() - reading data");
              return false;
            }
          }
        }
        // set the component index to the first
        componentIndex = 0;
        // finally, close the random access file
        lnr.close();
        } catch (IOException io) {
          printMessage("Cannot Open the LineNumberReader in loadMap()");
          return false;
        }
    } catch (FileNotFoundException fnfe) {
      printMessage("File Not Found Exception from FileReader in loadMap()");
      return false;
    }

    // set the loaded and saved flags - if the code gets here everything went
    // ok - note that if the file is loaded, it is also saved since no
    // modifications have been done to the map yet
    saveFile = loadFile;
    saveFileString = loadFileString;
    mapLoaded = true;
    mapModified = false;

    // compute the default block size from the loaded values
    defaultBlockSize();

    // set the slice index
    sliceIndex = 1;
    // set the component index
    componentIndex = 0;

    return true;

  }

  public void saveMap() {
    boolean fileexists;
    // check to see if the current map has been modified - if the map is not
    // modified there is no need to do this work.
    if(!mapModified) { 
      printMessage("Map Has Not Been Modified Since Last Save");
      return;
    }
    //String outputString;
    //int index;
    //File outputFile;
    // determine the default output file name - this can be used later when the
    // value in the field can be set externally
    //outputString = null;
    //index = 1;
    //fileexists = true;
    //while(fileexists) {
    //  // build the next map output file name
    //  outputString = "MapEditor" + Integer.toString(index) + ".map";
    //  outputFile = new File(directoryString,outputString);
    //  index = index + 1;
    //  if(!outputFile.isFile()) fileexists = false;
    //}

    // test to see if the file exists
    if(getMapFileInfo("Save")) {
      if(writeMap()) mapModified = false;
    } else {
      printMessage("File Does Not Exist In saveMap");
    }

  }

  public void overwriteMap() {
    // check to see if the map has been saved or loaded before - if not, the
    // file name must be retrieved
    if(saveFile != null) {
      if(writeMap()) mapModified = false;
    } else {
      mapModified = true;
      saveMap();
    }
  }

  public boolean writeMap() {
    FileWriter fw;
    String line, numberString;
    int index, nx, ny, nz;
    int power;
    double mantissa=0.0,value;
    int sign=1, exponent=1;
    boolean dataleft = false;

    // build a file writer
    try{fw = new FileWriter(saveFile);

        // print out the number of blocks in each direction
        try{
            // first coordinate
            numberString = Integer.toString(nmapx);
            line = "  " + numberString +
                   "           *** number of grid block in first coordinate\n";
            fw.write(line,0,line.length());
            // second coordinate
            numberString = Integer.toString(nmapy);
            line = "  " + numberString +
                 "           *** number of grid block in second coordinate\n";
            fw.write(line,0,line.length());
            // third coordinate
            numberString = Integer.toString(1);
            line = "  " + numberString +
                 "           *** number of grid block in third coordinate\n";
            fw.write(line,0,line.length());

            // loop over the components in the tensor
            for(int ll=0; ll<9; ll++) {

              index = 0;
              line = " ";
              while (index < ntot) {
                value = Math.abs(mapValues[index][ll].doubleValue());
                if(value == 0.0) {
                  line = line + "  0.000000E+00";
                } else {
                  power = 0;
                  if(value > 0.1) {
                    mantissa = value;
                    while (mantissa >= 1.0) {
                      power = power + 1;
                      mantissa = value / Math.pow(10.,(double) power);
                    }
                    numberString = new Double(mantissa).toString();
                    int cnt = numberString.length();
                    while (cnt < 10) {
                      numberString = numberString + "0";
                      cnt ++;
                    }
                    String mantissaString = numberString.substring(0,8);
                    String exponentString = new Integer(power).toString();
                    if(exponentString.length() == 0)
                          exponentString = "00";
                    if(exponentString.length() == 1)
                          exponentString = "0" + exponentString;
                    if(exponentString.length() > 2)
                          exponentString = "**";
                    if(mapValues[index][ll].doubleValue() >= 0.0) {
                      line = line +"  "+ mantissaString + "E+" + exponentString;
                    } else {
                      line = line +" -"+ mantissaString + "E+" + exponentString;
                    }
                  } else {
                    mantissa = value;
                    while (mantissa <= 0.1) {
                      power = power + 1;
                      mantissa = value * Math.pow(10.,(double)power);
                    }
                    numberString = new Double(mantissa).toString();
                    int cnt = numberString.length();
                    while (cnt < 10) {
                      numberString = numberString + "0";
                      cnt ++;
                    }
                    String mantissaString = numberString.substring(0,8);
                    String exponentString = new Integer(power).toString();
                    if(exponentString.length() == 0)
                          exponentString = "00";
                    if(exponentString.length() == 1)
                          exponentString = "0" + exponentString;
                    if(exponentString.length() > 2)
                          exponentString = "**";
                    if(mapValues[index][ll].doubleValue() >= 0.0) {
                      line = line +"  "+ mantissaString + "E-" + exponentString;
                    } else {
                      line = line +" -"+ mantissaString + "E-" + exponentString;
                    }
                  }
                }

                dataleft = true;
              
                index++;
                if(mod(index,5) == 0) {
                  line = line + "\n";
                  fw.write(line,0,line.length());
                  line = " ";
                  dataleft = false;
                }
              }

              // check for any left over data
              if(dataleft) {
                line = line + "\n";
                fw.write(line,0,line.length());
              }

            }

        } catch (IOException ioea) {
          printMessage("Cannot Write to the File");
          return false;
        }

        // close the file writer
        fw.close();

    } catch (IOException ioe) {
      printMessage("File Not Found Exception from FileWriter");
      return false;
    }

    return true;

  }

  public int mod(int i1, int i2) {
    return i1 - i2 * ((int) i1 / i2);
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   * method for building a map from scratch                                  *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  public void buildAMap() {
    // set the dialog box and do the deal
    BuildAMapDialog bamd
           = new BuildAMapDialog(containingFrame, "Build A Map Frame",
                                 nmapx, nmapy, nmapz);
    bamd.show();
    // check the state of the dialog box to see if the user wants to reset
    // the values
    if(!bamd.getState()) return;

    // try getting the necessary data and test to make sure it is ok
    int tnmapx = bamd.getNumberOfXBlocks();
    int tnmapy = bamd.getNumberOfYBlocks();
    int tnmapz = bamd.getNumberOfZBlocks();
    Double value = bamd.getBackgroundValue();
    // get rid of the dialog for now
    bamd.dispose();
    if(tnmapx<2) return;
    if(tnmapy<2) return;
    // if all is well reset the parameters in the code and show the basic map
    nmapx = tnmapx;
    nmapy = tnmapy;
    nmapz = tnmapz;
    ntot = nmapx * nmapy * nmapz;
    buildMapArray();
    for(int i=0; i<ntot; i++) {
      mapValues[i][0] = value;
      mapValues[i][1] = new Double(0.0);
      mapValues[i][2] = new Double(0.0);
      mapValues[i][3] = new Double(0.0);
      mapValues[i][4] = value;
      mapValues[i][5] = new Double(0.0);
      mapValues[i][6] = new Double(0.0);
      mapValues[i][7] = new Double(0.0);
      mapValues[i][8] = value;
    }
    displayMapValues = mapValues;
    dispnmapx = nmapx;
    dispnmapy = nmapy;
    setTempMapValues();
    defaultBlockSize();
    redrawMap();
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   * methods for printing messages from the code to the gui                  *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  public void printMessage(String s) {
    message = s;
    if (message.endsWith("\n")) {
      messageTextArea.insert(message,0);
    } else {
      messageTextArea.insert(message + "\n",0);
    }
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   * methods for processing input strings and strings for output to files    *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  public int getFirstInteger(String s) {
    // set up a string tokenizer
    StringTokenizer st = new StringTokenizer(s, " \t\r\n");
    // go through the tokens in the string
    while (st.hasMoreTokens()) {
      String token = st.nextToken();
      // try to build an integer
      try{Integer intval = Integer.valueOf(token);
          return intval.intValue();
      } catch (NumberFormatException nfe) {
        if(debug) System.out.println("Number Format Exception");
      }
    }
    return -1;
  }

  public void getNumbers(String s) {
    double mantissa=0.0,value;
    int sign=1, exponent=1; 
    boolean problem;
    // set up a string tokenizer
    StringTokenizer st = new StringTokenizer(s, " \t\r\n");
    // go through the tokens in the string
    while (st.hasMoreTokens()) {
      problem = false;
      String number = st.nextToken();
      StringTokenizer nst = new StringTokenizer(number, " eE");
      int nTokens = nst.countTokens();

      // if the number of tokens above is more than one, then this must be an
      // exponential format if it is a number
      if(nTokens > 1) {

        // get the two pieces - these are the mantissa and the exponent
        String mantissaString = nst.nextToken();
        String signedExponentString = nst.nextToken();
        String signString = signedExponentString.substring(0,1);
        String exponentString = signedExponentString.substring(1,3);

        // try to build the mantissa
        try{Double mantissaDouble = Double.valueOf(mantissaString);
            mantissa = mantissaDouble.doubleValue();
        } catch (NumberFormatException nfe) {
          if(debug) System.out.println("Number Format Exception");
          problem = true;
        }

        // get the sign
        if (signString.equalsIgnoreCase("-")) {
          sign = -1;
        } else {
          sign = 1;
        }

        // try to build the exponent
        try{Integer exponentInteger = Integer.valueOf(exponentString);
            exponent = exponentInteger.intValue();
        } catch (NumberFormatException nfe) {
          if(debug) System.out.println("Number Format Exception");
          problem = true;
        }

        // get the number
        if(!problem) {
          value = Math.exp(Math.log(mantissa) + sign*exponent*Math.log(10.0));
          mapValues[nextMapInputIndex][componentIndex] = new Double(value);
          nextMapInputIndex++;
          if(nextMapInputIndex>=ntot) return;
        } else {
          // this probably means the file is not what you think
          printMessage("Skipping a String");
        }

      // the following is supposed to handle floating point and integer number
      // input
      } else {
        // try to form a number from the string
        try {
          mapValues[nextMapInputIndex][componentIndex] = new Double(number);
          nextMapInputIndex++;
        } catch(NumberFormatException nfe) {
        }
      }

    }

  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *  methods for displaying and clearing the map                            *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  public void showMap() {

    int xval, yval;
    int i, j;
    int j2, k2, ind;
    float blockvalue, rval, gval, bval;
    Color color;

    // set the dimensiosn of the map
    mapDimensions();

    // set the canvas size
    setMapCanvasSize(dispnmapx, dispnmapy);

    // get the minimum and maximum values of the data
    if(rescale) computeMinAndMaxValues();
    int nl = 0;
    if(Math.abs(minimumValue) > 0.0) nl = (int)Math.log(Math.abs(minimumValue));
    int nu = 0;
    if(Math.abs(maximumValue) > 0.0) nu = (int)Math.log(Math.abs(maximumValue));
    int norders = nu - nl;
    if(norders < 1) norders = 1;
    double dorder = 0.25 / ((double) norders);

    // build the off screen image to draw to for the computations
    offScreenImage = null;
    offScreenImage = createImage(mapCanvasWidth,mapCanvasHeight);
    offScreenImageGraphics = offScreenImage.getGraphics();

    // set the current color for the graphics context
    color = offScreenImageGraphics.getColor();

    // clear the whole deal first
    offScreenImageGraphics.setColor(panelColor);
    offScreenImageGraphics.fillRect(0, 0, mapCanvasWidth, mapCanvasHeight);

    // compute the third dimension shift
    k2 = ( sliceIndex - 1 ) * dispnmapx * dispnmapy;

    // set the colors of the blocks in the map
    for (j=1; j<=dispnmapy; j++) {
      j2 = ( j - 1 ) * dispnmapx + k2;
      // get the cell boundary pixels in the second coordinate
      yval = ylength - j * dy + 1 + yoffset; 
      for (i=1; i<=dispnmapx; i++) {
        ind = i + j2 - 1;
        // get the cell boundary pixels in the first coordinate
        xval = ( i - 1 ) * dx + 1 + xoffset;

        // get the value in the map block
        if(displayMapValues[ind][componentIndex] != null) {
          blockvalue = displayMapValues[ind][componentIndex].floatValue();
	  if(blockvalue > 0.0) {
            if(minimumValue != maximumValue) {
              float val = (float) (( blockvalue - minimumValue )
                                / ( maximumValue - minimumValue));
	      int frac = norders - ((int) Math.log((maximumValue / val))) - 1;
	      double extra = 0.75 * dorder * ((double) (norders - frac));
              rval = (float) (0.2 * val * ( 1.0 - val ) + extra);
              rval = (float) (0.0);
              gval = (float) (0.2 * val + extra);
              bval = (float) (0.2 * ( 1.0 - val ) + extra);
            } else {
              rval = gval = bval = (float) averageValue;
            }
            if(rval < (float) 0.0) rval = (float) 0.0;
            if(gval < (float) 0.0) gval = (float) 0.0;
            if(bval < (float) 0.0) bval = (float) 0.0;
            if(rval > (float) 1.0) rval = (float) 1.0;
            if(gval > (float) 1.0) gval = (float) 1.0;
            if(bval > (float) 1.0) bval = (float) 1.0;

          } else {
            rval = (float) (1.0);
            gval = (float) (1.0);
            bval = (float) (1.0);
          }
        } else {
          rval = (float) (1.0);
          gval = (float) (1.0);
          bval = (float) (1.0);
        }
        
        offScreenImageGraphics.setColor(new Color(rval,gval,bval));
        offScreenImageGraphics.fillRect(xval, yval, dx, dy);
      }
    }

    // draw the boundaries
    offScreenImageGraphics.setColor(foreGroundColor);
    offScreenImageGraphics.drawLine(xoffset,yoffset,
            mapCanvasWidth-xoffset,yoffset);
    offScreenImageGraphics.drawLine(mapCanvasWidth-xoffset,yoffset,
            mapCanvasWidth-xoffset,mapCanvasHeight-yoffset);
    offScreenImageGraphics.drawLine(xoffset,mapCanvasHeight-yoffset,
            mapCanvasWidth-xoffset,mapCanvasHeight-yoffset);
    offScreenImageGraphics.drawLine(xoffset,yoffset,
            xoffset,mapCanvasHeight-yoffset);

    // check to see if the grid is toggled on
    if (gridOn) {
      offScreenImageGraphics.setColor(foreGroundColor);

      // draw the horizontal grid lines
      j = 1;
      yval = yoffset;
      while (j<dispnmapy) {
        yval = yval + dy;
        offScreenImageGraphics.drawLine(xoffset,yval,
                mapCanvasWidth-xoffset,yval);
        j++;
      }
  
      // draw the horizontal grid lines
      i = 1;
      xval = xoffset;
      while (i<dispnmapx) {
        xval = xval + dx;
        offScreenImageGraphics.drawLine(xval,yoffset,
                xval,mapCanvasHeight-yoffset);
        i++;
      }

    }

    // check to see if a zoom/choice box is being dragged on the image
    if(grabSome) {
    }

    // reset the color for the context to the original
    offScreenImageGraphics.setColor(color);

    // get rid of the graphics context for now
    offScreenImageGraphics.dispose();

    mapCanvas.redraw(0,0,offScreenImage);
    mapShowing = true;
    subregionShowing = false;

  }

  public void clearMap() {
    int xval, yval;
    int i, j;
    int j2, ind;
    float rval = 0, gval = 0, bval = 0;

    // set the canvas size
    setMapCanvasSize(dispnmapx, dispnmapy);

    // get the graphics context for the off screen image
    offScreenImage = null;
    offScreenImage = createImage(mapCanvasWidth,mapCanvasHeight);
    offScreenImageGraphics = offScreenImage.getGraphics();

    // draw the boundaries
    offScreenImageGraphics.drawLine(xoffset,yoffset,
            mapCanvasWidth-xoffset,yoffset);
    offScreenImageGraphics.drawLine(mapCanvasWidth-xoffset,yoffset,
            mapCanvasWidth-xoffset,mapCanvasHeight-yoffset);
    offScreenImageGraphics.drawLine(xoffset,mapCanvasHeight-yoffset,
            mapCanvasWidth-xoffset,mapCanvasHeight-yoffset);
    offScreenImageGraphics.drawLine(xoffset,yoffset,
            xoffset,mapCanvasHeight-yoffset);

    // draw the horizontal grid lines
    j = 1;
    yval = yoffset;
    while (j<dispnmapy) {
      yval = yval + dy;
      offScreenImageGraphics.drawLine(xoffset,yval,
              mapCanvasWidth-xoffset,yval);
      j++;
    }

    // draw the horizontal grid lines
    i = 1;
    xval = xoffset;
    while (i<dispnmapx) {
      xval = xval + dx;
      offScreenImageGraphics.drawLine(xval,yoffset,
              xval,mapCanvasHeight-yoffset);
      i++;
    }

    // set the colors of the blocks in the map
    for (j=1; j<=dispnmapy; j++) {
      j2 = ( j - 1 ) * dispnmapx;
      // get the cell boundary pixels in the second coordinate
      yval = ( j - 1 ) * dy + 1 + yoffset; 
      for (i=1; i<=dispnmapx; i++) {
        ind = i + j2 - 1;
        // get the cell boundary pixels in the first coordinate
        xval = ( i - 1 ) * dx + 1 + xoffset;

        offScreenImageGraphics.setColor(new Color(rval,gval,bval));
        offScreenImageGraphics.fillRect(xval, yval, dx-1, dy-1);
      }
    }

    // get rid of the graphics context for now
    offScreenImageGraphics.dispose();

    mapCanvas.redraw(0,0,offScreenImage);

    // print message
    printMessage("Map Cleared");
    mapShowing = false;
    subregionShowing = false;

  }

  public void redrawMap() {
    displayMapValues = mapValues;
    dispnmapx = nmapx;
    dispnmapy = nmapy;
    dispxlength = xlength;
    dispylength = ylength;
    showMap();
  }

  public void redrawSubregionMap() {
    displayMapValues = subregionValues;
    dispnmapx = subnmapx;
    dispnmapy = subnmapy;
    dispxlength = subxlength;
    dispylength = subylength;
    showMap();
  }

  public void toggleGridOn() {

    // check to make sure that the grid is big enough to draw grid lines
    if (dx < 3 || dy < 3 ) {
      printMessage("Block Sizes Too Small to Include a Grid");
      return;
    }

    if (gridOn) {
      gridOn = false;
      mapEditorGridOnOffButton.setLabel("Grid On");
      showMap();
    } else {
      gridOn = true;
      mapEditorGridOnOffButton.setLabel("Grid Off");
      showMap();
    }

  }

  public void togglePixelVsBlock() {

    if (blockMap) {
      blockMap = false;
      mapEditorPixelVsBlockButton.setLabel("Show Block Map");
      gridOn = false;
      mapEditorGridOnOffButton.setLabel("Grid On");
      dx = 1;
      dy = 1;
      redrawMap();
    } else {
      blockMap = true;
      mapEditorPixelVsBlockButton.setLabel("Show Pixel Map");
      dx = dxBlock;
      dy = dyBlock;
      redrawMap();
    }

  }

  public void nextSlice() {
    sliceIndex = sliceIndex + 1;
    if(sliceIndex > nmapz) sliceIndex = 1;
    showMap();
  }

  public void nextComponent() {
    componentIndex = componentIndex + 1;
    if(componentIndex > 8) componentIndex = 0;
    showMap();
  }

  public void previousSlice() {
    sliceIndex = sliceIndex + 1;
    if(sliceIndex < 1) sliceIndex = nmapz;
    showMap();
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *  methods for maniplating values in the map                              *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  public void setTempMapValues() {
    if(mapValues == null) return;
    tempMapValues = new Double [ntot][9];
    for(int i=0; i<ntot; i++) {
      for(int l=0; l<9; l++) {
        tempMapValues[i][l] = mapValues[i][l];
      }
    }
  }

  public void getTempMapValues() {
    if(tempMapValues == null) return;
    for(int i=0; i<ntot; i++) {
      for(int l=0; l<9; l++) {
        mapValues[i][l] = tempMapValues[i][l];
      }
    }
  }

  public void expOrLogOfValues () {
    if(takeExp) {
      expOfValues();
      takeExp = false;
      expOrLogOfMapButton.setLabel("Log of Current Map");
    } else {
      logOfValues();
      expOrLogOfMapButton.setLabel("Exp of Current Map");
      takeExp = true;
    }
  }

  public void expOfValues () {
    int i,l;

    // check to see if the map exists or not
    if(!mapLoaded) return;

    // compute the exp of the data
    for (i=0; i<ntot; i++) {
      for (l=0; l<9; l++) {
        tempMapValues[i][l]
              = new Double(Math.exp(mapValues[i][l].doubleValue()));
      }
    }
    // check for infinite values
    for (i=0; i<ntot; i++) {
      for (l=0; l<9; l++) {
       if(Double.toString(tempMapValues[i][l].doubleValue()).equals("Infinity")) {
          printMessage("Transformation of Map Contains Infinite Values");
          return;
        }
      }
    }
    getTempMapValues();
    redrawMap();
    // set the saved value to false since there is a new map
    mapModified = true;
    newMap = true;
  }

  public void logOfValues () {
    int i;

    // check to see if the map exists or not
    if(!mapLoaded) return;

    // first check for negative values
    for (i=0; i<ntot; i++) {
      for(int l=0; l<9; l++) {
        if(mapValues[i][l].doubleValue() <= 0.0) {
          printMessage("Map Contains Negative Values");
          return;
        }
      }
    }
    // compute the log of the data
    for (i=0; i<ntot; i++) {
      for(int l=0; l<9; l++) {
        mapValues[i][l] = new Double(Math.log(mapValues[i][l].doubleValue()));
      }
    }
    redrawMap();
    // set the saved value to false since there is a new map
    mapModified = true;
    newMap = true;
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *  methods for setting default map types                                  *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  public void uniformRandomValues () {
    int i;
    double scaling;
    componentIndex = 0;
    if(maximumValue != minimumValue) {
      scaling = maximumValue - minimumValue;
    } else {
      scaling = 1.0;
    }
    for (i=0; i<ntot; i++) {
      mapValues[i][0] = new Double(scaling*Math.random());
      mapValues[i][1] = new Double(0.0);
      mapValues[i][2] = new Double(0.0);
      mapValues[i][3] = new Double(0.0);
      mapValues[i][4] = mapValues[i][0];
      mapValues[i][5] = new Double(0.0);
      mapValues[i][6] = new Double(0.0);
      mapValues[i][7] = new Double(0.0);
      mapValues[i][8] = mapValues[i][0];
    }
    redrawMap();
    // set the saved value to false since there is a new map
    mapModified = true;
    newMap = true;
    mapLoaded = true;
  }

  public void normalRandomValues () {
    int i;
    double a0 = 2.30753;
    double a1 = 0.27061;
    double b1 = 0.99229;
    double b2 = 0.04481;
    double tval;
    double tinv;
    double earg;

    double smean = 0.5;
    double stddev = 0.20;

    componentIndex = 0;

    for (i=0; i<ntot; i++) {
      // compute a uniform value to use in the computation of the sample from
      // the normal distribution
      tval = new Double(Math.random()).doubleValue();
      if(tval > 0.5) {
        tval = 1.0 - tval;
        tinv = Math.sqrt(Math.log(1.0/(tval*tval)));
        earg = tinv - ( a0 + a1 * tinv ) /
                ( 1.0 + b1 * tinv + b2 * tinv * tinv );
        mapValues[i][0] = new Double(smean - stddev * earg);
      } else {
        tinv = Math.sqrt( Math.log(1.0/( tval * tval )));
        earg = tinv - ( a0 + a1 * tinv ) /
                ( 1.0 + b1 * tinv + b2 * tinv * tinv );
        mapValues[i][0] = new Double(smean + stddev * earg);
      }
      mapValues[i][1] = new Double(0.0);
      mapValues[i][2] = new Double(0.0);
      mapValues[i][3] = new Double(0.0);
      mapValues[i][4] = mapValues[i][0];
      mapValues[i][5] = new Double(0.0);
      mapValues[i][6] = new Double(0.0);
      mapValues[i][7] = new Double(0.0);
      mapValues[i][8] = mapValues[i][0];
    }
    redrawMap();
    // set the saved value to false since there is a new map
    mapModified = true;
    newMap = true;
    mapLoaded = true;
  }

  public void rectangularFractureValues() {
    double scaling;
    componentIndex = 0;
    if(maximumValue != minimumValue) {
      scaling = maximumValue - minimumValue;
      fractureValue = maximumValue;
      matrixValue = minimumValue;
    } else {
      scaling = 1.0;
    }
    // get the fracture information
    SetFractureValuesDialog sfvd
        = new SetFractureValuesDialog(containingFrame,
			 fractureValue, matrixValue, fractureWidth);
    sfvd.show();

    if(!sfvd.getState()) {
      printMessage("Fracture Values not Set");
      return;
    }

    // get the values
    fractureValue = sfvd.getFractureValue();
    matrixValue = sfvd.getMatrixValue();
    fractureWidth = sfvd.getFractureWidth();
    // set the background value to be the matrix values
    for (int i=0; i<ntot; i++) {
      mapValues[i][0] = new Double(matrixValue);
      mapValues[i][1] = new Double(0.0);
      mapValues[i][2] = new Double(0.0);
      mapValues[i][3] = new Double(0.0);
      mapValues[i][4] = mapValues[i][0];
      mapValues[i][5] = new Double(0.0);
      mapValues[i][6] = new Double(0.0);
      mapValues[i][7] = new Double(0.0);
      mapValues[i][8] = mapValues[i][0];
    }
    // now set the boundaries where the fractures live
    int fw = fractureWidth;
    for(int k=0; k<nmapz; k++) {
      int k2 = k * nmapx * nmapy;
      for(int j=0; j<nmapy; j++) {
        int j2 = j * nmapx + k2;
        for(int i=0; i<nmapx; i++) {
	  int ind = i + j2;
	  boolean test = false;
          if(nmapz>2 && (k<fw || k>=nmapz-fw)) test = true;
          if(nmapy>2 && (j<fw || j>=nmapy-fw)) test = true;
          if(nmapx>2 && (i<fw || i>=nmapx-fw)) test = true;
	  if(test) {
            mapValues[ind][0] = new Double(fractureValue);
            mapValues[ind][4] = mapValues[ind][0];
            mapValues[ind][8] = mapValues[ind][0];
          }
        }
      }
    }

    // show the map
    redrawMap();
    // set the saved value to false since there is a new map
    mapModified = true;
    newMap = true;
    mapLoaded = true;

  }

  public void setData(Double [][] data, int nx, int ny, int nz) {
    // set the integer block numbers
    nmapx = nx;
    nmapy = ny;
    nmapz = nz;
    // set the length of the data and build the new map array
    ntot = data.length;
    mapValues = new Double [ntot][9];
    // load the data into the map array
    for(int i=0; i<ntot; i++) {
      mapValues[i][0] = data[i][0];
      mapValues[i][1] = data[i][1];
      mapValues[i][2] = data[i][2];
      mapValues[i][3] = data[i][3];
      mapValues[i][4] = data[i][4];
      mapValues[i][5] = data[i][5];
      mapValues[i][6] = data[i][6];
      mapValues[i][7] = data[i][7];
      mapValues[i][8] = data[i][8];
    }
    setTempMapValues();
    defaultBlockSize();
    redrawMap();
    // set the saved value to false since there is a new map
    mapModified = true;
    newMap = true;
    mapLoaded = true;
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *  method for computing eigenvalues of a tensor; i.e., onoc the map values*
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */
  public double [] eignvals(double [] vals) {
    double xlam0, xlam1, xavg;
    double xlsq, xlcu, pval, dval;
    double error = 1.0;
    double [] eigs = new double [3];

    // set the tensor values
    double txx =  vals[0];
    double txy =  vals[1];
    double txz =  vals[2];
    double tyx =  vals[3];
    double tyy =  vals[4];
    double tyz =  vals[5];
    double tzx =  vals[6];
    double tzy =  vals[7];
    double tzz =  vals[8];

    // check for the diagonal case
    int ichk = 0;
    if(txy != 0.0) ichk = 1;
    if(txz != 0.0) ichk = 1;
    if(tyx != 0.0) ichk = 1;
    if(tyz != 0.0) ichk = 1;
    if(tzx != 0.0) ichk = 1;
    if(tzy != 0.0) ichk = 1;
    if(ichk == 0) {
      eigs[0] = txx;
      eigs[1] = tyy;
      eigs[2] = tzz;
      return eigs;
    }

    // check for a two dimensional case
    // x-y case
    ichk = 0;
    if(txz != 0.0) ichk = 1;
    if(tyz != 0.0) ichk = 1;
    if(tzx != 0.0) ichk = 1;
    if(tzy != 0.0) ichk = 1;
    if(ichk == 0) {
      double b = - ( txx + tyy );
      double c = txx * tyy - txy * tyx;
      double discr = Math.pow(b,2) - 4.0 * c;
      if(discr > 0.0) {
        eigs[0] = ( -b + Math.sqrt(discr) ) / 2.0;
        eigs[1] = ( -b - Math.sqrt(discr) ) / 2.0;
        eigs[2] = tzz;
      }
      return eigs;
    }
    // x-z case
    ichk = 0;
    if(txy != 0.0) ichk = 1;
    if(tyx != 0.0) ichk = 1;
    if(tyz != 0.0) ichk = 1;
    if(tzy != 0.0) ichk = 1;
    if(ichk == 0) {
      double b = - ( txx + tzz );
      double c = txx * tzz - txz * tzx;
      double discr = Math.pow(b,2) - 4.0 * c;
      if(discr > 0.0) {
        eigs[0] = ( -b + Math.sqrt(discr) ) / 2.0;
        eigs[1] = tyy;
        eigs[2] = ( -b - Math.sqrt(discr) ) / 2.0;
      }
      return eigs;
    }
    // y-z case
    ichk = 0;
    if(txy != 0.0) ichk = 1;
    if(txz != 0.0) ichk = 1;
    if(tyx != 0.0) ichk = 1;
    if(tzx != 0.0) ichk = 1;
    if(ichk == 0) {
      double b = - ( tyy + tzz );
      double c = tyy * tzz - tyz * tzy;
      double discr = Math.pow(b,2) - 4.0 * c;
      if(discr > 0.0) {
        eigs[0] = txx;
        eigs[1] = ( -b + Math.sqrt(discr) ) / 2.0;
        eigs[2] = ( -b - Math.sqrt(discr) ) / 2.0;
      }
      return eigs;
    }

    // do a simple iteration on the cubic polynomial
    int iter = 0;

    // set the coefficients of the cubic polynomial
    double a0 = txx * tyy * tzz + txy * tyz * tzx + txz * tyx * tzy
                - txz * tyy * tzx - tyz * tzy * txx - txy * tyx * tzz;
    double a1 = - txx * tyy - txx * tzz * tyy - tzz - txz * tzx
                - tyz * tzy - txy * tyx;
    double a2 = - txx - tyy - tzz;
    double a3 = 1.0;
    double b0 = a1;
    double b1 = 2.0 * a2;
    double b2 = 3.0 * a3;
    xlam0 = txx;
    xlam1 = txx;
    xavg = ( txx + tyy + tzz ) / 3.0;

    while(error > 0.0001 && iter < 100) {
      xlsq = xlam0 * xlam0;
      xlcu = xlam0 * xlsq;
      pval = a0 + a1 * xlam0 + a2 * xlsq + a3 * xlcu;
      dval = b0 + b1 * xlam0 + b2 * xlsq;
      if(dval == 0.0) {
        if(debug) System.out.println("Zero in Newton iteration in eignvals()");
        eigs[0] = 0.0;
        eigs[1] = 0.0;
        eigs[2] = 0.0;
        return eigs;
      }
      xlam1 = xlam0 - pval / dval;
      error = Math.abs( (xlam1 - xlam0) / xavg );
      iter++;
    }
    if(iter >= 100)
       if(debug)
           System.out.println("Number of Newton iterations in eignvals()>99");

    // now figure out the other two
    double c2 = a3;
    double c1 = a2 + c2 * xlam1;
    double c0 = a1 + c1 * xlam1;
    double discr = c1 * c1 + 4.0 * c0 * c2;

    if(discr<0.0) {
      if(debug) System.out.println("Discriminant is <0 in eignvals()");
      if(debug) System.out.println("Discriminant = " + discr);
    }

    eigs[0] = xlam0;
    double d0 = 2.0 * discr;
    discr = Math.sqrt(discr);
    eigs[1] = ( - c1 + discr ) / d0;
    eigs[2] = ( - c1 - discr ) / d0;

    return eigs;

  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *  methods for getting values from the gui                                *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  public void toggleSetMapValues() {
    if(grabSome) grabSubregion();
    if(doSetValues) {
      doSetValues = false;
      killSetValuesFrame();
      mapEditorSetValueButton.setLabel("Set Values in the Map");
    } else {
      if(!mapShowing) redrawMap();
      // popup the map values frame to let the user set values
      popupSetValuesFrame();
      mapEditorSetValueButton.setLabel("Quit Setting Values");
      doSetValues = true;
    }

  }

  public double getMinimumValue() {
    Double d;
    try {d = Double.valueOf(minimumValueTextField.getText());
         minimumValue = d.doubleValue();
    } catch(NumberFormatException e) {
      if(debug)
           System.out.println("--- Number Format Exception Encountered ---");
      if(debug)
           System.out.println("---  Make sure the minimum is a number  ---");
    }
    return minimumValue;
  }
  
  public double getMaximumValue() {
    Double d;
    try {d = Double.valueOf(maximumValueTextField.getText());
         maximumValue = d.doubleValue();
    } catch(NumberFormatException e) {
      if(debug)
           System.out.println("--- Number Format Exception Encountered ---");
      if(debug)
           System.out.println("---  Make sure the maximum is a number  ---");
    }
    return maximumValue;
  }
  
  public void getCurrentValue() {
    double diff;

    // get the values
    setValue = getdoubleSetValues();

    // loop over the components
    for(int i=0; i<9; i++) {
      diff = 0.00001 * Math.abs(maximumValue-minimumValue);
      // the value must be checked against the minimum and maximum values
      if(setValue[i]-minimumValue < diff) {
        minimumValue = setValue[i];
        this.setMinimumValue(setValue[i]);
        showMap();
      } 
      if(setValue[i]-maximumValue > diff) {
        maximumValue = setValue[i];
        this.setMaximumValue(setValue[i]);
        showMap();
      } 
    }

  }
  
  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *  methods for setting values for the object                              *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */
  
  public void setMapValues(Double [][] d, int nx, int ny, int nz) {
    // test the values to make sure they are consistent
    if(d == null) return;
    int n = d.length;
    if(n != nx*ny*nz) return;
    n = d[0].length;
    if(n != 9) return;
    // at this point things check out
    nmapx = nx;
    nmapy = ny;
    nmapz = nz;
    buildMapArray();
    // copy the data into the mapValues array
    for(int i=0; i<ntot; i++) {
      for(int l=0; l<9; l++) {
        double dval = d[i][l].doubleValue();
	mapValues[i][l] = new Double(dval);
      }
    }
    setTempMapValues();
    defaultBlockSize();
    displayMapValues = mapValues;
    dispnmapx = nmapx;
    dispnmapy = nmapy;
    dispxlength = xlength;
    dispylength = ylength;
    showMap();
  }

  public void setMapValue() {
    int indx; 
    int indy; 
    int index; 

    // test to see if the toggle is set on
    if(!doSetValues) return;
    // get the current location in the map
    xLocation = mapCanvasX - xoffset;
    if(xLocation < 0) xLocation = 0;
    if(xLocation > xlength) xLocation = xlength;
    // this line flips the graph vertically with respect to the 
    // y direction
    yLocation = ylength + yoffset - mapCanvasY;
    if(yLocation < 0) yLocation = 0;
    if(yLocation > ylength) yLocation = ylength;
    // compute the appropriate indices
    indx = (int) xLocation / dx + 1;
    indy = (int) yLocation / dy + 1;
    index = indx + ( indy - 1 ) * nmapx - 1; 
    // check the index to make sure that it is in range
    if(index < 0 || index >= ntot) return;
    for(int i=0; i<9; i++) mapValues[index][i] = new Double(setValue[i]);

    xLocationTextField.setText(Integer.toString(indx));
    yLocationTextField.setText(Integer.toString(indy));

  }
  
  public void setMinimumValue(double d) {
    String s;
    minimumValue = d;
    s = new Float((float)d).toString();
    if(s.length() > 11) {
      minimumValueTextField.setText(s.substring(0,10));
    } else {
      minimumValueTextField.setText(s);
    }
  }
  
  public void setMaximumValue(double d) {
    String s;
    maximumValue = d;
    s = new Float((float)d).toString();
    if(s.length() > 11) {
      maximumValueTextField.setText(s.substring(0,11));
    } else {
      maximumValueTextField.setText(s);
    }
  }
  
  public void setScale(double dmin, double dmax) {
    String s;
    minimumValue = dmin;
    s = new Float((float)dmin).toString();
    if(s.length() > 11) {
      minimumValueTextField.setText(s.substring(0,10));
    } else {
      minimumValueTextField.setText(s);
    }
    maximumValue = dmax;
    s = new Float((float)dmax).toString();
    if(s.length() > 11) {
      maximumValueTextField.setText(s.substring(0,11));
    } else {
      maximumValueTextField.setText(s);
    }
    rescale = false;
  }

  public boolean isRescaled() {
    return rescale;
  }

  public void setRescaled(boolean b) {
    rescale = b;
  }
  
  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *  method for computing the minimum and maximum values in the current map *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  public void computeMinAndMaxValues() {
    double [] tens = new double [9];
    double [] eigs = new double [3];
    int i;
    // get the minimum and maximum values of the data
    // set the value to be the value in the first block in the map
    tens[0] =  mapValues[0][0].doubleValue();
    tens[1] =  mapValues[0][1].doubleValue();
    tens[2] =  mapValues[0][2].doubleValue();
    tens[3] =  mapValues[0][3].doubleValue();
    tens[4] =  mapValues[0][4].doubleValue();
    tens[5] =  mapValues[0][5].doubleValue();
    tens[6] =  mapValues[0][6].doubleValue();
    tens[7] =  mapValues[0][7].doubleValue();
    tens[8] =  mapValues[0][8].doubleValue();
    eigs = eignvals(tens);
    minimumValue = eigs[0];
    maximumValue = eigs[0];
    if(minimumValue > eigs[1]) minimumValue = eigs[1];
    if(maximumValue < eigs[1]) maximumValue = eigs[1];
    if(minimumValue > eigs[2]) minimumValue = eigs[2];
    if(maximumValue < eigs[2]) maximumValue = eigs[2];
    for(int l=0; l<9; l++) {
      if(minimumValue > tens[l]) minimumValue = tens[l];
      if(maximumValue < tens[l]) maximumValue = tens[l];
    }
    // now loop over the rest of the blocks
    for (i=1; i<ntot; i++) {
      tens[0] =  mapValues[i][0].doubleValue();
      tens[1] =  mapValues[i][1].doubleValue();
      tens[2] =  mapValues[i][2].doubleValue();
      tens[3] =  mapValues[i][3].doubleValue();
      tens[4] =  mapValues[i][4].doubleValue();
      tens[5] =  mapValues[i][5].doubleValue();
      tens[6] =  mapValues[i][6].doubleValue();
      tens[7] =  mapValues[i][7].doubleValue();
      tens[8] =  mapValues[i][8].doubleValue();
      eigs = eignvals(tens);
      tens[0] =  mapValues[i][0].doubleValue();
      if(minimumValue > eigs[0]) minimumValue = eigs[0];
      if(maximumValue < eigs[0]) maximumValue = eigs[0];
      if(minimumValue > eigs[1]) minimumValue = eigs[1];
      if(maximumValue < eigs[1]) maximumValue = eigs[1];
      if(minimumValue > eigs[2]) minimumValue = eigs[2];
      if(maximumValue < eigs[2]) maximumValue = eigs[2];
      for(int l=0; l<9; l++) {
        if(minimumValue > tens[l]) minimumValue = tens[l];
        if(maximumValue < tens[l]) maximumValue = tens[l];
      }
    }
    // finally, look at the values in the set values boxes
    if(doSetValues) {
      tens = getdoubleSetValues();
      eigs = eignvals(tens);
      if(minimumValue > eigs[0]) minimumValue = eigs[0];
      if(maximumValue < eigs[0]) maximumValue = eigs[0];
      if(minimumValue > eigs[1]) minimumValue = eigs[1];
      if(maximumValue < eigs[1]) maximumValue = eigs[1];
      if(minimumValue > eigs[2]) minimumValue = eigs[2];
      if(maximumValue < eigs[2]) maximumValue = eigs[2];
    }

    // set the values in the gui
    setMinimumValue(minimumValue);
    setMaximumValue(maximumValue);

    // compute the average value for some uses
    averageValue = 0.5 * ( maximumValue + minimumValue );

  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   * methods for getting the values assigned to cells in the map             *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  public Dimension getPreferredSize() {
    return new Dimension(WIDTH, HEIGHT);
  }

  public Dimension getMinimumSize() {
    return new Dimension(minimumWidth,minimumHeight);
  }

  public Dimension getMaximumSize() {
    return new Dimension(WIDTH, HEIGHT);
  }

  public int getNumberOfBlocksX() {
    return nmapx;
  }

  public int getNumberOfBlocksY() {
    return nmapy;
  }

  public int getNumberOfBlocksZ() {
    return nmapz;
  }

  public int getTotalNumberOfBlocks() {
    return ntot;
  }

  public Double [][] getMapValues() {
    return mapValues;
  }

  public int getNumberOfSubregionBlocksX() {
    return subnmapx;
  }

  public int getNumberOfSubregionBlocksY() {
    return subnmapy;
  }

  public int getNumberOfSubregionBlocksZ() {
    return subnmapz;
  }

  public int getTotalNumberOfSubregionBlocks() {
    return subntot;
  }

  public Double [][] getSubregionValues() {
    return subregionValues;
  }

  public float getfloatValue(int x, int y, int l) {
    int i,j,ind;
    // first, locate the block
    i = (int) (dx * ( x / xlength ));
    j = (int) (dy * ( y / ylength ));
    ind = i + ( j - 1 ) * nmapx - 1;
    // get the value
    return mapValues[ind][l].floatValue();
  }

  public double getdoubleValue(int x, int y, int l) {
    int i,j,ind;
    // first, locate the block
    i = (int) (dx * ( x / xlength ));
    j = (int) (dy * ( y / ylength ));
    ind = i + ( j - 1 ) * nmapx - 1;
    // get the value
    return mapValues[ind][l].doubleValue();
  }

  public Double getDoubleValue(int x, int y, int l) {
    int i,j,ind;
    // first, locate the block
    i = (int) (dx * ( x / xlength ));
    j = (int) (dy * ( y / ylength ));
    ind = i + ( j - 1 ) * nmapx - 1;
    // get the value
    return mapValues[ind][l];
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *  methods for building the components of the gui                         *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  public void buildMapCanvas() {

    // instantiate the canvas
    mapCanvas = new ImageCanvas(mapCanvasWidth,mapCanvasHeight);
    mapCanvas.addMouseListener(this);
    mapCanvas.addMouseMotionListener(this);

    // set the map canvas size
    setMapCanvasSize(nmapx, nmapy);

    // build a scroll panel for display of the canvas/drawing area
    mapCanvasPanel = new ScrollPane(mapCanvasPanel.SCROLLBARS_ALWAYS);
    mapCanvasPanel.setBackground(Color.white);

    // add the canvas to the scrolling region
    mapCanvasPanel.add(mapCanvas);

    mapCanvasPanel.setSize(mapCanvasPanelWidth,mapCanvasPanelHeight);

  }

  public void buildMapDomainControl() {
    mapDomainPanel = new Panel();
    mapDomainPanel.setLayout(new BorderLayout());

      Panel pl = new Panel();
      pl.setLayout(new GridLayout(1,0));
      pl.setBackground(labelColor);
      mapDomainPanelLabel = new Label("Domain Information", Label.LEFT);
      pl.add(mapDomainPanelLabel);
      mapDomainPanel.add(pl, "North");

      // put together the map value text fields 
      Panel p0 = new Panel();
      p0.setLayout(new GridLayout(0,1));

      // set up the minimum value field
      Panel p1 = new Panel();
      p1.setLayout(new GridLayout(1,0));
        p1.add(new Label("Min Value"));
        minimumValueTextField = new TextField("",5);
          minimumValueTextField.setBackground(Color.white);
          minimumValueTextField.setForeground(Color.black);
        p1.add(minimumValueTextField);
        p1.add(new Label("Max Value"));
        maximumValueTextField = new TextField("",5);
          maximumValueTextField.setBackground(Color.white);
          maximumValueTextField.setForeground(Color.black);
        p1.add(maximumValueTextField);
        p1.add(new Label(""));
        p1.add(new Label(""));

      // set up the index locator textfields for the map
      Panel p2 = new Panel();
      p2.setLayout(new GridLayout(1,0));
        p2.add(new Label("X Index"));
        xBlockTextField = new TextField("",5);
          xBlockTextField.setBackground(Color.white);
          xBlockTextField.setForeground(Color.black);
        p2.add(xBlockTextField);
        p2.add(new Label("Y Index"));
        yBlockTextField = new TextField("",5);
          yBlockTextField.setBackground(Color.white);
          yBlockTextField.setForeground(Color.black);
        p2.add(yBlockTextField);
        p2.add(new Label("Slice in 3D"));
        slice3DTextField = new TextField(new Integer(sliceIndex).toString(), 5);
        p2.add(slice3DTextField);

      // set up the index locator textfields for the map
      Panel p3 = new Panel();
      p3.setLayout(new GridLayout(1,0));
        p3.add(new Label("X Location"));
        xLocationTextField = new TextField("",5);
          xLocationTextField.setBackground(Color.white);
          xLocationTextField.setForeground(Color.black);
        p3.add(xLocationTextField);
        p3.add(new Label("Y Location"));
        yLocationTextField = new TextField("",5);
          yLocationTextField.setBackground(Color.white);
          yLocationTextField.setForeground(Color.black);
        p3.add(yLocationTextField);
        p3.add(new Label("Z Location"));
        zLocationTextField = new TextField("0.0", 5);
          zLocationTextField.setBackground(Color.white);
          zLocationTextField.setForeground(Color.black);
        p3.add(zLocationTextField);

    // add the pieces
    p0.add(p1);
    p0.add(p2);
    p0.add(p3);

    mapDomainPanel.add(p0, "Center");

  }

  public void buildMapEditorControls() {

    // set up the panel
    mapEditorControls = new Panel();;
    mapEditorControls.setLayout(new GridLayout(0,1));

    // build the label
    mapControlLabel = new Label("Editor Controls",Label.LEFT);
    mapControlLabel.setBackground(labelColor);
    mapEditorControls.add(mapControlLabel);

    // build the load button
    mapEditorLoadNewButton = new Button("Load New Map");
    mapEditorLoadNewButton.addActionListener(this);

    // build the reload button
    mapEditorReloadButton = new Button("Reload the Current Map");
    mapEditorReloadButton.addActionListener(this);

    // build the save button
    mapEditorSaveButton = new Button("Save Current Map");
    mapEditorSaveButton.addActionListener(this);

    // build the save button
    mapEditorSaveSameButton = new Button("Overwrite Map File");
    mapEditorSaveSameButton.addActionListener(this);

    // build the clear button
    mapEditorClearButton = new Button("Clear Canvas");
    mapEditorClearButton.addActionListener(this);

    // build the redraw button
    mapEditorRedrawButton = new Button("Redraw Current Map");
    mapEditorRedrawButton.addActionListener(this);

    // build the build a map button
    mapEditorBuildAMapButton = new Button("Build A Map");
    mapEditorBuildAMapButton.addActionListener(this);

    // build the grid on/off toggle button
    mapEditorGridOnOffButton = new Button("Grid Off");
    mapEditorGridOnOffButton.addActionListener(this);

    // build the zoom up button
    mapEditorGridRefine = new Button("Refine Computational Grid");
    mapEditorGridRefine.addActionListener(this);

    // build the zoom up button
    mapEditorGridUnRefine = new Button("UnRefine Computational Grid");
    mapEditorGridUnRefine.addActionListener(this);

    // build the zoom up button
    mapEditorZoomUpButton = new Button("Zoom Up (2x)");
    mapEditorZoomUpButton.addActionListener(this);

    // build the zoom down button
    mapEditorZoomDownButton = new Button("Zoom Down (x2)");
    mapEditorZoomDownButton.addActionListener(this);

    // build the default size button
    mapEditorDefaultSizeButton = new Button("Default Size");
    mapEditorDefaultSizeButton.addActionListener(this);

    // build the pixel map vs block map button
    mapEditorPixelVsBlockButton = new Button("Show Pixel Map");
    mapEditorPixelVsBlockButton.addActionListener(this);

    // build the next slice button
    mapEditorNextSliceButton = new Button("Next Slice");
    mapEditorNextSliceButton.addActionListener(this);

    // build the next component button
    mapEditorNextComponentButton = new Button("Next Component");
    mapEditorNextComponentButton.addActionListener(this);

    // build the grab some button
    mapEditorGrabSomeButton = new Button("Grab A Subregion");
    mapEditorGrabSomeButton.addActionListener(this);

    // build a button to display the current subregion
    mapEditorShowSubregionButton = new Button("Display Current Subregion");
    mapEditorShowSubregionButton.addActionListener(this);

    // build the set value button
    mapEditorSetValueButton = new Button("Set Values in the Map");
    mapEditorSetValueButton.addActionListener(this);

    // build the exit button
    mapEditorExitButton = new Button("Exit");
    mapEditorExitButton.addActionListener(this);

    // add the buttons
    if(controlMask[0]) mapEditorControls.add(mapEditorLoadNewButton);
    if(controlMask[1]) mapEditorControls.add(mapEditorReloadButton);
    if(controlMask[2]) mapEditorControls.add(mapEditorSaveButton);
    if(controlMask[3]) mapEditorControls.add(mapEditorSaveSameButton);
    if(controlMask[4]) mapEditorControls.add(mapEditorClearButton);
    if(controlMask[5]) mapEditorControls.add(mapEditorRedrawButton);
    if(controlMask[6]) mapEditorControls.add(mapEditorBuildAMapButton);
    if(controlMask[7]) mapEditorControls.add(mapEditorGridOnOffButton);
    if(controlMask[8]) mapEditorControls.add(mapEditorGridRefine);
    if(controlMask[9]) mapEditorControls.add(mapEditorGridUnRefine);
    if(controlMask[10]) mapEditorControls.add(mapEditorZoomUpButton);
    if(controlMask[11]) mapEditorControls.add(mapEditorZoomDownButton);
    if(controlMask[12]) mapEditorControls.add(mapEditorDefaultSizeButton);
    if(controlMask[13]) mapEditorControls.add(mapEditorPixelVsBlockButton);
    if(controlMask[14]) mapEditorControls.add(mapEditorNextSliceButton);
    if(controlMask[15]) mapEditorControls.add(mapEditorNextComponentButton);
    if(controlMask[16]) mapEditorControls.add(mapEditorGrabSomeButton);
    if(controlMask[17]) mapEditorControls.add(mapEditorShowSubregionButton);
    if(controlMask[18]) mapEditorControls.add(mapEditorSetValueButton);
    if(controlMask[19]) mapEditorControls.add(mapEditorExitButton);

  }

  public void buildMapDefaultGeneration() {

    // set up the panel and label
    defaultGenerationPanel = new Panel();
    defaultGenerationPanel.setLayout(new BorderLayout());

    Panel pl = new Panel();
    pl.setLayout(new GridLayout(1,0));
    pl.setBackground(labelColor);
      defaultGenerationPanelLabel
            = new Label("Data Manipulation/Generation Functions", Label.LEFT);
    pl.add(defaultGenerationPanelLabel);

      // build a simple panel of buttons
      Panel p1 = new Panel();
      p1.setLayout(new GridLayout(0,3));
  
        // uniform distribution choice
        uniformValuesGeneration = new Button("Uniform Random Map");
        uniformValuesGeneration.addActionListener(this);

        // normal distribution choice
        normalValuesGeneration = new Button("Normal Random Map");
        normalValuesGeneration.addActionListener(this);

        // exp/log of the current distribution choice
        expOrLogOfMapButton = new Button("Exp of Current Map");
        expOrLogOfMapButton.addActionListener(this);

        // rectangular fracture choice
        rectangularFracture = new Button("Rectangular Fracture Map");
        rectangularFracture.addActionListener(this);

      // add the buttons to the panel
      p1.add(uniformValuesGeneration);
      p1.add(normalValuesGeneration);
      p1.add(expOrLogOfMapButton);
      p1.add(rectangularFracture);

    defaultGenerationPanel.add(pl, "North");
    defaultGenerationPanel.add(p1, "Center");

  }

  public void enableExitButton() {
    mapEditorExitButton.setEnabled(true);
  }

  public void disableExitButton() {
    mapEditorExitButton.setEnabled(false);
  }

  public void buildMessagePanel() {

    // build the panel and text area
    messagePanel = new Panel();
    messagePanel.setLayout(new BorderLayout());
      messageLabel = new Label("Messages:", Label.LEFT);
      messageLabel.setBackground(labelColor);
      messageTextArea = new TextArea(4,60);
        messageTextArea.setBackground(Color.white);
        messageTextArea.setForeground(Color.black);
      messageTextArea.setEditable(false);
    messagePanel.add("North", messageLabel);
    messagePanel.add("Center", messageTextArea);

  }

  public void popupCurrentValuesFrame() {
    double sum;
    currentValuesFrame = new Frame("Current Tensor Values Frame");

    currentValuesPanel = new Panel();
    currentValuesPanel.setLayout(new BorderLayout());

    currentValuesPanelLabel = new Label("Current Tensor Values", Label.CENTER);
    currentValuesPanel.add("North", currentValuesPanelLabel);

      cvaluePanel = new Panel();
      cvaluePanel.setLayout(new GridLayout(3,6));
      cvaluePanel.setBackground(panelColor);

        cvaluePanel.add(clxx);
        cvaluePanel.add(cxxTextField);
          sum = 0.0;
          for(int i=0; i<ntot; i++) {
            sum = sum + mapValues[i][0].doubleValue();
          }
          sum = sum / ((double) ntot);
          cxxTextField.setText(digitFilter(sum, 3));
        cvaluePanel.add(clxy);
        cvaluePanel.add(cxyTextField);
          sum = 0.0;
          for(int i=0; i<ntot; i++) {
            sum = sum + mapValues[i][1].doubleValue();
          }
          sum = sum / ((double) ntot);
          cxyTextField.setText(digitFilter(sum, 3));
        cvaluePanel.add(clxz);
        cvaluePanel.add(cxzTextField);
          sum = 0.0;
          for(int i=0; i<ntot; i++) {
            sum = sum + mapValues[i][2].doubleValue();
          }
          sum = sum / ((double) ntot);
          cxzTextField.setText(digitFilter(sum, 3));
        cvaluePanel.add(clyx);
        cvaluePanel.add(cyxTextField);
          sum = 0.0;
          for(int i=0; i<ntot; i++) {
            sum = sum + mapValues[i][3].doubleValue();
          }
          sum = sum / ((double) ntot);
          cyxTextField.setText(digitFilter(sum, 3));
        cvaluePanel.add(clyy);
        cvaluePanel.add(cyyTextField);
          sum = 0.0;
          for(int i=0; i<ntot; i++) {
            sum = sum + mapValues[i][4].doubleValue();
          }
          sum = sum / ((double) ntot);
          cyyTextField.setText(digitFilter(sum, 3));
        cvaluePanel.add(clyz);
        cvaluePanel.add(cyzTextField);
          sum = 0.0;
          for(int i=0; i<ntot; i++) {
            sum = sum + mapValues[i][5].doubleValue();
          }
          sum = sum / ((double) ntot);
          cyzTextField.setText(digitFilter(sum, 3));
        cvaluePanel.add(clzx);
        cvaluePanel.add(czxTextField);
          sum = 0.0;
          for(int i=0; i<ntot; i++) {
            sum = sum + mapValues[i][6].doubleValue();
          }
          sum = sum / ((double) ntot);
          czxTextField.setText(digitFilter(sum, 3));
        cvaluePanel.add(clzy);
        cvaluePanel.add(czyTextField);
          sum = 0.0;
          for(int i=0; i<ntot; i++) {
            sum = sum + mapValues[i][7].doubleValue();
          }
          sum = sum / ((double) ntot);
          czyTextField.setText(digitFilter(sum, 3));
          cvaluePanel.add(clzz);
          cvaluePanel.add(czzTextField);
          sum = 0.0;
          for(int i=0; i<ntot; i++) {
            sum = sum + mapValues[i][8].doubleValue();
          }
          sum = sum / ((double) ntot);
          czzTextField.setText(digitFilter(sum, 3));

      // add the panel for the tensor values
      currentValuesPanel.add("Center",cvaluePanel);

    // add the panel to the box
    currentValuesFrame.add(currentValuesPanel);

    // set the thing to a fixed size
    int xl = this.getLocation().x;
    currentValuesFrame.setSize(width, height);
    currentValuesFrame.setLocation(xl-width, 0);
    currentValuesFrame.show();

  }

  public void removeCurrentValuesFrame() {
    currentValuesFrame.setVisible(false);
    currentValuesFrame = null;
  }

  public void popupSetValuesFrame() {
    double sum;
    if(setValuesFrame == null) {
      setValuesFrame = new Frame("Set Tensor Values Frame");

        setValuesPanel = new Panel();
        setValuesPanel.setLayout(new BorderLayout());

          setValuesPanelLabel = new Label("Set Tensor Values", Label.CENTER);
          setValuesPanel.add("North", setValuesPanelLabel);

          valuePanel = new Panel();
          valuePanel.setLayout(new GridLayout(3,6));
          valuePanel.setBackground(panelColor);

          valuePanel.add(lxx);
            txxTextField.setBackground(Color.white);
            txxTextField.setForeground(Color.black);
          valuePanel.add(txxTextField);
            sum = 0.0;
            for(int i=0; i<ntot; i++) {
              sum = sum + mapValues[i][0].doubleValue();
            }
            sum = sum / ((double) ntot);
            txxTextField.setText(digitFilter(sum, 3));
          valuePanel.add(lxy);
            txyTextField.setBackground(Color.white);
            txyTextField.setForeground(Color.black);
          valuePanel.add(txyTextField);
            sum = 0.0;
            for(int i=0; i<ntot; i++) {
              sum = sum + mapValues[i][1].doubleValue();
            }
            sum = sum / ((double) ntot);
            txyTextField.setText(digitFilter(sum, 3));
          valuePanel.add(lxz);
            txzTextField.setBackground(Color.white);
            txzTextField.setForeground(Color.black);
          valuePanel.add(txzTextField);
            sum = 0.0;
            for(int i=0; i<ntot; i++) {
              sum = sum + mapValues[i][2].doubleValue();
            }
            sum = sum / ((double) ntot);
            txzTextField.setText(digitFilter(sum, 3));
          valuePanel.add(lyx);
            tyxTextField.setBackground(Color.white);
            tyxTextField.setForeground(Color.black);
          valuePanel.add(tyxTextField);
            sum = 0.0;
            for(int i=0; i<ntot; i++) {
              sum = sum + mapValues[i][3].doubleValue();
            }
            sum = sum / ((double) ntot);
            tyxTextField.setText(digitFilter(sum, 3));
          valuePanel.add(lyy);
            tyyTextField.setBackground(Color.white);
            tyyTextField.setForeground(Color.black);
          valuePanel.add(tyyTextField);
            sum = 0.0;
            for(int i=0; i<ntot; i++) {
              sum = sum + mapValues[i][4].doubleValue();
            }
            sum = sum / ((double) ntot);
            tyyTextField.setText(digitFilter(sum, 3));
          valuePanel.add(lyz);
            tyzTextField.setBackground(Color.white);
            tyzTextField.setForeground(Color.black);
          valuePanel.add(tyzTextField);
            sum = 0.0;
            for(int i=0; i<ntot; i++) {
              sum = sum + mapValues[i][5].doubleValue();
            }
            sum = sum / ((double) ntot);
            tyzTextField.setText(digitFilter(sum, 3));
          valuePanel.add(lzx);
            tzxTextField.setBackground(Color.white);
            tzxTextField.setForeground(Color.black);
          valuePanel.add(tzxTextField);
            sum = 0.0;
            for(int i=0; i<ntot; i++) {
              sum = sum + mapValues[i][6].doubleValue();
            }
            sum = sum / ((double) ntot);
            tzxTextField.setText(digitFilter(sum, 3));
          valuePanel.add(lzy);
            tzyTextField.setBackground(Color.white);
            tzyTextField.setForeground(Color.black);
          valuePanel.add(tzyTextField);
            sum = 0.0;
            for(int i=0; i<ntot; i++) {
              sum = sum + mapValues[i][7].doubleValue();
            }
            sum = sum / ((double) ntot);
            tzyTextField.setText(digitFilter(sum, 3));
          valuePanel.add(lzz);
            tzzTextField.setBackground(Color.white);
            tzzTextField.setForeground(Color.black);
          valuePanel.add(tzzTextField);
            sum = 0.0;
            for(int i=0; i<ntot; i++) {
              sum = sum + mapValues[i][8].doubleValue();
            }
            sum = sum / ((double) ntot);
            tzzTextField.setText(digitFilter(sum, 3));

        // add the panel for the tensor values
        setValuesPanel.add("Center",valuePanel);

      // add the panel to the box
      setValuesFrame.add(setValuesPanel);

      // set the thing to a fixed size
      setValuesFrame.setSize(width, height);
      int xl = this.getLocation().x;
      setValuesFrame.setLocation(xl-width, height);
      setValuesFrame.show();

    } else {
      setValuesFrame.setVisible(true);
    }

  }

  public void killSetValuesFrame() {
    setValuesFrame.setVisible(false);
    setValuesFrame = null;
  }

  public Double [] getDoubleSetValues() {
    Double [] values = new Double [9];
    try{values[0] = Double.valueOf(txxTextField.getText());
    } catch (NumberFormatException nfe) {
      values[0] = new Double(0.0);
    }
    try{values[1] = Double.valueOf(txyTextField.getText());
    } catch (NumberFormatException nfe) {
      values[1] = new Double(0.0);
    }
    try{values[2] = Double.valueOf(txzTextField.getText());
    } catch (NumberFormatException nfe) {
      values[2] = new Double(0.0);
    }
    try{values[3] = Double.valueOf(tyxTextField.getText());
    } catch (NumberFormatException nfe) {
      values[3] = new Double(0.0);
    }
    try{values[4] = Double.valueOf(tyyTextField.getText());
    } catch (NumberFormatException nfe) {
      values[4] = new Double(0.0);
    }
    try{values[5] = Double.valueOf(tyzTextField.getText());
    } catch (NumberFormatException nfe) {
      values[5] = new Double(0.0);
    }
    try{values[6] = Double.valueOf(tzxTextField.getText());
    } catch (NumberFormatException nfe) {
      values[6] = new Double(0.0);
    }
    try{values[7] = Double.valueOf(tzyTextField.getText());
    } catch (NumberFormatException nfe) {
      values[7] = new Double(0.0);
    }
    try{values[8] = Double.valueOf(tzzTextField.getText());
    } catch (NumberFormatException nfe) {
      values[8] = new Double(0.0);
    }
    return values;
  }

  public double [] getdoubleSetValues() {
    double [] values = new double [9];
    try{values[0] = Double.valueOf(txxTextField.getText()).doubleValue();
    } catch (NumberFormatException nfe) {
      values[0] = 0.0;
    }
    try{values[1] = Double.valueOf(txyTextField.getText()).doubleValue();
    } catch (NumberFormatException nfe) {
      values[1] = 0.0;
    }
    try{values[2] = Double.valueOf(txzTextField.getText()).doubleValue();
    } catch (NumberFormatException nfe) {
      values[2] = 0.0;
    }
    try{values[3] = Double.valueOf(tyxTextField.getText()).doubleValue();
    } catch (NumberFormatException nfe) {
      values[3] = 0.0;
    }
    try{values[4] = Double.valueOf(tyyTextField.getText()).doubleValue();
    } catch (NumberFormatException nfe) {
      values[4] = 0.0;
    }
    try{values[5] = Double.valueOf(tyzTextField.getText()).doubleValue();
    } catch (NumberFormatException nfe) {
      values[5] = 0.0;
    }
    try{values[6] = Double.valueOf(tzxTextField.getText()).doubleValue();
    } catch (NumberFormatException nfe) {
      values[6] = 0.0;
    }
    try{values[7] = Double.valueOf(tzyTextField.getText()).doubleValue();
    } catch (NumberFormatException nfe) {
      values[7] = 0.0;
    }
    try{values[8] = Double.valueOf(tzzTextField.getText()).doubleValue();
    } catch (NumberFormatException nfe) {
      values[8] = 0.0;
    }
    return values;
  }

  public void setValues(Double [] values) {
    txxTextField.setText(values[0].toString());
    txyTextField.setText(values[1].toString());
    txzTextField.setText(values[2].toString());
    tyxTextField.setText(values[3].toString());
    tyyTextField.setText(values[4].toString());
    tyzTextField.setText(values[5].toString());
    tzxTextField.setText(values[6].toString());
    tzyTextField.setText(values[7].toString());
    tzzTextField.setText(values[8].toString());
  }

  public void setValues(double [] values) {
    txxTextField.setText(digitFilter(values[0], 3));
    txyTextField.setText(digitFilter(values[1], 3));
    txzTextField.setText(digitFilter(values[2], 3));
    tyxTextField.setText(digitFilter(values[3], 3));
    tyyTextField.setText(digitFilter(values[4], 3));
    tyzTextField.setText(digitFilter(values[5], 3));
    tzxTextField.setText(digitFilter(values[6], 3));
    tzyTextField.setText(digitFilter(values[7], 3));
    tzzTextField.setText(digitFilter(values[8], 3));
  }

  public void setCurrentValues(double [] values) {
    cxxTextField.setText(digitFilter(values[0], 3));
    cxyTextField.setText(digitFilter(values[1], 3));
    cxzTextField.setText(digitFilter(values[2], 3));
    cyxTextField.setText(digitFilter(values[3], 3));
    cyyTextField.setText(digitFilter(values[4], 3));
    cyzTextField.setText(digitFilter(values[5], 3));
    czxTextField.setText(digitFilter(values[6], 3));
    czyTextField.setText(digitFilter(values[7], 3));
    czzTextField.setText(digitFilter(values[8], 3));
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *    mouse event methods                                                  *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  public void mousePressed(MouseEvent e) {
    Point tmpPoint = e.getPoint();
    if(tmpPoint != null) {
      mapCanvasX = tmpPoint.x;
      mapCanvasY = tmpPoint.y;
      getCurrentValue();
      setMapValue();
      // finally, show the new map
      showMap();
      // set the map modified flag to true
      mapModified = true;
    }
    if(grabSome) {
      if(!subregionStart) {
        subregionStart = true;
        subX1 = subX2 = tmpPoint.x;
        subY1 = subY2 = tmpPoint.y;
      } else {
        subregionStart = false;
        subX2 = tmpPoint.x;
        subY2 = tmpPoint.y;
        grabSubregion();
      }
    }
  }

  public void mouseReleased(MouseEvent e) {
  }

  public void mouseEntered(MouseEvent e) {
  }

  public void mouseExited(MouseEvent e) {
  }

  public void mouseClicked(MouseEvent e) {
  }

  public void mouseDragged(MouseEvent e) {
    Point tmpPoint = e.getPoint();
    if(tmpPoint != null) {
      mapCanvasX = tmpPoint.x;
      mapCanvasY = tmpPoint.y;
      getCurrentValue();
      setMapValue();
      // finally, show the new map
      showMap();
      // set the map modified flag to true
      mapModified = true;
    }
  }

  public void mouseMoved(MouseEvent e) {
    double [] values = new double[9];
    Point tmpPoint = e.getPoint();
    if(tmpPoint != null) {
      mapCanvasX = tmpPoint.x;
      mapCanvasY = tmpPoint.y;
      xLocationTextField.setText(new Integer(mapCanvasX).toString());
      yLocationTextField.setText(new Integer(mapCanvasY).toString());
      int indx = (int) mapCanvasX / dx + 1;
      int indy = nmapy - (int) mapCanvasY / dy;
      xBlockTextField.setText(new Integer(indx).toString());
      yBlockTextField.setText(new Integer(indy).toString());
      int index = indx + ( indy - 1 ) * nmapx - 1;
      if(index < 0) index = 0;
      if(index >= ntot) index = ntot - 1;
      values[0] = mapValues[index][0].doubleValue();
      values[1] = mapValues[index][1].doubleValue();
      values[2] = mapValues[index][2].doubleValue();
      values[3] = mapValues[index][3].doubleValue();
      values[4] = mapValues[index][4].doubleValue();
      values[5] = mapValues[index][5].doubleValue();
      values[6] = mapValues[index][6].doubleValue();
      values[7] = mapValues[index][7].doubleValue();
      values[8] = mapValues[index][8].doubleValue();
      setCurrentValues(values);
    }
  }

  public void actionPerformed(ActionEvent e) {
    Object source = e.getSource();

    if(source.equals(mapEditorLoadNewButton)) loadNewMap();
    if(source.equals(mapEditorReloadButton)) reloadMap();
    if(source.equals(mapEditorSaveButton)) saveMap();
    if(source.equals(mapEditorSaveSameButton)) overwriteMap();
    if(source.equals(mapEditorClearButton)) clearMap();
    if(source.equals(mapEditorRedrawButton)) redrawMap();
    if(source.equals(mapEditorBuildAMapButton)) buildAMap();
    if(source.equals(mapEditorGridOnOffButton)) toggleGridOn();
    if(source.equals(mapEditorGridRefine)) refineGrid();
    if(source.equals(mapEditorGridUnRefine)) unrefineGrid();
    if(source.equals(mapEditorZoomUpButton)) zoomUp();
    if(source.equals(mapEditorZoomDownButton)) zoomDown();
    if(source.equals(mapEditorPixelVsBlockButton)) togglePixelVsBlock();
    if(source.equals(mapEditorDefaultSizeButton)) defaultSize();
    if(source.equals(mapEditorNextSliceButton)) nextSlice();
    if(source.equals(mapEditorNextComponentButton)) nextComponent();
    if(source.equals(mapEditorGrabSomeButton)) grabSubregion();
    if(source.equals(mapEditorShowSubregionButton)) showSubregion();
    if(source.equals(mapEditorSetValueButton)) toggleSetMapValues();
    if(source.equals(mapEditorExitButton)) System.exit(0);

    if(source.equals(uniformValuesGeneration)) uniformRandomValues();
    if(source.equals(normalValuesGeneration)) normalRandomValues();
    if(source.equals(expOrLogOfMapButton)) expOrLogOfValues();
    if(source.equals(rectangularFracture)) rectangularFractureValues();

  }

  private void plot1d() {
    printMessage("This Ain't Working Yet");
    // first get the beginning and ending point of the 1-d data
    //Plot1D plot1d = new Plot1D(data);
  }

  private void fft1d() {
    printMessage("This Ain't Working Yet");
    // first get the beginning and ending point of the 1-d data
    // next, set up the line of data and get the data - 256 points
    // build an fft object and do the fft calculation
    //FFT1D fft = new FFT1D(data);
    //fft.compute();
    // now display the results
  }

  private void wavelet1d() {
    printMessage("This Ain't Working Yet");
  }

  public static String digitFilter(double value, int n) {
    String integerPart = "0";
    String dot = ".";
    String fractionPart = "0";
    String exponent = "";
    String t = null;

    String s = Double.toString(value);
    StringTokenizer st = new StringTokenizer(s, " .eE");
    if(st.hasMoreTokens()) integerPart = st.nextToken();
    if(st.hasMoreTokens()) t = st.nextToken();
    if(t != null) {
      fractionPart = t;
      if(t.length() > n) fractionPart = t.substring(0,n);
    }
    if(st.hasMoreTokens()) exponent = st.nextToken();
    if(exponent.length() > 0) exponent = "E" + exponent;

    return integerPart + dot + fractionPart + exponent;
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   * main() - called when the application is run.                            *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  public static void main(String args[]) {
    Frame f = new Frame("Map Editor Frame");
    MapEditor mapEditor = new MapEditor(f);
    f.add("Center", mapEditor);
    f.show();
    mapEditor.popupCurrentValuesFrame();
    //f.setSize(600,800);
  }

  // the basic panel variables for the pieces
  private Frame containingFrame;
  private Panel mainPanel;
    private Panel displayPanel;
    private Panel manipulationPanel;
      private Panel domainPanel;
      private Panel otherPanel;

  // the map canvas area
  private ScrollPane mapCanvasPanel;
  private int mapCanvasPanelWidth = 400;
  private int mapCanvasPanelHeight = 400;
  private ImageCanvas mapCanvas;
  private int mapCanvasWidth = 400;
  private int mapCanvasHeight = 400;
  private int mapCanvasX;
  private int mapCanvasY;

  // the controls panel variables
  private boolean [] controlMask;
  private int ncontrols = 20;
  private Panel mapEditorControls;
    private Label mapControlLabel;
    private Button mapEditorLoadNewButton;
    private Button mapEditorReloadButton;
    private Choice saveChoice;
    private Button mapEditorSaveButton;
    private Button mapEditorSaveSameButton;
    private Button mapEditorClearButton;
    private Button mapEditorGridOnOffButton;
    private Button mapEditorRedrawButton;
    private Button mapEditorBuildAMapButton;
    private Button mapEditorGridRefine;
    private Button mapEditorGridUnRefine;
    private Button mapEditorZoomUpButton;
    private Button mapEditorZoomDownButton;
    private Button mapEditorDefaultSizeButton;
    private Button mapEditorPixelVsBlockButton;
    private Button mapEditorNextSliceButton;
    private Button mapEditorNextComponentButton;
      private boolean blockMap = true;
    private Button mapEditorGrabSomeButton;
    private Button mapEditorShowSubregionButton;
    private Button mapEditorSetValueButton;
    private Button mapEditorExitButton;

  // the domain panel
  private Panel mapDomainPanel;
    private Label mapDomainPanelLabel;
    private TextField minimumValueTextField;
    private TextField maximumValueTextField;
    private boolean rescale = true;
    private TextField xLocationTextField;
    private TextField yLocationTextField;
    private TextField zLocationTextField;
    private TextField xBlockTextField;
    private TextField yBlockTextField;
    private TextField slice3DTextField;

  private double minimumValue = 0.0;
  private double maximumValue = 1.0;
  private double averageValue;
  private int xLocation = 0;
  private int yLocation = 0;
  private int zLocation = 0;
  private int xBlock = 0;
  private int yBlock = 0;
  private int sliceIndex = 1;
  private int componentIndex = 0;

  // variables for the default cases in the map
  private Panel defaultGenerationPanel;
    private Label defaultGenerationPanelLabel;
    private Button uniformValuesGeneration;
    private Button normalValuesGeneration;
    private Button expOrLogOfMapButton;
    private Button rectangularFracture;
    private boolean takeExp = true;

  // variables for the message panel
  private Panel messagePanel;
    private Label messageLabel;
    private TextArea messageTextArea;

  // variables for the off screen image being created
  private Image offScreenImage;
  private Graphics offScreenImageGraphics;

  // map grid variables
  private int xoffset = 5;
  private int yoffset = 5;
  private int nmapx = 20;
  private int nmapy = 20;
  private int nmapz = 1;
  private int ntot;
  private int dx, dxBlock;
  private int dy, dyBlock;
  private int xlength;
  private int ylength;

  // map values in a linear array
  private Double [][] mapValues;
  private Double [][] tempMapValues;
  private int nextMapInputIndex = 0;
  public boolean gridOn = true;
  public boolean mapShowing = false;
  public boolean gridRefined = false;
  public int gridRefinementLevel = 0;

  // variables for grabbing a chunk of the region
  private Rectangle subregion;
  private Double [][] subregionValues = null;
  private int subX1;
  private int subY1;
  private int subX2;
  private int subY2;
  private int subindx1;
  private int subindy1;
  private int subindx2;
  private int subindy2;
  private boolean subregionStart = false;
  private int subnmapx;
  private int subnmapy;
  private int subnmapz = 1;
  private int subntot;
  private int subdx, subdxBlock;
  private int subdy, subdyBlock;
  private int subxlength;
  private int subylength;
  private boolean grabSome = false;
  public boolean subregionShowing = false;

  // the folowing is the map that will be drawn
  private Double [][] displayMapValues = null;
  private int dispnmapx;
  private int dispnmapy;
  private int dispxlength;
  private int dispylength;

  private double fractureValue = 100;
  private double matrixValue = 1.;
  private int fractureWidth = 1;

  // color map
  private Vector colorMap;

  // file stuff
  private File directory;
  private String directoryString = "./";
  private File loadFile;
  private String loadFileString;
  private File saveFile;
  private String saveFileString;
  private String message;
  private boolean mapModified = false;
  private boolean newMap = false;
  private boolean mapLoaded = false;

  private String fileSeparator;

  // the runtime object for the application
  Runtime runtime;
  Toolkit toolkit;

  // screen size variables
  private Dimension screensize;
  private int screenresolution;

  // color variables
  private Color backGroundColor = Color.white;
  private Color foreGroundColor = Color.black;
  private Color panelColor = new Color(196,196,132);
  private Color labelColor = new Color(110,132,255);

  // frame size
  private static int WIDTH = 650;
  private static int HEIGHT = 800;
  private static int minimumWidth = 500;
  private static int minimumHeight = 800;
  private int widthInput = -1;
  private int heightInput = -1;
  private int xPosition = 0;
  private int yPosition = 0;

  private double [] setValue = new double [9];
  private boolean doSetValues = false;
  private int width = 450;
  private int height = 150;

  private Frame setValuesFrame;
    private Label setValuesPanelLabel;
    private Panel setValuesPanel;
      private Panel valuePanel;
        private Label lxx = new Label("txx", Label.CENTER);
        private Label lxy = new Label("txy", Label.CENTER);
        private Label lxz = new Label("txz", Label.CENTER);
        private Label lyx = new Label("tyx", Label.CENTER);
        private Label lyy = new Label("tyy", Label.CENTER);
        private Label lyz = new Label("tyz", Label.CENTER);
        private Label lzx = new Label("tzx", Label.CENTER);
        private Label lzy = new Label("tzy", Label.CENTER);
        private Label lzz = new Label("tzz", Label.CENTER);
        private TextField txxTextField = new TextField(10);
        private TextField txyTextField = new TextField(10);
        private TextField txzTextField = new TextField(10);
        private TextField tyxTextField = new TextField(10);
        private TextField tyyTextField = new TextField(10);
        private TextField tyzTextField = new TextField(10);
        private TextField tzxTextField = new TextField(10);
        private TextField tzyTextField = new TextField(10);
        private TextField tzzTextField = new TextField(10);

  private Frame currentValuesFrame;
    private Label currentValuesPanelLabel;
    private Panel currentValuesPanel;
      private Panel cvaluePanel;
        private Label clxx = new Label("txx", Label.CENTER);
        private Label clxy = new Label("txy", Label.CENTER);
        private Label clxz = new Label("txz", Label.CENTER);
        private Label clyx = new Label("tyx", Label.CENTER);
        private Label clyy = new Label("tyy", Label.CENTER);
        private Label clyz = new Label("tyz", Label.CENTER);
        private Label clzx = new Label("tzx", Label.CENTER);
        private Label clzy = new Label("tzy", Label.CENTER);
        private Label clzz = new Label("tzz", Label.CENTER);
        private TextField cxxTextField = new TextField(10);
        private TextField cxyTextField = new TextField(10);
        private TextField cxzTextField = new TextField(10);
        private TextField cyxTextField = new TextField(10);
        private TextField cyyTextField = new TextField(10);
        private TextField cyzTextField = new TextField(10);
        private TextField czxTextField = new TextField(10);
        private TextField czyTextField = new TextField(10);
        private TextField czzTextField = new TextField(10);

  private boolean debug = false;

}
