/*
 * SimulatorGUI class: MapObject
 *
 */

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.Arrays;
import java.util.Observable;
import java.util.StringTokenizer;
import java.util.Vector;

/*
 * this object holds a map and any other associated maps
 */
public class MapObject extends Object {

  // null object for putting stuff into the object
  public MapObject()
       throws BadArraySizeException {
    this(null, null,
         0, 0, 0,
         null, null,
         null, null, 
         null, null, 
         null, null, null,
         null); 
  }

  // name only object with no data
  public MapObject(String name)
       throws BadArraySizeException {
    this(name, null,
         0, 0, 0,
         null, null,
         null, null, 
         null, null, 
         null, null, null,
         null); 
  }

  // name only object with no data
  public MapObject(String name, MapObject pmo)
       throws BadArraySizeException {
    this(name, null,
         0, 0, 0,
         null, null,
         null, null, 
         null, null, 
         null, null, null,
         pmo); 
  }

  // name and data only object with no data
  public MapObject(String name, Double [][] d, int nx, int ny, int nz)
       throws BadArraySizeException {
    this(name, d,
         nx, ny, nz,
         null, null,
         null, null, 
         null, null, 
         null, null, null,
         null); 
  }

  // name and data only object with no data
  public MapObject(String name, Double [][] d,
                   int nx, int ny, int nz, MapObject pmo)
       throws BadArraySizeException {
    this(name, d,
         nx, ny, nz,
         null, null,
         null, null, 
         null, null, 
         null, null, null,
         pmo); 
  }

  // full constructor for the class
  public MapObject(String name, Double [][] d,
                   int nx, int ny, int nz,
                   Double xl, Double xr,
                   Double yl, Double yr,
                   Double zl, Double zr,
                   Double [] xg, Double [] yg, Double [] zg,
                   MapObject pmo)
       throws BadArraySizeException {
    mapName = name;
    data = d;
    nmapx = nx;
    nmapy = ny;
    nmapz = nz;
    nxny = nmapx * nmapy;
    nnmap = nxny * nmapz;
    // check to see if this is a null object for later use by the instantiation
    // source
    if(data != null) {
      // number of components or layers in the map
      nComp = data[0].length;
      if(nComp <= 0) throw
         new BadArraySizeException("Number of Layers Nonpositive in MapObject");
      if(data.length != nnmap) throw
           new BadArraySizeException("Array the wrong size in MapObject");
      initialize(xl, xr, yl, yr, zl, zr, xg, yg, zg);
    } else {
      nComp = 1;
      initialize(new Double(0.0), new Double(1.0), new Double(0.0),
                 new Double(1.0), new Double(0.0), new Double(1.0),
                 null, null, null);
    }

    // at this point the map has not been modified and does not need to be
    // saved
    modified = false;
    saved = true;
    parentMapObject = pmo;
    if(parentMapObject != null)
               parentMapObjectName = parentMapObject.getMapName();
  }

  // intialize the object
  public void initialize(Double xl, Double xr,
                    Double yl, Double yr,
                    Double zl, Double zr,
                    Double [] xg, Double [] yg, Double [] zg)
       throws BadArraySizeException {

    // set up the grid
    if(xl != null) {
      xleft = xl;
    } else {
      xleft = new Double(0.0);
    }
    if(xr != null) {
      xright= xr;
    } else {
      xright = new Double(1.0);
    }
    if(yl != null) {
      yleft = yl;
    } else {
      yleft = new Double(0.0);
    }
    if(yr != null) {
      yright= yr;
    } else {
      yright = new Double(1.0);
    }
    if(zl != null) {
      zleft = zl;
    } else {
      zleft = new Double(0.0);
    }
    if(zr != null) {
      zright= zr;
    } else {
      zright = new Double(1.0);
    }

    // make sure the end points are ok
    if(xleft.doubleValue() > xright.doubleValue()) {
      Double temp = xleft;
      xleft = xright;
      xright = temp;
    }
    if(yleft.doubleValue() > yright.doubleValue()) {
      Double temp = yleft;
      yleft = yright;
      yright = temp;
    }
    if(zleft.doubleValue() > zright.doubleValue()) {
      Double temp = zleft;
      zleft = zright;
      zright = temp;
    }

    // set or compute the grid lines in the domain
    if(xg != null) {
      xgrid = xg;
      nmapx = xgrid.length;
    } else {
      xgrid = computeGrid(xleft, xright, nmapx);
    }
    Arrays.sort(xgrid);
    if(yg != null) {
      ygrid = yg;
      nmapy = ygrid.length;
    } else {
      ygrid = computeGrid(yleft, yright, nmapy);
    }
    Arrays.sort(ygrid);
    if(zg != null) {
      zgrid = zg;
      nmapz = zgrid.length;
    } else {
      zgrid = computeGrid(zleft, zright, nmapz);
    }
    Arrays.sort(zgrid);

    // test the end points against the grid - the only time this matters is if
    // the grid lines are specified
    if(xleft.doubleValue() != xgrid[0].doubleValue()) xgrid[0] = xleft;
    if(xright.doubleValue() != xgrid[nmapx].doubleValue())
            xgrid[nmapx] = xright;
    if(yleft.doubleValue() != ygrid[0].doubleValue()) ygrid[0] = yleft;
    if(yright.doubleValue() != ygrid[nmapy].doubleValue())
            ygrid[nmapy] = yright;
    if(zleft.doubleValue() != zgrid[0].doubleValue()) zgrid[0] = zleft;
    if(zright.doubleValue() != zgrid[nmapz].doubleValue())
            zgrid[nmapz] = zright;

  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *    method for computing a grid if this has not been given               *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  public Double [] computeGrid(Double xl, Double xr, int nx) {
    Double [] grid = null;
    grid = new Double [nx+1];
    double dx = (xr.doubleValue() - xl.doubleValue()) / ((double) nx);
    grid[0] = xl;
    for(int i=1; i<nx; i++) {
      grid[i] = new Double(grid[i-1].doubleValue() + dx);
    }
    grid[nx] = xr;
    return grid;
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *    method for getting the internal variables                            *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */
 
  public String getMapName() {
    return mapName;
  }
 
  public Double [][] getData() {
    return data;
  }
 
  public int getNumberOfComponents() {
    return nComp;
  }
 
  public int getNumberOfXBlocks() {
    return nmapx;
  }
 
  public int getNumberOfYBlocks() {
    return nmapy;
  }
 
  public int getNumberOfZBlocks() {
    return nmapz;
  }
 
  public int getTotalBlocks() {
    return nnmap;
  }

  public Double getXLeft() {
    return xleft;
  }

  public double getxLeft() {
    return xleft.doubleValue();
  }

  public Double getXRight() {
    return xright;
  }

  public double getxRight() {
    return xright.doubleValue();
  }

  public Double getYLeft() {
    return yleft;
  }

  public double getyLeft() {
    return yleft.doubleValue();
  }

  public Double getYRight() {
    return yright;
  }

  public double getyRight() {
    return yright.doubleValue();
  }

  public Double getZLeft() {
    return zleft;
  }

  public double getzLeft() {
    return zleft.doubleValue();
  }

  public Double getZRight() {
    return zright;
  }

  public double getzRight() {
    return zright.doubleValue();
  }

  public Double [] getXGrid() {
    return xgrid;
  }

  public Double [] getYGrid() {
    return ygrid;
  }

  public Double [] getZGrid() {
    return zgrid;
  }

  public double [] getxGrid() {
    double [] xg = new double [nmapx+1];
    for(int i=0; i<=nmapx; i++) xg[i] = xgrid[i].doubleValue();
    return xg;
  }

  public double [] getyGrid() {
    double [] yg = new double [nmapy+1];
    for(int j=0; j<=nmapy; j++) yg[j] = ygrid[j].doubleValue();
    return yg;
  }

  public double [] getzGrid() {
    double [] zg = new double [nmapz+1];
    for(int k=0; k<=nmapz; k++) zg[k] = zgrid[k].doubleValue();
    return zg;
  }

  public void setSaved(boolean b) {
    if(b) {
      saved = true;
    } else {
      saved = false;
    }
  }

  public boolean isSaved() {
    if(saved) {
      if(associatedMapObjects != null) {
        int n = associatedMapObjects.size();
        if(n <= 0) return true;
        for(int i=0; i<n; i++) {
          MapObject mo = (MapObject) associatedMapObjects.get(i);
          if(!mo.isSaved()) return false;
        }
      }
      return true;
    } else {
      return false;
    }
  }

  public boolean isModified() {
    if(modified) {
      if(associatedMapObjects != null) {
        int n = associatedMapObjects.size();
        if(n <= 0) return true;
        for(int i=0; i<n; i++) {
          MapObject mo = (MapObject) associatedMapObjects.get(i);
          if(!mo.isModified()) return false;
        }
      }
      return true;
    } else {
      return false;
    }
  }

  public MapObject getParentMap() {
    return parentMapObject;
  }

  public String getParentMapName() {
    return parentMapObjectName;
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *    method for setting the internal variables                            *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  public void setMapName(String s) {
    if(s == null) return;
    mapName = s;
    modified = true;
    saved = false;
  }
 
  public void setData(Double [][] d) {
    if(d == null) return;
    // check the dimensions
    int n = d.length;
    if(n != d.length) return;
    n = d[0].length;
    if(n != d[0].length) return;
    data = d;
    modified = true;
    saved = false;
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *    methods for adding associated maps                                   *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  // the following methods qre NOT recrusive and are intended to address a
  // specific level in the map structure

  public void addAssociatedMap(MapObject mo) {
    if(associatedMapObjects == null) associatedMapObjects = new Vector();
    associatedMapObjects.addElement(mo);
    modified = true;
    saved = false;
  }

  public void addAssociatedMap(String name, Double [][] d,
                               int nx, int ny, int nz) {
    if(associatedMapObjects == null) associatedMapObjects = new Vector();
    try {
      MapObject mo = new MapObject(name, d, nx, ny, nz);
      associatedMapObjects.addElement(mo);
      modified = true;
      saved = false;
    } catch(BadArraySizeException bse) {
      System.out.println("Bad Array Size in Adding an Associated Map");
    }
  }

  public MapObject getAssociatedMap(String name) {
    if(name == null) return null;
    if(associatedMapObjects == null) return null;
    String s = name.trim();
    int n = associatedMapObjects.size();
    for(int i=0; i<n; i++) {
      MapObject mo = (MapObject) associatedMapObjects.get(i);
      if(s.equals(mo.getMapName().trim())) return mo;
    }
    return null;
  }

  public MapObject getAssociatedMap(int index) {
    if(associatedMapObjects == null) return null;
    int n = associatedMapObjects.size();
    if(index < 0 || index >= n) return null;
    MapObject mo = (MapObject) associatedMapObjects.get(index);
    return mo;
  }

  public Vector getAssociatedMaps() {
    return associatedMapObjects;
  }

  public boolean anyAssociatedMaps() {
    if(associatedMapObjects == null) {
      return false;
    } else {
      if(associatedMapObjects.size() <= 0) return false;
    }
    return true;
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *    methods for setting up map relationships and hierarchies for map     *
   *    storage qnd organization - some of the methods below are recursive   *
   *    while others only reference a given level in the structure of        *
   *    associated maps                                                      *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  // the following methods ARE NOT recursive

  public int getNumberOfAssociatedMapObjects() {
    if(associatedMapObjects == null) return 0;
    return associatedMapObjects.size();
  }

  public String [] getAssociatedMapNames() {
    String [] names = null;
    if(associatedMapObjects != null) {
      int n = associatedMapObjects.size();
      names = new String[n];
      for(int i=0; i<n; i++) {
        MapObject mo = (MapObject) associatedMapObjects.get(i);
        names[i] = mo.getMapName();
      }
    }
    return names;
  }

  // the following methods ARE recursive

  public int depth() {
    if(associatedMapObjects == null) {
      return 0;
    }
    int n = associatedMapObjects.size();
    if(n <= 0) return 0;
    int cnt = 1;
    boolean test = false;
    for(int i=0; i<n; i++) {
      MapObject mo = (MapObject) associatedMapObjects.get(i);
      if(mo.depth() > 0) test = true;
    }
    if(test) cnt++;
    return cnt;
  }

  public int getTotalNumberOfMapObjects() {
    int cnt = 1;
    cnt = cnt + getTotalNumberOfAssociatedMaps();
    return cnt;
  }

  public int getTotalNumberOfAssociatedMaps() {
    int cnt = 0;
    if(associatedMapObjects != null) {
      int n = associatedMapObjects.size();
      cnt = cnt + n;
      for(int i=0; i<n; i++) {
        MapObject mo = (MapObject) associatedMapObjects.get(i);
        cnt = cnt + mo.getTotalNumberOfAssociatedMaps();
      }
    }
    return cnt;
  }

  public String [] getMapObjectHierarchy() {
    Vector ov = getMapObjectHierarchy(".");
    String [] s = null;
    if(ov != null) {
      int n = ov.size();
      s = new String [n];
      for(int i=0; i<n; i++) s[i] = (String) ov.get(i);
    }
    return s;
  }

  public Vector getMapObjectHierarchy(String head) {
    Vector ov = new Vector();
    String nextHead = head + File.separator + getMapName();
    ov.addElement(nextHead);
    if(associatedMapObjects != null) {
      int n = associatedMapObjects.size();
      for(int i=0; i<n; i++) {
        MapObject mo = (MapObject) associatedMapObjects.get(i);
        Vector va = mo.getMapObjectHierarchy(nextHead);
	int m = va.size();
	for(int j=0; j<m; j++) ov.addElement((Object) va.get(j));
      }
    }
    return ov;
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *    methods for writing and loading a map structure to this object       *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  public boolean buildMapFile() {
    String s = getMapName();
    if(s == null) return false;
    File f = new File(s);
    if(!f.exists()) f.mkdirs();
    return true;
  }

  public boolean buildDirectoryStructure() {
    String [] s = getMapObjectHierarchy();
    if(s == null) return false;
    int n = s.length;
    for(int i=0; i<n; i++) {
      File f = new File(s[i]);
      if(!f.exists()) f.mkdirs();
    }
    return true;
  }

  public String [] loadMapHierarchy() {
    String [] sa = getMapObjectHierarchy();
    if(sa == null) return null;
    if(sa.length <= 0) return null;
    String s = getMapName();
    if(s == null) return null;
    File f = new File(s);
    if(!f.exists()) f.mkdirs();
    File fa = new File(f, "map_hierarchy.dat");
    try {
      FileWriter fw = new FileWriter(fa);
      try {
        int n = sa.length;
	for(int i=0; i<n; i++) {
          String line = sa[i];
          fw.write(line,0,line.length());
        }
      } catch (IOException ioea) {
        return null;
      }
      fw.close();
    } catch (IOException ioe) {
      return null;
    }
    return sa;
  }

  public boolean loadMap() {
    String s = getMapName() + File.separator + getMapName() + ".map";
    if(!loadMap(s)) return false;
    checkInputMap();
    modified = false;
    saved = true;
    return true;
  }

  public boolean loadMaps() {
    if(!loadMaps(".")) return false;
    checkInputMap();
    return true;
  }

  public boolean loadMaps(String head) {
    // set up the folder and file names for a particular map
    String nextHead = head + File.separator + getMapName();
    String filename = nextHead + File.separator + getMapName() + ".map";
    // load this map;
    if(!loadMap(filename)) return false;
    // search the folder for folders that possibly contain other maps - start
    // by getting a list of files in the folder
    File f = new File(nextHead);
    String [] files = f.list();
    // see if there are any files to look at
    if(files != null) {
      // loop over the files testing for directories
      int nf = files.length;
      for(int i=0; i<nf; i++) {
        File tf = new File(nextHead, files[i]);
	if(tf.isDirectory()) {
          // now the recursive loading part of things
          try {
            MapObject mo = new MapObject(files[i]);
            if(mo.loadMaps(nextHead)) addAssociatedMap(mo);
          } catch(BadArraySizeException bse) {
          }
        }
      }
    }
    modified = false;
    saved = true;
    return true;
  }

  public boolean reloadMaps() {
    reloadMaps(".");
    checkInputMap();
    return true;
  }

  public boolean reloadMaps(String head) {
    String nextHead = head + File.separator + getMapName();
    String filename = nextHead + File.separator + getMapName() + ".map";
    loadMap(filename);
    if(associatedMapObjects != null) {
      int n = associatedMapObjects.size();
      for(int i=0; i<n; i++) {
        MapObject mo = (MapObject) associatedMapObjects.get(i);
        mo.reloadMaps(nextHead);
      }
    }
    modified = false;
    saved = true;
    return true;
  }

  public void checkInputMap() {
    if(nCompRead < nComp) {
      for(int ll=nCompRead; ll<nComp; ll++) {
        for(int i=0; i<nnmap; i++) data[i][ll] = new Double(0.0);
      }
    }
  }

  public boolean loadMap(String loadString) {
    FileReader fr;
    LineNumberReader lnr;
    String line;
    int nx, ny, nz;

    // build a file reader
    File loadFile = new File(loadString);
    if(loadFile == null) return false;
    try{fr = new FileReader(loadFile);
      // set up the input stream
      try {lnr = new LineNumberReader(fr);
        // read in the dimensions
        // get the number of values in the first dimension
        line = lnr.readLine();
        if(line != null) {
          nx = getFirstInteger(line);
        } else {
          return false;
        }
        if(nx < 1) {
          return false;
        }
        // get the number of values in the second dimension
        line = lnr.readLine();
        if(line != null) {
          ny = getFirstInteger(line);
        } else {
          return false;
        }
        if(ny < 1) {
          return false;
        }
        // get the number of values in the third dimension
        line = lnr.readLine();
        if(line != null) {
          nz = getFirstInteger(line);
        } else {
          return false;
        }
        if(nz < 1) {
          return false;
        }
	// set the map object values
	nmapx = nx;
	nmapy = ny;
	nmapz = nz;
	nnmap = nmapx * nmapy * nmapz;
        if(nnmap == 0) return false;
	data = new Double[nnmap][9];
        // loop over the components
	nCompRead = 0;
	componentIndex = 0;
	line = "start it up";
        while(line != null && componentIndex < 9) {
          // set the load index location to 0
          nextMapInputIndex = 0;
          // loop through until enough numbers are retrieved for the present
          // component
          while(nextMapInputIndex<nnmap) {
            line = lnr.readLine();
            if(line != null) {
              getNumbers(line);
            } else {
              // in this case there is a problem with the data file
              if(nCompRead<1) return false;
              nComp = nCompRead;
              return true;
            }
          }
          nCompRead++;
          componentIndex++;
        }
        // finally, close the random access file
        lnr.close();
        } catch (IOException io) {
          return false;
        }
    } catch (FileNotFoundException fnfe) {
      return false;
    }

    try {
      initialize(xleft, xright, yleft, yright, zleft, zright, null, null, null);
    } catch(BadArraySizeException base) {
    }
    nComp = nCompRead;
    return true;

  }

  public boolean writeMapHierarchy() {
    if(!buildMapFile()) return false;
    String [] sa = getMapObjectHierarchy();
    if(sa == null) return false;
    if(sa.length <= 0) return false;
    String s = getMapName();
    if(s == null) return false;
    File f = new File(s);
    if(!f.exists()) f.mkdirs();
    File fa = new File(f, "map_hierarchy.dat");
    try {
      FileWriter fw = new FileWriter(fa);
      try {
        int n = sa.length;
	for(int i=0; i<n; i++) {
          String line = sa[i] + "\n";
          fw.write(line,0,line.length());
        }
      } catch (IOException ioea) {
        return false;
      }
      fw.close();
    } catch (IOException ioe) {
      return false;
    }
    return true;
  }

  public boolean writeMap() {
    if(!buildMapFile()) return false;
    String s = "."
             + File.separator
             + getMapName()
             + File.separator
             + getMapName()
             + ".map";
    if(!writeMap(s)) return false;
    modified = false;
    saved = true;
    return true;
  }

  public boolean writeMaps() {
    if(!buildDirectoryStructure()) return false;
    writeMaps(".");
    return true;
  }

  public boolean writeMaps(String head) {
    String nextHead = head + File.separator + getMapName();
    String filename = nextHead + File.separator + getMapName() + ".map";
    writeMap(filename);
    if(associatedMapObjects != null) {
      int n = associatedMapObjects.size();
      for(int i=0; i<n; i++) {
        MapObject mo = (MapObject) associatedMapObjects.get(i);
        mo.writeMaps(nextHead);
      }
    }
    modified = false;
    saved = true;
    return true;
  }

  // the following method needs a bunch of work - it is not at all functional
  // at this point in time.
  public boolean writeLevelMaps(int l1, int l2) {
    // get the correct ordering of the levels for the recursive invokation
    int l1tmp = l1;
    int l2tmp = l2;
    if(l1 > l2) {
      l1tmp = l2;
      l2tmp = l1;
    }
    // test the trivial case of writing only the zeroth level of the map
    if(l1tmp < 0 && l2tmp < 0) return false;
    if(l1tmp == 0) writeMap();
    if(l2tmp == 0) return true;
    // now drop into the rest of the structure if needed
    if(associatedMapObjects == null && (l1tmp > 0 || l2tmp > 0)) return false;
    // get the rest of the map
    int n = associatedMapObjects.size();
    if(n <= 0) return false;
    int cnt = 1;
    boolean test = false;
    // now for the recursive part
    for(int i=0; i<n; i++) {
      MapObject mo = (MapObject) associatedMapObjects.get(i);
      if(mo.depth() > 0) test = true;
    }
    if(test) cnt++;
    return true;
  }

  public boolean writeMap(String saveString) {
    FileWriter fw;
    String line, numberString;
    int index, nx, ny, nz;
    int power;
    double mantissa=0.0,value;
    int sign=1, exponent=1;
    boolean dataleft = false;

    // build a file writer
    File saveFile = new File(saveString);
    try{fw = new FileWriter(saveFile);

        // print out the number of blocks in each direction
        try{
            // first coordinate
            numberString = Integer.toString(nmapx);
            line = "  " + numberString +
                   "           *** number of grid block in first coordinate\n";
            fw.write(line,0,line.length());
            // second coordinate
            numberString = Integer.toString(nmapy);
            line = "  " + numberString +
                 "           *** number of grid block in second coordinate\n";
            fw.write(line,0,line.length());
            // third coordinate
            numberString = Integer.toString(nmapz);
            line = "  " + numberString +
                 "           *** number of grid block in third coordinate\n";
            fw.write(line,0,line.length());

            // loop over the components in the tensor
            for(int ll=0; ll<nComp; ll++) {

              index = 0;
              line = " ";
              while (index < nnmap) {
                if(data[index][ll] != null) {
                  value = Math.abs(data[index][ll].doubleValue());
                } else {
                  value = 0.0;
                }
                if(value == 0.0) {
                  line = line + "  0.000000E+00";
                } else {
                  power = 0;
                  if(value > 0.1) {
                    mantissa = value;
                    while (mantissa >= 1.0) {
                      power = power + 1;
                      mantissa = value / Math.pow(10.,(double) power);
                    }
                    numberString = new Double(mantissa).toString();
                    int cnt = numberString.length();
                    while (cnt < 10) {
                      numberString = numberString + "0";
                      cnt ++;
                    }
                    String mantissaString = numberString.substring(0,8);
                    String exponentString = new Integer(power).toString();
                    if(exponentString.length() == 0)
                          exponentString = "00";
                    if(exponentString.length() == 1)
                          exponentString = "0" + exponentString;
                    if(exponentString.length() > 2)
                          exponentString = "**";
                    if(data[index][ll].doubleValue() >= 0.0) {
                      line = line +"  "+ mantissaString + "E+" + exponentString;
                    } else {
                      line = line +" -"+ mantissaString + "E+" + exponentString;
                    }
                  } else {
                    mantissa = value;
                    while (mantissa <= 0.1) {
                      power = power + 1;
                      mantissa = value * Math.pow(10.,(double)power);
                    }
                    numberString = new Double(mantissa).toString();
                    int cnt = numberString.length();
                    while (cnt < 10) {
                      numberString = numberString + "0";
                      cnt ++;
                    }
                    String mantissaString = numberString.substring(0,8);
                    String exponentString = new Integer(power).toString();
                    if(exponentString.length() == 0)
                          exponentString = "00";
                    if(exponentString.length() == 1)
                          exponentString = "0" + exponentString;
                    if(exponentString.length() > 2)
                          exponentString = "**";
                    if(data[index][ll].doubleValue() >= 0.0) {
                      line = line +"  "+ mantissaString + "E-" + exponentString;
                    } else {
                      line = line +" -"+ mantissaString + "E-" + exponentString;
                    }
                  }
                }

                dataleft = true;
              
                index++;
                if(mod(index,5) == 0) {
                  line = line + "\n";
                  fw.write(line,0,line.length());
                  line = " ";
                  dataleft = false;
                }
              }

              // check for any left over data
              if(dataleft) {
                line = line + "\n";
                fw.write(line,0,line.length());
              }

            }

        } catch (IOException ioea) {
          return false;
        }

        // close the file writer
        fw.close();

    } catch (IOException ioe) {
      return false;
    }

    return true;

  }

  public int mod(int i1, int i2) {
    return i1 - i2 * ((int) i1 / i2);
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   * methods for processing input strings and strings for output to files    *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  public int getFirstInteger(String s) {
    // set up a string tokenizer
    StringTokenizer st = new StringTokenizer(s, " \t\r\n");
    // go through the tokens in the string
    while (st.hasMoreTokens()) {
      String token = st.nextToken();
      // try to build an integer
      try{Integer intval = Integer.valueOf(token);
          return intval.intValue();
      } catch (NumberFormatException nfe) {
        System.out.println("Number Format Exception");
      }
    }
    return -1;
  }

  public void getNumbers(String s) {
    double mantissa = 0.0;
    double value;
    int sign = 1; 
    int exponent = 1; 
    boolean problem;
    // set up a string tokenizer
    StringTokenizer st = new StringTokenizer(s, " \t\r\n");
    // go through the tokens in the string
    while (st.hasMoreTokens()) {
      problem = false;
      String number = st.nextToken();
      StringTokenizer nst = new StringTokenizer(number, " eE");
      int nTokens = nst.countTokens();

      // if the number of tokens above is more than one, then this must be an
      // exponential format if it is a number
      if(nTokens > 1) {

        // get the two pieces - these are the mantissa and the exponent
        String mantissaString = nst.nextToken();
        double numberSign = 1.0; 
        if(mantissaString.startsWith("-")) {
          numberSign = - 1.0; 
          int len = mantissaString.length();
          mantissaString = mantissaString.substring(1);
        }
        String signedExponentString = nst.nextToken();
        String signString = signedExponentString.substring(0,1);
        String exponentString = signedExponentString.substring(1,3);

        // try to build the mantissa
        try{Double mantissaDouble = Double.valueOf(mantissaString);
            mantissa = mantissaDouble.doubleValue();
        } catch (NumberFormatException nfe) {
          System.out.println("Number Format Exception");
          problem = true;
        }

        // get the sign
        if (signString.equalsIgnoreCase("-")) {
          sign = -1;
        } else {
          sign = 1;
        }

        // try to build the exponent
        try{Integer exponentInteger = Integer.valueOf(exponentString);
            exponent = exponentInteger.intValue();
        } catch (NumberFormatException nfe) {
          System.out.println("Number Format Exception");
          problem = true;
        }

        // get the number
        if(!problem) {
          value = numberSign
            * Math.exp(Math.log(mantissa) + sign*exponent*Math.log(10.0));
          data[nextMapInputIndex][componentIndex] = new Double(value);
          nextMapInputIndex++;
          if(nextMapInputIndex>=nnmap) return;
        } else {
        }

      // the following is supposed to handle floating point and integer number
      // input
      } else {
        // try to form a number from the string
        try {
          data[nextMapInputIndex][componentIndex] = new Double(number);
          nextMapInputIndex++;
        } catch(NumberFormatException nfe) {
        }
      }

    }

  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *    basic methods like toString and others for convenience               *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */
 
  public String toString() {
    return toString(0);
  }
 
  public String toString(int n) {
    String s = "";
    s = "\n"
      + getTabs(n)
      + "String output for the MapObject\n\n";
    s = s
      + getTabs(n)
      + "MapObject Name:  " + getMapName()
      + "\n\n";
    s = s
      + getTabs(n)
      + "  Number of blocks in first coordinate:  "
      + new Integer(getNumberOfXBlocks()).toString()
      + "\n";
    s = s
      + getTabs(n)
      + "  Number of blocks in second coordinate:  "
      + new Integer(getNumberOfYBlocks()).toString()
      + "\n";
    s = s
      + getTabs(n)
      + "  Number of blocks in third coordinate:  "
      + new Integer(getNumberOfZBlocks()).toString()
      + "\n";
    if(data != null) {
      s = s
        + getTabs(n)
        + "  Data is set?  Yes"
        + "\n";
    } else {
      s = s
        + getTabs(n)
        + "  Data is set?  No"
        + "\n";
    }
    s = s
      + getTabs(n)
      + "  Map Depth:  "
      + new Integer(depth()).toString()
      + "\n";
    if(associatedMapObjects != null) {
      s = s
        + "\n"
	+ getTabs(n)
        + "  Associated Maps:  \n";
      int k = associatedMapObjects.size();
      for(int i=0; i<k; i++) {
        MapObject mo = (MapObject) associatedMapObjects.get(i);
        s = s
          + getTabs(n)
          + mo.toString(1);
      }
      String [] sa = getMapObjectHierarchy();
      if(sa != null) {
        s = s
          + "\n\n"
          + getTabs(n)
          + "  Associated Map List:  \n\n";
        int l = sa.length;
        for(int i=0; i<l; i++) {
          s = s
            + getTabs(n+1)
            + sa[i]
            + "\n";
        }
      }
    } else {
      s = s
        + "\n"
	+ getTabs(n)
	+ "  No Associated Maps:  \n";
    }
    return s;
  }

  public String getTabs(int n) {
    String tabChar = "\t";
    String s = "";
    for(int i=0; i<n; i++) s = s + tabChar;
    return s;
  }

  public boolean equals(MapObject mo) {
    if(!getMapName().equals(mo.getMapName())) return false;
    if(getNumberOfXBlocks() != mo.getNumberOfXBlocks()) return false; 
    if(getNumberOfYBlocks() != mo.getNumberOfYBlocks()) return false; 
    if(getNumberOfZBlocks() != mo.getNumberOfZBlocks()) return false; 
    Double [][] d = mo.getData();
    for(int i=0; i<nnmap; i++) {
      for(int l=0; l<nComp; l++) if(!data[i][l].equals(d[i][l])) return false; 
    }
    return true;
  }
 
  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *    main method for testing the interface                                *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */
  public static void main(String args[]) {
    int i, j, k, l;
    int nx = 10;
    int ny = 10;
    int nz = 1;
    int ntot = nx * ny * nz;
    // a first simple test on a single block
    System.out.println("************* next test output ******************\n\n");
    MapObject mo = null;
    Double [][] data = new Double[ntot][9];
    for(i=0; i<ntot; i++) for(l=0; l<9; l++) data[i][l] = new Double(i);
    try {
      mo = new MapObject("main_map", data, nx, ny, nz);
      System.out.println(mo.toString());
    } catch(BadArraySizeException bse) {
    }
    System.out.println("************* end test output ******************\n\n");
    // add an associate map to this map
    System.out.println("************* next test output ******************\n\n");
    MapObject amo = null;
    Double [][] adata = new Double[4][9];
    for(i=0; i<4; i++) for(l=0; l<9; l++) adata[i][l] = new Double(i);
    try {
      amo = new MapObject("first_level_map", adata, 2, 2, 1);
      if(mo != null) mo.addAssociatedMap(amo);
      System.out.println(mo.toString());
    } catch(BadArraySizeException bse) {
    }
    System.out.println("************* end test output ******************\n\n");
    // go one more level to make sure that this crap works correctly
    System.out.println("************* next test output ******************\n\n");
    MapObject a11mo = null;
    MapObject a12mo = null;
    Double [][] a11data = new Double[4][9];
    Double [][] a12data = new Double[4][9];
    for(i=0; i<4; i++) for(l=0; l<9; l++) {
      a11data[i][l] = new Double(i);
      a12data[i][l] = new Double(l);
    }
    try {
      a11mo = new MapObject("second_level_map_11", a11data, 2, 2, 1);
      a12mo = new MapObject("second_level_map_12", a12data, 2, 2, 1);
      if(amo != null) amo.addAssociatedMap(a11mo);
      if(amo != null) amo.addAssociatedMap(a12mo);
      System.out.println(mo.toString());
    } catch(BadArraySizeException bse) {
    }
    mo.writeMapHierarchy();
    mo.buildDirectoryStructure();
    mo.writeMaps();
    System.out.println("load the map  " + mo.loadMap());
    System.out.println("************* end test output ******************\n\n");
  }

  // local variables
  private String mapName = null;
  private Double [][] data;
  private int nComp = 1;
  private int nCompRead = 0;
  private int componentIndex;
  private int nextMapInputIndex;
  private int nmapx;
  private int nmapy;
  private int nmapz;
  private int nxny;
  private int nnmap;
  private Double xleft;
  private Double xright;
  private Double yleft;
  private Double yright;
  private Double zleft;
  private Double zright;
  private Double [] xgrid;
  private Double [] ygrid;
  private Double [] zgrid;
  private MapObject parentMapObject = null;
  private String parentMapObjectName = null;

  private Vector associatedMapObjects = null;

  private boolean modified = false;
  private boolean saved = true;

  private boolean debug = false;

}
