c
c this common block contains information and storage for the
c fracture translation code.
c
c General comments about the data structures:
c
c 1. The description of the variables is given below.
c
c 2. the parameter statements are included in this file to keep things a little
c    easier in terms of coding subroutines.
c
c parameter definitions for the fracture blocks
c
c  the following parameter statements set up the storage and the like for the
c 2-d fracture code
c
c   nfrac - the maximum number of fracture map nodes allowable in the
c          simulation
c   nodefrac - the maximum number of nodes allowed in either coordinate
c              direction. Note that this should be the maximum of the
c              desired nodes in the x direction and the y direction
c
c relationships
c
c   nnfrac = nxfrac * nyfrac * nzfrac
c   nnvx = ( nxfrac + 1 ) * nyfrac * nzfrac
c   nnvy = ( nyfrac + 1 ) * nxfrac * nzfrac
c   nnvz = ( nzfrac + 1 ) * nxfrac * nyfrac
c
c and then
c
c   nfrac >= nnfrac
c   nodefrac >= max(nxfrac,nyfrac,nzfrac)
c
c for the next to lowest order mixed finite element method the values
c are computed internally and can be taken by multiplying nxfrac and nyfrac by
c two. also, one should make sure that nvelx is the maximum number of
c the velocity nodes, since it is used in any ambiguous definitions,
c such as routines that are symmetrically in the x and y components.
c routines lusol and reduce are examples of such routines.
c
c note that upon change of the parameter file here, the entire library
c and all drivers must be recompiled and linked in order to ensure that
c the simulators are on the same wavelength. There are facilities
c already built to handle this in the utility directory under sim2p2d.
c
      parameter(nodefrac=5000, nfrac=1000000)
      parameter(nfracd=200,nfraca=20,nfracl=20,nfraccdf=100)
c
c fracture map variables
c
c   fracmap - the array of fracture map at the next time step (2d array)
c   nxfrac - the number of fracture map nodes in first coordinate direction
c   nyfrac - the number of fracture map nodes in second coordinate direction
c   nzfrac - the number of fracture map nodes in third coordinate direction
c
      common /frac/ fracmap(nfrac),nxfrac,nyfrac,nzfrac,nxfracm1
     1,nyfracm1,nzfracm1,nxfracp1,nyfracp1,nzfracp1,nnfrac,nnvx
c
c some extra parameters for the higher dimensional comptuations
c
c   nxnyfrac = nxfrac * nyfrac
c   nynzfrac = nyfrac * nzfrac
c   nxnzfrac = nxfrac * nzfrac
c
     1,nxnyfrac,nynzfrac,nxnzfrac
c
c fracture mesh variables
c
c   xfrac - the cell centers of the first coordinate in the fracture mesh
c   yfrac - the cell centers of the second coordinate in the fracture map mesh
c   zfrac - the cell centers of the third coordinate in the fracture map mesh
c   xgrfrac - the grid lines of the first coordinate in the fracture map mesh
c   ygrfrac - the grid lines of the second coordinate in the fracture map mesh
c   zgrfrac - the grid lines of the third coordinate in the fracture map mesh
c   dxfrac - the cell sizes of the first coordinate in the fracture map mesh
c   dyfrac - the cell sizes of the second coordinate in the fracture map mesh
c   dzfrac - the cell sizes of the third coordinate in the fracture map mesh
c   fracxleft - the original left end point for the domain in first coord.
c   fracxright - the original right end point for the domain in first coord.
c   fracyleft - the original left end point for the domain in second coord.
c   fracyright - the original right end point for the domain in second coord.
c   fraczleft - the original left end point for the domain in third coord.
c   fraczright - the original right end point for the domain in third coord.
c
     1,xfrac(0:nodefrac+1),yfrac(0:nodefrac+1),zfrac(0:nodefrac+1)
     1,xgrfrac(0:nodefrac+1),ygrfrac(0:nodefrac+1),zgrfrac(0:nodefrac+1)
     1,dxfrac(nodefrac),dyfrac(nodefrac),dzfrac(nodefrac),fracxleft
     1,fracxright,fracyleft,fracyright,fraczleft,fraczright
c
c mask array for the permeability variable
c
c   maskfrac - vector to mask out some permeability nodes - normally used to
c             communicate input to the code or to fix permeability that are
c             constant and known in the domain
c
     1,maskfrac(nfrac)
c
c data for the type 2 bedform generation - archetypes
c
c   nfracarchedata - the number of data points used in the determination of the
c                   archetypes
c   nfracarches - the number of archetypes to read in
c   nfracarchelen - the length of the archetype vectors
c   fracarches - the vector of archetypes
c   fracarchebeta - the coefficients for the construction of the archetypes from
c                  the data
c   fracarchealpha - the coefficients for computing the projections of the data
c                   to the archetypes
c   fracarcheamean - the means for the archetype multipliers
c   fracarcheastdv - the standard deviations for the archetype multipliers
c   ifracarchedist - the type of distribution to sample from
c                   0 -> uniform
c                   1 -> normal
c                   2 -> log-normal
c   fracarchecdf - cdf values for the archetype projection coefficients - alpha's
c
     1,nfracarchedata,nfracarches,nfracarchelen
     1,fracarches(nfracl,nfraca),fracarchebeta(nfracd,nfraca)
     1,fracarchealpha(nfraca,nfracd),fracarcheamean(nfraca)
     1,fracarcheastdv(nfraca),ifracarchedist
     1,fracarchecdf(0:nfraccdf+1,nfraca)
c
c permeability values on the fracture map grid
c
c   frackxx - (1,1) tensor entry in the permeability
c   frackxy - (1,2) tensor entry in the permeability
c   frackxz - (1,3) tensor entry in the permeability
c   frackyx - (2,1) tensor entry in the permeability
c   frackyy - (2,2) tensor entry in the permeability
c   frackyz - (2,3) tensor entry in the permeability
c   frackzx - (3,1) tensor entry in the permeability
c   frackzy - (3,2) tensor entry in the permeability
c   frackzz - (3,3) tensor entry in the permeability
c                fracimentary map simulator.
c
     1,frackxx(nfrac),frackxy(nfrac),frackxz(nfrac)
     1,frackyx(nfrac),frackyy(nfrac),frackyz(nfrac)
     1,frackzx(nfrac),frackzy(nfrac),frackzz(nfrac)
     1,t1frackxx(nfrac),t1frackxy(nfrac),t1frackxz(nfrac)
     1,t1frackyx(nfrac),t1frackyy(nfrac),t1frackyz(nfrac)
     1,t1frackzx(nfrac),t1frackzy(nfrac),t1frackzz(nfrac)
c
c restart parameter
c
c   ifracrstrt - a parameter to tell the code whether or not to restart the
c               simulator from the last output step
c   fracrstrt = restart string:
c              RESTART -> restart the simulation
c              RUN COMPLETE -> the simulation is complete and there is no need
c                              to run.
c              any thing else will let the simulation start from the beginning
c
     1,ifracrstrt,fracrstrt
      character*12 fracrstrt
c
c force the map to be integers
c ----------------------------
c
      integer fracmap
      data fracarchebw / 0.05 /
