c
c this common block contains information and storage for the
c mixed finite element method approach to the homogenization
c process
c
c General comments about the data structures:
c
c 1. the arrays are two-dimensional; the first dimension containing the nodal
c    values and the second dimension containing the phases and the average
c    or velocity. the average value is always stored in the first
c    component of the second argument of the array. this facilitates easier
c    cases when nhommfeph=2.
c
c 2. The description of the variables is given below.
c
c 3. the parameter statements are included in this file to keep things a little
c    easier in terms of coding subroutines.
c
c 4. this common block MUST appear before all other common blocks
c    like hommfe_matrix_common so that the parameter statement can be loaded
c    for setting up the array lengths
c
c parameter definitions for the mfe homogenization blocks
c
c  the following parameter statements set up the storage and the like for the
c 3-d mfe homogenization code
c
c   nhommfe - the maximum number of mfe homogenization nodes allowable in the
c          simulation
c   nhomvelx - the maximum number of x component of velocity nodes allowable
c          in the simulation - always choose nhomvelx > nhomvely
c   nhomvely - the maximum number of y component of velocity nodes allowable
c          in the simulation
c   nhomvelz - the maximum number of z component of velocity nodes allowable
c          in the simulation
c   nnhomvar - the sum of the velocity and mfe homogenization dimensions
c   nodehommfe - the maximum number of nodes allowed in either coordinate
c              direction. Note that this should be the maximum of the
c              desired nodes in the x direction and the y direction
c
c relationships
c
c   the maximum number of mfe homogenization nodes and the maximum of nhomvelx,
c nhomvely and nhomvelz are should be related by the following. if nxs is the
c number of c nodes in the x direction and ny is the number of unknowns in the y
c direction then for the lowest order mixed finite element method one
c has
c
c   nnhommfe = nxhommfe * nyhommfe * nzhommfe
c   nnhommfevx = ( nxhommfe + 1 ) * nyhommfe * nzhommfe
c   nnhommfevy = ( nyhommfe + 1 ) * nxhommfe * nzhommfe
c   nnhommfevz = ( nzhommfe + 1 ) * nxhommfe * nyhommfe
c
c and then
c
c   nhommfe>= nnhommfe
c   nhomvelx >= nnhommfevx
c   nhomvely >= nnhommfevy
c   nhomvelz >= nnhommfevz
c   nodehommfe >= max(nxhommfe,nyhommfe,nzhommfe)
c
c for the next to lowest order mixed finite element method the values are
c computed internally and can be taken by multiplying nxhommfe and nyhommfe by
c two. also, one should make sure that nhomvelx is the maximum number of
c the velocity nodes, since it is used in any ambiguous definitions,
c such as routines that are symmetrically in the x and y components.
c routines lusol and reduce are examples of such routines.
c
c note that upon change of the parameter file here, the entire library
c and all drivers must be recompiled and linked in order to ensure that
c the simulators are on the same wavelength. There are facilities
c already built to handle this in the utility directory under sim2p2d.
c
c      parameter(nodehommfe=300, nhommfe=80000)
c      parameter(nhomvelx=90000, nhomvely=90000, nhomvelz=90000)
c      parameter(nnhomvar=nhommfe+nhomvelx+nhomvely+nhomvelz)
      parameter(nodehommfe=100, nhommfe=10000)
      parameter(nhomvelx=20000, nhomvely=20000, nhomvelz=20000)
      parameter(nnhomvar=nhommfe+nhomvelx+nhomvely+nhomvelz)
c
c homogenization variables
c
c   w - the array of at the next time step (2d array)
c   nxhommfe - the number of nodes in the first coordinate direction
c   nyhommfe - the number of nodes in the second coordinate direction
c   nzhommfe - the number of nodes in the third coordinate direction
c   ihommfeeq - the problem number being treated
c
      common /hommfe/ w(nhommfe,3),nxhommfe,nyhommfe,nzhommfe,nxhommfem1       
     1,nyhommfem1,nzhommfem1,nxhommfep1,nyhommfep1,nzhommfep1,nnhommfe
     1,nnhommfevx,nnhommfevy,nnhommfevz,nnhommfec,ihommfeeq
c
c some extra parameters for the higher dimensional comptuations
c
c   nxnyhommfe = nxhommfe * nyhommfe
c   nynzhommfe = nyhommfe * nzhommfe
c   nxnzhommfe = nxhommfe * nzhommfe
c   nhvtot = nnhommfe + nnhommfevx + nnhommfevy + nnhommfevz
c
c some older parameters left in for convenience
c
c   nshftv2 = nnhommfevx
c   nshftv3 = nshftv2 + nnhommfevy
c   nshftp  = nshftv3 + nnhommfevz
c
c 2d velocity grid parameters
c
c   nnhommfevx2d = nxhommfep1 * nyhommfe
c   nnhommfevy2d = nyhommfep1 * nzhommfe
c   nnhommfevz2d = nxhommfe * nzhommfep1
c
     1,nxnyhommfe,nynzhommfe,nxnzhommfe,nhvtot
     1,nnhommfevx2d,nnhommfevy2d,nnhommfevz2d
     1,nshftv2,nshftv3,nshftp
c
c velocity variables
c
c   hommfevelx - the x component of the total or phase fluid velocity
c   hommfevely - the y component of the total or phase fluid velocity
c   hommfevelz - the z component of the total or phase fluid velocity
c
     1,hommfevelx(nhomvelx,3),hommfevely(nhomvely,3)
     1,hommfevelz(nhomvelz,3),nxvel,nyvel,nzvel
c
c hommfe mesh variables
c
c   xhommfe - the cell centers of the first coordinate in the mesh
c   yhommfe - the cell centers of the second coordinate in the mesh
c   zhommfe - the cell centers of the third coordinate in the mesh
c   xgrhommfe - the grid lines of the first coordinate in the mesh
c   ygrhommfe - the grid lines of the second coordinate in the mesh
c   zgrhommfe - the grid lines of the third coordinate in the mesh
c   dxhommfe - the cell sizes of the first coordinate in the mesh
c   dyhommfe - the cell sizes of the second coordinate in the mesh
c   dzhommfe - the cell sizes of the third coordinate in the mesh
c   hommfexleft - the original left end point for the domain in first coord.
c   hommfexright - the original right end point for the domain in first coord.
c   hommfeyleft - the original left end point for the domain in second coord.
c   hommfeyright - the original right end point for the domain in second coord.
c   hommfezleft - the original left end point for the domain in third coord.
c   hommfezright - the original right end point for the domain in third coord.
c
     1,xhommfe(0:nodehommfe+1),yhommfe(0:nodehommfe+1)
     1,zhommfe(0:nodehommfe+1),xgrhommfe(0:nodehommfe+1)
     1,ygrhommfe(0:nodehommfe+1),zgrhommfe(0:nodehommfe+1)
     1,dxhommfe(nodehommfe),dyhommfe(nodehommfe)
     1,dzhommfe(nodehommfe),hommfexleft,hommfexright
     1,hommfeyleft,hommfeyright,hommfezleft,hommfezright
c
c mask array for the variable
c
c   maskhommfe - vector to mask out some nodes - normally used to
c             communicate input to the code or to fix values that are
c             constant and known in the domain
c             maskhommfe(i)=0; no mask out of the node i (default)
c             maskhommfe(i)!=0; mask out of the node i
c   maskvx - vector to mask out some first coordinate velocity nodes - same
c            comment as for the masking vector.        
c   maskvy - vector to mask out some second coordinate velocity nodes - same
c            comment as for the masking vector.        
c   maskvz - vector to mask out some third coordinate velocity nodes - same
c            comment as for the masking vector.        
c
     1,maskhommfe(nhommfe),maskvx(nhomvelx),maskvy(nhomvely)
     1,maskvz(nhomvelz)
c
c the temporary mask vectors for the and velocity variables
c
c   ibithommfe - temporary masking vector for the variable
c   ibithomvx - temporary masking vector for the first component of the velocity
c   ibithomvy - temporary masking vector for the second component of the
c               velocity
c   ibithomvz - temporary masking vector for the third component of the velocity
c
     1,ibithommfe(nhommfe),ibithomvx(nhomvelx),ibithomvy(nhomvely)
     1,ibithomvz(nhomvelz)
c
c permeability values on the grid
c
c   hommfekxx - (1,1) tensor entry in the permeability
c   hommfekxy - (1,2) tensor entry in the permeability
c   hommfekxz - (1,3) tensor entry in the permeability
c   hommfekyx - (2,1) tensor entry in the permeability
c   hommfekyy - (2,2) tensor entry in the permeability
c   hommfekyz - (2,3) tensor entry in the permeability
c   hommfekzx - (3,1) tensor entry in the permeability
c   hommfekzy - (3,2) tensor entry in the permeability
c   hommfekzz - (3,3) tensor entry in the permeability
c   ihommfekdiag - a parameter that keeps track of whether or not the
c                  permeability tensor on the grid is diagonal or not.
c                  ihommfeprmdiag=0; the tensor is full
c                  ihommfeprmdiag=1; the tensor is diagonal
c                  the code assumes the tensor is full and will compute values
c                  based on this.
c   ihommfescalprm - a flag to tell the simulator whether or not to scale
c                 the permeability.
c                 ihommfescalprm=0; no scaling
c                 ihommfescalprm=1; orders of magnitude scaling
c   hommfescalprm - a scaling factor for the permeability to get things away
c                   from terribly small numbers that cause the matrix algebra to
c                   have problems - this is used to rescale the input
c                   permeabilities, the initial , and the gravitational constantc                   in the simulator.
c
     1,hommfekxx(nhommfe),hommfekxy(nhommfe),hommfekxz(nhommfe)
     1,hommfekyx(nhommfe),hommfekyy(nhommfe),hommfekyz(nhommfe)
     1,hommfekzx(nhommfe),hommfekzy(nhommfe),hommfekzz(nhommfe)
     1,ihommfekdiag,ihommfescalprm,hommfescalprm
c
c the inverted permeabilities that are used in the mixed finite element
c solution where the system is reduced to a system in only the 
c variable
c
c   t1homkxx - (1,1) tensor entry in the inverted permeability tensor
c   t1homkxy - (1,2) tensor entry in the inverted permeability tensor
c   t1homkxz - (1,3) tensor entry in the inverted permeability tensor
c   t1homkyx - (2,1) tensor entry in the inverted permeability tensor
c   t1homkyy - (2,2) tensor entry in the inverted permeability tensor
c   t1homkyz - (2,3) tensor entry in the inverted permeability tensor
c   t1homkzx - (3,1) tensor entry in the inverted permeability tensor
c   t1homkzy - (3,2) tensor entry in the inverted permeability tensor
c   t1homkzz - (3,3) tensor entry in the inverted permeability tensor
c
c t for temporary
c n for the number of the temporary array in the sequence
c
     1,t1homkxx(nhommfe),t1homkxy(nhommfe),t1homkxz(nhommfe)
     1,t1homkyx(nhommfe),t1homkyy(nhommfe),t1homkyz(nhommfe)
     1,t1homkzx(nhommfe),t1homkzy(nhommfe),t1homkzz(nhommfe)
c
c numerical methods information for the routine
c
c   hommfecgtol - the tolerance to use in the conjugate gradient routine for
c              convergence
c
     1,hommfecgtol
c
c boundary condition parameters
c
c each of the following parameters will be set to one of the values
c
c   value =  0 => no flow condition on the given plane
c   value = -1 => fixed velocity condition on the given plane for which the
c                 code must pad the arrays and domain
c   value =  1 => fixed velocity condition on the given plane for which the
c                 code has already padded the arrays and domain
c   value = -2 => time dependent velocity condition on the given plane for
c                 which the code must pad the arrays and domain
c   value =  2 => time dependent velocity condition on the given plane for
c                 which the code has already padded the arrays and domain
c   value = -3 => fixed condition on the given plane for which the
c                 code must pad the arrays and domain
c   value =  3 => fixed condition on the given plane for which the
c                 code has already padded the arrays and domain
c   value = -4 => time dependent condition on the given plane for
c                 which the code must pad the arrays and domain
c   value =  4 => time dependent condition on the given plane for
c                 which the code has already padded the arrays and domain
c
c   note that the negative value will normally be set external to the simulator
c   and will be reset to a positive value after the padding is done (when
c   necessary).
c
c   ihommfebndxyl - the flag for the type of boundary condition to impose on the
c                plane perpendicular to the third coordinate axis on the left
c                end of the domain
c   ihommfebndxyr - the flag for the type of boundary condition to impose on the
c                plane perpendicular to the third coordinate axis on the right
c                end of the domain
c   ihommfebndxzl - the flag for the type of boundary condition to impose on the
c                plane perpendicular to the second coordinate axis on the left
c                end of the domain
c   ihommfebndxzr - the flag for the type of boundary condition to impose on the
c                plane perpendicular to the second coordinate axis on the right
c                end of the domain
c   ihommfebndyzl - the flag for the type of boundary condition to impose on the
c                plane perpendicular to the first coordinate axis on the left
c                end of the domain
c   ihommfebndyzr - the flag for the type of boundary condition to impose on the
c                plane perpendicular to the first coordinate axis on the right
c                end of the domain
c
     1,ihommfebndxyl,ihommfebndxyr
     1,ihommfebndxzl,ihommfebndxzr
     1,ihommfebndyzl,ihommfebndyzr
