c
c this common block contains information and storage for the
c pressure and velocities in the simulation.
c
c General comments about the data structures:
c
c 1. the arrays are two-dimensional; the first dimension containing the nodal
c    values and the second dimension containing the phases and the average
c    pressure or velocity. the average value is always stored in the first
c    component of the second argument of the array. this facilitates easier
c    cases when nprsph=2.
c
c 2. The description of the variables is given below.
c
c 3. the parameter statements are included in this file to keep things a little
c    easier in terms of coding subroutines.
c
c 4. this common block MUST appear before all other pressure common blocks
c    like prs_matrix_common so that the parameter statement can be loaded
c    for setting up the array lengths
c
c parameter definitions for the pressure blocks
c
c  the following parameter statements set up the storage and the like for the
c 3-d pressure code
c
c   npres - the maximum number of pressure nodes allowable in the
c          simulation
c   nvelx - the maximum number of x component of velocity nodes allowable
c          in the simulation - always choose nvelx > nvely
c   nvely - the maximum number of y component of velocity nodes allowable
c          in the simulation
c   nvelz - the maximum number of z component of velocity nodes allowable
c          in the simulation
c   nnvar - the sum of the velocity and pressure dimensions
c   nodeprs - the maximum number of nodes allowed in either coordinate
c              direction. Note that this should be the maximum of the
c              desired nodes in the x direction and the y direction
c   nprsw - the maximum number of wells allowed in the simulator
c   nprsph - maximum number of phases allowed in the simulator
c   nprsrp - maximum number of data points in the relative permeability arrays
c            for the phases
c   nprsdd - maximum number of data points in the density arrays for the phases
c   nprsvd - maximum number of data points in the viscosity arrays for the
c            phases
c
c relationships
c
c   the maximum number of pressure nodes and the maximum of nvelx, nvely and
c nvelz are should be related by the following. if nxs is the number of
c nodes in the x direction and ny is the number of unknowns in the y
c direction then for the lowest order mixed finite element method one
c has
c
c   nnprs = nxprs * nyprs * nzprs
c   nnvx = ( nxprs + 1 ) * nyprs * nzprs
c   nnvy = ( nyprs + 1 ) * nxprs * nzprs
c   nnvz = ( nzprs + 1 ) * nxprs * nyprs
c
c and then
c
c   npres >= nnprs
c   nvelx >= nnvx
c   nvely >= nnvy
c   nvelz >= nnvz
c   nodeprs >= max(nxprs,nyprs,nzprs)
c
c for the next to lowest order mixed finite element method the values
c are computed internally and can be taken by multiplying nxprs and nyprs by
c two. also, one should make sure that nvelx is the maximum number of
c the velocity nodes, since it is used in any ambiguous definitions,
c such as routines that are symmetrically in the x and y components.
c routines lusol and reduce are examples of such routines.
c
c note that upon change of the parameter file here, the entire library
c and all drivers must be recompiled and linked in order to ensure that
c the simulators are on the same wavelength. There are facilities
c already built to handle this in the utility directory under sim2p2d.
c
      parameter(nodeprs=330, npres=50000)
      parameter(nvelx=60000, nvely=60000, nvelz=90000)
      parameter(nnvar=npres+nvelx+nvely+nvelz)
      parameter(nprsw=10)
      parameter(nprsph=5, nprsrp=50, nprsdd=20, nprsvd=20, nprscd=20)
c
c pressure variables
c
c   p - the array of pressures at the next time step (2d array)
c   pold - the array of pressures at the old time step
c   nxprs - the number of pressure nodes in the first coordinate direction
c   nyprs - the number of pressure nodes in the second coordinate direction
c   nzprs - the number of pressure nodes in the third coordinate direction
c
      common /press/ p(npres,0:nprsph),pold(npres,0:nprsph),nxprs,nyprs
     1,nzprs,nxprsm1,nyprsm1,nzprsm1,nxprsp1,nyprsp1,nzprsp1,nnprs,nnvx
     1,nnvy,nnvz,nnprsc
c
c minimum and maximum values
c
c   prsprsmin - the minimum pressure computed in the simulation
c   prsprsmax - the maximum pressure computed in the simulation
c   iprsprsdmin - the index where the minimum pressure exists
c   iprsprsdmax - the index where the maximum pressure exists
c   velxmin - the minimum velocity computed in the simulation in the x direction
c   velxmax - the maximum velocity computed in the simulation in the x direction
c   velymin - the minimum velocity computed in the simulation in the y direction
c   velymax - the maximum velocity computed in the simulation in the y direction
c   velzmin - the minimum velocity computed in the simulation in the z direction
c   velzmax - the maximum velocity computed in the simulation in the z direction
c   prssatmin - the minimum saturation computed on the pressure grid
c   prssatmax - the maximum saturation computed on the pressure grid
c   iprssatdmin - the index where the minimum saturation exists
c   iprssatdmax - the index where the maximum saturation exists
c
     1,prsprsmin(0:nprsph),prsprsmax(0:nprsph)
     1,iprsprsmin(nprsph),iprsprsmax(nprsph)
     1,velxmin(0:nprsph),velxmax(0:nprsph)
     1,velymin(0:nprsph),velymax(0:nprsph)
     1,velzmin(0:nprsph),velzmax(0:nprsph)
     1,prssatmin(nprsph),prssatmax(nprsph)
     1,iprssatmin(nprsph),iprssatmax(nprsph)
c
c some extra parameters for the higher dimensional comptuations
c
c   nxnyprs = nxprs * nyprs
c   nynzprs = nyprs * nzprs
c   nxnzprs = nxprs * nzprs
c   npvtot = nnprs + nnvx + nnvy + nnvz
c
c some older parameters left in for convenience
c
c   nshftv2 = nnvx
c   nshftv3 = nshftv2 + nnvy
c   nshftp  = nshftv3 + nnvz
c
     1,nxnyprs,nynzprs,nxnzprs,npvtot
     1,nshftv2,nshftv3,nshftp
c
c velocity variables
c
c   velx - the x component of the total or phase fluid velocity
c   vely - the y component of the total or phase fluid velocity
c   velz - the z component of the total or phase fluid velocity
c
     1,velx(nvelx,0:nprsph),vely(nvely,0:nprsph),velz(nvelz,0:nprsph)
     1,nxvel,nyvel,nzvel
c
c pressure mesh variables
c
c   xprs - the cell centers of the first coordinate in the pressure mesh
c   yprs - the cell centers of the second coordinate in the pressure mesh
c   zprs - the cell centers of the third coordinate in the pressure mesh
c   xgrprs - the grid lines of the first coordinate in the pressure mesh
c   ygrprs - the grid lines of the second coordinate in the pressure mesh
c   zgrprs - the grid lines of the third coordinate in the pressure mesh
c   dxprs - the cell sizes of the first coordinate in the pressure mesh
c   dyprs - the cell sizes of the second coordinate in the pressure mesh
c   dzprs - the cell sizes of the third coordinate in the pressure mesh
c   prsxleft - the original left end point for the domain in first coord.
c   prsxright - the original right end point for the domain in first coord.
c   prsyleft - the original left end point for the domain in second coord.
c   prsyright - the original right end point for the domain in second coord.
c   prszleft - the original left end point for the domain in third coord.
c   prszright - the original right end point for the domain in third coord.
c   xprshyd - the first coordinate of the hydrostatic pressure reference point
c   yprshyd - the second coordinate of the hydrostatic pressure reference point
c   zprshyd - the third coordinate of the hydrostatic pressure reference point
c   prshyd - the base hydrostatic pressure at the pressure nodes
c   iprshyd - a parameter to let the code know if the pressure field is driven
c             or is due only to hydrostatic forces on the fluids. this is
c             determined in the computations for the right hand side of the
c             system of equations - see prsrhs.f for more
c             iprshyd=0; the field is driven
c             iprshyd=1; the field is hyrdrostatic
c
     1,xprs(0:nodeprs+1),yprs(0:nodeprs+1),zprs(0:nodeprs+1)
     1,xgrprs(0:nodeprs+1),ygrprs(0:nodeprs+1),zgrprs(0:nodeprs+1)
     1,dxprs(nodeprs),dyprs(nodeprs),dzprs(nodeprs),prsxleft,prsxright
     1,prsyleft,prsyright,prszleft,prszright,xprshyd,yprshyd,zprshyd
     1,prshyd(npres,0:nprsph),iprshyd
c
c mask array for the pressure variable
c
c   maskprs - vector to mask out some pressure nodes - normally used to
c             communicate input to the code or to fix pressures that are
c             constant and known in the domain
c             maskprs(i)=0; no mask out of the pressure node i (default)
c             maskprs(i)!=0; mask out of the pressure node i
c   maskvx - vector to mask out some first coordinate velocity nodes - same
c            comment as for the pressure masking vector.        
c   maskvy - vector to mask out some second coordinate velocity nodes - same
c            comment as for the pressure masking vector.        
c   maskvz - vector to mask out some third coordinate velocity nodes - same
c            comment as for the pressure masking vector.        
c
     1,maskprs(npres),maskvx(nvelx),maskvy(nvely),maskvz(nvelz)
c
c phase information
c
c   nprsphase - the number of pressure phases - note that the arrays are
c               dimensioned one more than the number of phases to account for
c               the total fluid velocity storage - this is always in the first
c               entry of the second index in the array definition.
c
     1,nprsphase 
c
c pressure drawdown indicator
c
c   idrwdn - parameter to tell the code to compute drawdown from the initial
c            head profile
c            idrwdn=1; compute the drawdown
c            idrwdn.ne.1; do not compute the drawdown
     1,idrwdn
c
c the temporary mask vectors for the pressure and velocity variables
c
c   ibitprs - temporary masking vector for the pressure variable
c   ibitvx - temporary masking vector for the first component of the velocity
c   ibitvy - temporary masking vector for the second component of the velocity
c   ibitvz - temporary masking vector for the third component of the velocity
c
     1,ibitprs(npres),ibitvx(nvelx),ibitvy(nvely),ibitvz(nvelz)
c
c smooth forcing function acting on the pressure equation
c
c   iprssff - parameter to determine what type of smooth forcing
c             function will be used in the simulation
c             iprssff = 0; the forcing function is 0 everywhere
c             iprssff anything else - see the routine sff.f
c
     1,iprssff
c
c permeability values on the pressure grid
c
c   prskxx - (1,1) tensor entry in the permeability
c   prskxy - (1,2) tensor entry in the permeability
c   prskxz - (1,3) tensor entry in the permeability
c   prskyx - (2,1) tensor entry in the permeability
c   prskyy - (2,2) tensor entry in the permeability
c   prskyz - (2,3) tensor entry in the permeability
c   prskzx - (3,1) tensor entry in the permeability
c   prskzy - (3,2) tensor entry in the permeability
c   prskzz - (3,3) tensor entry in the permeability
c   iprskdiag - a parameter that keeps track of whether or not the permeability
c               tensor on the pressure grid is diagonal or not.
c               iprsprmdiag=0; the tensor is full
c               iprsprmdiag=1; the tensor is diagonal
c               the code assumes the tensor is full and will compute values
c               based on this.
c   iprsscalprm - a flag to tell the simulator whether or not to scale
c                 the permeability.
c                 iprsscalprm=0; no scaling
c                 iprsscalprm=1; orders of magnitude scaling
c   prsscalprm - a scaling factor for the permeability to get things away from
c                terribly small numbers that cause the matrix algebra to have
c                problems - this is used to rescale the input permeabilities,
c                the initial pressure, and the gravitational constant in the
c                pressure simulator.
c
     1,prskxx(npres),prskxy(npres),prskxz(npres)
     1,prskyx(npres),prskyy(npres),prskyz(npres)
     1,prskzx(npres),prskzy(npres),prskzz(npres)
     1,iprskdiag,iprsscalprm,prsscalprm
c
c the inverted permeabilities that are used in the mixed finite element
c solution where the system is reduced to a system in only the pressure
c variable
c
c   t1prskxx - (1,1) tensor entry in the inverted permeability tensor
c   t1prskxy - (1,2) tensor entry in the inverted permeability tensor
c   t1prskxz - (1,3) tensor entry in the inverted permeability tensor
c   t1prskyx - (2,1) tensor entry in the inverted permeability tensor
c   t1prskyy - (2,2) tensor entry in the inverted permeability tensor
c   t1prskyz - (2,3) tensor entry in the inverted permeability tensor
c   t1prskzx - (3,1) tensor entry in the inverted permeability tensor
c   t1prskzy - (3,2) tensor entry in the inverted permeability tensor
c   t1prskzz - (3,3) tensor entry in the inverted permeability tensor
c
c t for temporary
c n for the number of the temporary array in the sequence
c
     1,t1prskxx(npres),t1prskxy(npres),t1prskxz(npres)
     1,t1prskyx(npres),t1prskyy(npres),t1prskyz(npres)
     1,t1prskzx(npres),t1prskzy(npres),t1prskzz(npres)
c
     1,t2prskxx(npres),t2prskxy(npres),t2prskxz(npres)
     1,t2prskyx(npres),t2prskyy(npres),t2prskyz(npres)
     1,t2prskzx(npres),t2prskzy(npres),t2prskzz(npres)
c
c porosity values on the pressure grid
c
c   the porosity is assumed to be a linear function of the pressure variable
c   in the model and as such needs two coefficients for each node.
c
c   prspor0 - the coefficient for the constant term in the porosity linear
c             constituitve law for the pressure
c   prspor1 - the coefficient for the linear term in the porosity linear
c             constituitve law for the pressure
c
     1,prspor0(npres),prspor1(npres)
c
c saturation values on the pressure grid
c
c   prssat - the saturation values at the pressure grid nodes at the next level
c   prssatold - the saturation values at the pressure grid nodes at the old
c               level
c   maskprssat - vector to mask out computation of saturation values on some
c                pressure nodes
c                maskprssat(i)=0; no mask out of the pressure node i (default)
c                maskprssat(i)!=0; mask out of the pressure node i
c   ibitprssat - the temporary bit vector used for storing the saturation mask
c                on the pressure grid
c
     1,prssat(npres,nprsph),prssatold(npres,nprsph),maskprssat(npres)
     1,ibitprssat(npres)
c
c the relative permeability information for the phases and the reference
c densitites and viscosities for the phases in the pressure problem
c
c   prsrprm - the relative permeability curve information - the first index
c             addresses the data used in the definition of the relative
c             permeability curve. the first entry prsrprm(1,l) gives the
c             method of evaluation for the relative permeability for phase
c             l. the second entry in the column is the number of data points
c             (if necessary) used in the definition of the relative permeability
c             curve. the rest of the entries in the column are the data points
c             themselves. they are paired in ordered pairs of numbers which
c             are the function data values. the second index addresses the
c             phase. this array thus contains all the info on the relative
c             permeability of the phases in the problem. as examples,
c
c                  prsrprm(1,lphase) = 0
c
c             means the code expects to interpolate a set of data to determine
c             the value of the relative permeability for phase, lphase, and
c
c                  prsrprm(1,lphase) > 0
c
c             means the code will use an analytic definition for the relative
c             permeabiity. see anarprm for the choices. the value
c
c                  prsrprm(2,lphase) = 5
c
c             indicates the code whould use five points in defining the linear
c             spline for the function. the third point in the data is stored in
c             prsrprm(7,lphase) and prsrprm(8,lphase). 
c   prsdens - the phase density information - the description for prsrprm works
c             here.
c   prsvisc - the phase viscosity information - the description for prsrprm
c             works here.
c
c   note that in all these definitions that the independent variable is NOT
c   specified. for example, the independent variable is probably going to be
c   saturation. however, the density will most likely be dependent on the
c   total pressure at the point. these must be handled by the various codes.
c   it should also be noted that the data is assumed to be a function of a
c   single variable. in the analytic case where all the data is ignored,
c   it is easier to use relationships between variables. finally, the
c   dimensioning in the arrays should be clear - the number of entries handles
c   the ordered pairs plus the type and number of data points. 
c
     1,prsrprm(2+2*nprsrp,nprsph),prsdens(2+2*nprsdd,nprsph)
     1,prsvisc(2+2*nprsvd,nprsph)
c
c residual saturations
c
c   prsresidsat - the residual saturation of the phases in the simulation
c
     1,prsresidsat(nprsph)
c
c total mobility of the fluid on the pressure grid
c
c   prslambda - array containing the mobilities of the total fluid and all
c               phases as compute in the routine prstotmob
c   prslamold - array containing the mobilities from the previous time step.
c               these are initialized to zero. this forces the simulator to
c               compute a pressure solution on the first time step
c
     1,prslambda(npres,0:nprsph),prslamold(npres)
c
c density values in the total pressure/velocity formulation
c
c   prsgamma - this is the weighted density at each node that occurs in the
c              total velocity definition.
c
     1,prsgamma(npres)
c
c capillary pressure variables
c
c   prscap - the capillary pressure curve information - the first index
c            data used in the definition of the cappilary pressure data.
c            the first entry prscap(1,l) gives the first of the two phases
c            involved in the relationship. The second entry, prscap(2,l), is
c            the same for the second phase involved.
c
c                  prscap(1,2) = 3
c                  prscap(2,2) = 5 
c
c             means the code will relate phase 3 and five using the data in
c             the second capillary data location. The value in prscap(3,l)
c             contains the resdiual saturation for the first phase involved
c             and prscap(4,l) contains the same for the second phase. The
c             value in prscap(5,l) gives the number of data points to follow
c             for defining the capillary pressure curve. The rest of the array
c             defines the data:
c
c                  prscap(6,2) = first saturation value
c                  prscap(7,2) = first capillary pressure value
c                       *                *
c                       *                *
c                       *                *
c
c   iprscapwtyp - the type of weighting to use in the computation of the averge
c                 pressure
c                 iprscapwtyp=0; simple average of the phase pressures
c                 iprscapwtyp=1; linear saturation multiplier on the phase
c                                pressures
c                 iprscapwtyp=2; use the input multipliers on the phase
c                                pressures
c   prscapwt - weights used in the inversion of the phase pressure variables
c              in terms of the average and capillary pressures
c   mapprscap - a vector that keeps track of the ordering of the capillary
c               relationships. it is the necessary ordering to transform the
c               given matrix into the "companion matrix" form.
c
     1,prscap(nprscd,nprsph),iprscapwtyp,prscapwt(nprsph)
     1,mapcapprs(nprsph)
c
c capillary pressure on the pressure grid
c
c   pcap - the capillary pressures at the nodes between successive phases -
c          there are one fewer capillary values since the relationship is
c          between two phase pressures.
c
     1,pcap(npres,nprsph-1)
c
c pressure time step - this allows the code to take a pressure time step
c independently of the other pieces of the code.
c
c   prstep - the pressure time step
c   prstrt - initial time when the pressure code is called
c   prsend - final time when the pressure code is to exit
c   prstim - the time local to the pressure solution routines
c   prstepmax - the maximum time step allowed in the pressure code
c   iprsteptyp - the type of time stepping to do - check the time step routine
c                for the options on this parameter.
c   ninpsteps - the number of input time steps read into the code from the
c               input file 'timesteps.inp'
c
     1,prstep,prstrt,prsend,prstim,prstepmax,iprsteptyp,ninpsteps
c
c numerical methods information for the pressure routine
c
c   methprsol - the method to use for the pressure solution
c               methprsol=1; use a standard mixed finite element method with
c                            straight tensor inversion
c               methprsol=2; use the reduced mixed finite element method with
c                            decomposition of the tensor
c   prscgtol - the tolerance to use in the conjugate gradient routine for
c              convergence
c   nprsmaxiter - the maximum number of iterations to use in the iterative
c                solver for the linear system of pressures equations
c
     1,methprsol,prscgtol,nprsmaxiter
c
c well information for the pressure simulator
c
c   nprswells - the number wells in the simulation
c   maskprswell - a well mask to indicate whether the well is active in the
c                 simulation or not.
c                 maskprswell(i)=0; the well is not active
c                 maskprswell(i)!=0; the well is active
c   iprswelltyp - the type of well
c                 1 - a fixed total pressure is specified at the well (default)
c                 2 - a fixed total flux is specified at the well
c   prswellprs - the total pressure at each of the wells
c   prswellfr - the flow rate of each of the phases for each of the wells
c               the total flow rate is computed into the zero index as in
c               the other arrays
c   prswellarea - the area of injection for the well - this can be thought of
c                 as the area of the completion zone in the well bore
c   iprswelloc - the index location of the wells
c   xprswell - the first coordinate of the location of the wells
c   yprswell - the second coordinate of the location of the wells
c   zprswell - the third coordinate of the location of the wells
c   prswellvol - the volume of the well block cells in the reservoir
c
     1,nprswells,iprswelltyp(nprsw),maskprswell(nprsw)
     1,prswellprs(nprsw),prswellfr(nprsw,nprsph),prswellarea(nprsph)
     1,iprswelloc(nprsw),xprswell(nprsw),yprswell(nprsw),zprswell(nprsw)
     1,prswellblkvol(nprsw)
c
c injection and production data for the wells
c
c   prstotpv - the total pore volume value for the reservoir
c   prspvinj - the number of pore volumes injected
c   prspvprd - the number of pore volumes produced
c   prstotpvinj - the total number of pore volumes injected
c   prstotpvprd - the total number of pore volumes produced
c
     1,prstotpv,prspvinj(nprsw,nprsph),prspvprd(nprsw,nprsph)
     1,prstotpvinj(0:nprsph),prstotpvprd(0:nprsph)
c
c boundary condition parameters
c
c each of the following parameters will be set to one of the values
c
c   value =  0 => no flow condition on the given plane
c   value = -1 => fixed velocity condition on the given plane
c   value =  1 => fixed velocity condition on the given plane
c   value = -2 => time dependent velocity condition on the given
c   value =  2 => time dependent velocity condition on the given
c   value = -3 => fixed pressure condition on the given plane for which the
c                 code must pad the arrays and domain
c   value =  3 => fixed pressure condition on the given plane for which the
c                 code has already padded the arrays and domain
c   value = -4 => time dependent pressure condition on the given plane for
c                 which the code must pad the arrays and domain
c   value =  4 => time dependent pressure condition on the given plane for
c                 which the code has already padded the arrays and domain
c
c   note that the negative value will normally be set external to the simulator
c   and will be reset to a positive value after the padding is done (when
c   necessary).
c
c   iprsbndxyl - the flag for the type of boundary condition to impose on the
c                plane perpendicular to the third coordinate axis on the left
c                end of the domain
c   iprsbndxyr - the flag for the type of boundary condition to impose on the
c                plane perpendicular to the third coordinate axis on the right
c                end of the domain
c   iprsbndxzl - the flag for the type of boundary condition to impose on the
c                plane perpendicular to the second coordinate axis on the left
c                end of the domain
c   iprsbndxzr - the flag for the type of boundary condition to impose on the
c                plane perpendicular to the second coordinate axis on the right
c                end of the domain
c   iprsbndyzl - the flag for the type of boundary condition to impose on the
c                plane perpendicular to the first coordinate axis on the left
c                end of the domain
c   iprsbndyzr - the flag for the type of boundary condition to impose on the
c                plane perpendicular to the first coordinate axis on the right
c                end of the domain
c
     1,iprsbndxyl,iprsbndxyr
     1,iprsbndxzl,iprsbndxzr
     1,iprsbndyzl,iprsbndyzr
c
c restart parameter
c
c   iprsrstrt - a parameter to tell the code whether or not to restart the
c               simulator from the last output step
c   nprsrstrtim - the output step to go to before restarting. this is used to
c                 restart the simulator at some intermediate point in the run.
c   prsrstrt = restart string:
c              RESTART -> restart the simulation
c              RUN COMPLETE -> the simulation is complete and there is no need
c                              to run.
c              any thing else will let the simulation start from the beginning
c
     1,iprsrstrt,nprsrstrtim,prsrstrt
c
c set a debug parameter in here
c
c   iprsdebug - if the parameter is set to 1 the routine can put an if statement
c               in a good location for use in debugging
c
     1,iprsdebug
      character*12 prsrstrt
