import java.awt.*;
import java.util.Random;
import java.util.StringTokenizer;
import java.util.Vector;
import java.io.*;
public class EllipticRectMFEObject extends Object {

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *  Constructors for the class                                             *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  // full constructor
  public EllipticRectMFEObject(
           String ctype,             // coefficient type (e.g. SYMBOLIC)
           String [][] ac,           // symbolic coefficient operator
           double [] as,             // sample array for the coefficient
           String ftype,             // forcing function type (e.g., SYMBOLIC)
           String ff,                // symbolic forcing function
           double [] fs,             // sample array for the forcing function
           PointSources ps,          // point source object input for forcing
           String bftype,            // bndry value type (e.g., SYMBOLIC)
           String bct,               // bndry val type - e.g., flux
           String fbfxl,             // 1st component symbolic flux - left side
           String fbfxr,             // 1st component symbolic flux - right side
           String fbfyl,             // 2nd component symbolic flux - left side
           String fbfyr,             // 2nd component symbolic flux - right side
           String fbfzl,             // 3rd component symbolic flux - left side
           String fbfzr,             // 3rd component symbolic flux - right side
           String pbfxl,             // 1st component symbolic pres - left side
           String pbfxr,             // 1st component symbolic pres - right side
           String pbfyl,             // 2nd component symbolic pres - left side
           String pbfyr,             // 2nd component symbolic pres - right side
           String pbfzl,             // 3rd component symbolic pres - left side
           String pbfzr,             // 3rd component symbolic pres - right side
           double [][] fbcx,         // 1st component flux bc array
           double [][] fbcy,         // 2nd component flux bc array
           double [][] fbcz,         // 3rd component flux bc array
           double [][] pbcx,         // 1st component pres bc array
           double [][] pbcy,         // 2nd component pres bc array
           double [][] pbcz,         // 3rd component pres bc array
           int [][] mbcx,            // 1st component bndry value mask array
           int [][] mbcy,            // 2nd component bndry value mask array
           int [][] mbcz,            // 3rd component bndry value mask array
           double xa,                // 1st component left domain boundary
           double xb,                // 1st component right boundary
           double ya,                // 2nd component left boundary
           double yb,                // 2nd component right boundary
           double za,                // 3rd component left boundary
           double zb,                // 3rd component right boundary
           int nxv,                  // number of nodes in 1st component
           int nyv,                  // number of nodes in 2nd component
           int nzv,                  // number of nodes in 3rd component
           int nxr,                  // number of refinements in 1st component
           int nyr,                  // number of refinements in 2nd component
           int nzr,                  // number of refinements in 3rd component
           String grtypx,            // grid type parameter in 1st component
           String grtypy,            // grid type parameter in 2nd component
           String grtypz,            // grid type parameter in 3rd component
           double [] xn,             // explicit node values in 1st component
           double [] yn,             // explicit node values in 2nd component
           double [] zn,             // explicit node values in 3rd component
           String fnx,               // function to define grid in 1st component
           String fny,               // function to define grid in 2nd component
           String fnz,               // function to define grid in 3rd component
           double [] xg,             // explicit grid line values in x
           double [] yg,             // explicit grid line values in y
           double [] zg,             // explicit grid line values in z
           String fgx,               // function to define grid in x
           String fgy,               // function to define grid in y
           String fgz,               // function to define grid in z
           double tc,                // time dependent coeff
           double dt,                // time step to use
           int ord,                  // order of the method
           TextArea ta,              // text area for output
           boolean cp)               // compute solution flag
                                 throws WellPosedProblemException {

    // set the outputArea for messages
    outputArea = ta;

    outputMessage("\nBeginning MFE Solution of the Elliptic Problem.....\n",
                  false);

    // set up the coefficient in the elliptic operator
    outputMessage("Setting up coefficient in the elliptic operator.....");
    setUpCoefficient(ctype, ac, as, nxv, nyv, nzv);

    // set up the forcing function on the right hand side
    outputMessage("Setting up the forcing function.....");
    setUpForcingFunction(ftype, ff, fs, ps, nxv, nyv, nzv);

    // set up the forcing function on the right hand side
    outputMessage("Setting up the boundary conditions.....");
    setUpBoundaryConditions(bftype, bct,
                            fbfxl, fbfxr, fbfyl, fbfyr, fbfzl, fbfzr,
                            pbfxl, pbfxr, pbfyl, pbfyr, pbfzl, pbfzr,
                            fbcx, fbcy, fbcz,
                            pbcx, pbcy, pbcz,
                            mbcx, mbcy, mbcz,
                            nxv, nyv, nzv);

    // set up the grid
    outputMessage("Setting up the computational grid.....");
    setUpComputationalGrid(xa, xb, ya, yb, za, zb, nxv, nyv, nzv,
                           grtypx, grtypy, grtypz, xn, yn, zn,
                           fnx, fny, fnz, xg, yg, zg, fgx, fgy, fgz);

    // set the refinement in the grid
    outputMessage("Setting up the computational mesh refinement.....");
    setComputationalRefinement(nxr, nyr, nzr, nxv, nyv, nzv);

    // set the time stuff
    if(tc > 0.0) tcoeff = tc;
    if(dt > 0.0) tstep = dt;

    // set up the order
    outputMessage("Setting up the computational order of method.....");
    order = ord;
    if(order < 1) order = 1;
    if(order > 2) order = 2;

    // start into the process of setting up the problem
    // build the computational mesh
    outputMessage("EllipticRectMFEObject -> making the mesh");
    makeMesh();

    // compute the solution if desired
    if(cp) compute();

  }

  public void compute() throws WellPosedProblemException {

    // now the mesh has been built and the domain is known, we can test the
    // boundary conditions, forcing function and coefficient in the operator
    // to make sure they are well defined at the point.
    outputMessage("EllipticRectMFEObject -> testing functions");
    testFunctions();

    // initialize the variables in the problem - that is the unknown and the
    // gradient of the variable
    outputMessage("EllipticRectMFEObject -> initializing unknowns");
    initializeUnknowns();

    // initialize the matrix information for the solution of the system
    outputMessage("EllipticRectMFEObject -> initializing matrix storage");
    initializeMatrixStorage();

    // initialize the boundary values
    outputMessage("EllipticRectMFEObject -> initializing boundary values");
    setBoundaryValues();

    // build the basis functions for the scheme
    outputMessage("EllipticRectMFEObject -> building basis functions");
    buildBasis();

    // build the matrix for the discretization
    outputMessage("EllipticRectMFEObject -> building velocity matrices");
    buildVelMatrices();
    outputMessage("EllipticRectMFEObject -> factor velocity matrices");
    reduceVelMatrices();

    // build the time dependent piece of the parabolic operator
    outputMessage("EllipticRectMFEObject -> set compressibility term");
    buildDiagH();

    outputMessage("EllipticRectMFEObject -> building mass matrices");
    buildMassMatrix();
    outputMessage("EllipticRectMFEObject -> factor mass matrices");
    reduceMassMatrices();

    // set up coefficient values at the cell centers
    outputMessage("EllipticRectMFEObject -> building cell center coeffs");
    buildCellCenteredCoefficientValues();
    
    // build the right hand side from the pressure contribution
    outputMessage("EllipticRectMFEObject -> building right hand side");
    buildPRSRHS();

    // build the right hand side from the boundary conditions
    outputMessage("EllipticRectMFEObject -> building right hand side");
    buildFluxBoundaryConditionRHS();
    buildPRSBoundaryConditionRHS();

    // compute the solution
    outputMessage("EllipticRectMFEObject -> computing solution");
    computeSolution(ibith);

    // compute the solution
    outputMessage("EllipticRectMFEObject -> computing velocities");
    computeVelocity();

    // all done
    outputMessage("EllipticRectMFEObject " +
                             "-> done computing elliptic solution");

  }

  private void setOutputArea(TextArea ta) {
    outputArea = ta;
  }

  private void outputMessage(String s) {
    outputMessage(s,true);
  }

  private void outputMessage(String s, boolean append) {
    if(outputArea == null) {
      System.out.println(s);
      return;
    }
    if(append) {
      outputArea.append(s);
    } else {
      outputArea.setText(s);
    }
    outputArea.append("\n");
    outputArea.setCaretPosition(10000000);
  }

  private void setUpCoefficient(String ctype, String [][] ac, double [] as,
                                int nxv, int nyv, int nzv)
             throws WellPosedProblemException { 
    // check the coefficient
    coefType = ctype;
    if(ctype.equals(SYMBOLIC_FUNCTION_DEFINITION)) {
      a = ac;
      if(ac == null) a = defaultCoeff;
    } else {
      // check the size of the sample
      int nnp = nxv * nyv * nzv;
      int nxny = nxv * nyv;
      if(as.length != (9*nnp)) {
        String s = "The coefficient sample array does not match dimensions";
        throw new WellPosedProblemException(s);
      }
      // pull off the tensor
      int shft = 0;
      asampxx = new double[nnp];
      for(int i=0; i<nnp; i++) asampxx[i] = as[i+shft];
      shft = shft + nnp;
      asampxy = new double[nnp];
      for(int i=0; i<nnp; i++) asampxy[i] = as[i+shft];
      shft = shft + nnp;
      asampxz = new double[nnp];
      for(int i=0; i<nnp; i++) asampxz[i] = as[i+shft];
      shft = shft + nnp;
      asampyx = new double[nnp];
      for(int i=0; i<nnp; i++) asampyx[i] = as[i+shft];
      shft = shft + nnp;
      asampyy = new double[nnp];
      for(int i=0; i<nnp; i++) asampyy[i] = as[i+shft];
      shft = shft + nnp;
      asampyz = new double[nnp];
      for(int i=0; i<nnp; i++) asampyz[i] = as[i+shft];
      shft = shft + nnp;
      asampzx = new double[nnp];
      for(int i=0; i<nnp; i++) asampzx[i] = as[i+shft];
      shft = shft + nnp;
      asampzy = new double[nnp];
      for(int i=0; i<nnp; i++) asampzy[i] = as[i+shft];
      shft = shft + nnp;
      asampzz = new double[nnp];
      for(int i=0; i<nnp; i++) asampzz[i] = as[i+shft];
    }

  }

  private void setUpForcingFunction(String ftype, String ff, double [] fs,
                        PointSources ps, int nxv, int nyv, int nzv)
             throws WellPosedProblemException { 
    // check the forcing function
    forcingType = ftype;
    if(ftype.equals(SYMBOLIC_FUNCTION_DEFINITION)) {
      // check the forcing function
      f = ff;
      if(ff == null) f = "0.0";
    } else if(ftype.equals(POINT_SOURCE_FUNCTION_DEFINITION)) {
      pointSources = ps;
      if(pointSources == null) pointSources = new PointSources();
    } else {
      // check the size of the sample
      int nnp =  nxv * nyv * nzv;
      int nxny =  nxv * nyv;
      if(fs == null) {
        String s = "The coefficient sample array is null";
        throw new WellPosedProblemException(s);
      }
      if(fs.length != (nnp)) {
        String s = "The coefficient sample array does not match dimensions";
        throw new WellPosedProblemException(s);
      }
      // pull off the forcing function values
      fsamp = new double[nnp];
      for(int i=0; i<nnp; i++) fsamp[i] = fs[i];
    }
  }

  private void setUpBoundaryConditions(String bftype, String bct,
                        String fbfxl, String fbfxr,
                        String fbfyl, String fbfyr,
                        String fbfzl, String fbfzr,
                        String pbfxl, String pbfxr,
                        String pbfyl, String pbfyr,
                        String pbfzl, String pbfzr,
                        double [][] fbcx, double [][] fbcy, double [][] fbcz,
                        double [][] pbcx, double [][] pbcy, double [][] pbcz,
                        int [][] mbcx, int [][] mbcy, int [][] mbcz,
                        int nxv, int nyv, int nzv)
             throws WellPosedProblemException { 
    boolean test = true;
    int nxny = nxv * nyv;
    int nxnz = nxv * nzv;
    int nynz = nyv * nzv;

    // set the boundary evaluation type
    boundaryEvaluationType = bftype;
    test = false;
    if(boundaryEvaluationType.equals(SYMBOLIC_FUNCTION_DEFINITION)) test = true;
    if(boundaryEvaluationType.equals(NUMERIC_FUNCTION_DEFINITION)) test = true;
    if(!test) {
      String s = "The boundary condition evaluation is not symbolic or numeric";
      throw new WellPosedProblemException(s);
    }

    // check the boundary condition information
    boundaryConditionType = bct;
    if(boundaryConditionType.equals(NEUMANN_BOUNDARY)) test = true;
    if(boundaryConditionType.equals(DIRICHLET_BOUNDARY)) test = true;
    if(boundaryConditionType.equals(MIXED_BOUNDARY)) test = true;
    if(!test) {
      String s = "The boundary condition is not Dirichlet, Neumann, or Mixed";
      throw new WellPosedProblemException(s);
    }

    // test for neumann conditions
    if(boundaryConditionType.equals(NEUMANN_BOUNDARY) ||
                    boundaryConditionType.equals(MIXED_BOUNDARY) ) {
      if(boundaryEvaluationType.equals(SYMBOLIC_FUNCTION_DEFINITION)) {
        // check the flux function
        fluxBoundaryFunctionXL = fbfxl;
        if(fluxBoundaryFunctionXL == null) {
          String s = "The flux boundary condition function is null";
          throw new WellPosedProblemException(s);
        }
        fluxBoundaryFunctionXR = fbfxr;
        if(fluxBoundaryFunctionXR == null) {
          String s = "The flux boundary condition function is null";
          throw new WellPosedProblemException(s);
        }
        fluxBoundaryFunctionYL = fbfyl;
        if(fluxBoundaryFunctionYL == null) {
          String s = "The flux boundary condition function is null";
          throw new WellPosedProblemException(s);
        }
        fluxBoundaryFunctionYR = fbfyr;
        if(fluxBoundaryFunctionYL == null) {
          String s = "The flux boundary condition function is null";
          throw new WellPosedProblemException(s);
        }
        fluxBoundaryFunctionZL = fbfzl;
        if(fluxBoundaryFunctionZL == null) {
          String s = "The flux boundary condition function is null";
          throw new WellPosedProblemException(s);
        }
        fluxBoundaryFunctionZR = fbfzr;
        if(fluxBoundaryFunctionZL == null) {
          String s = "The flux boundary condition function is null";
          throw new WellPosedProblemException(s);
        }
      } else if (boundaryEvaluationType.equals(NUMERIC_FUNCTION_DEFINITION)) {

        // see if the boundary condition arrays are well defined
        if(fbcx == null) {
          String s = "The x-flux boundary array is null";
          throw new WellPosedProblemException(s);
        }
        if(fbcy == null) {
          String s = "The y-flux boundary array is null";
          throw new WellPosedProblemException(s);
        }
        if(fbcz == null) {
          String s = "The z-flux boundary array is null";
          throw new WellPosedProblemException(s);
        }

        // test the two dimensional deal
        if(fbcx.length != 2) {
          String s = "The x-flux boundary array must have two parts";
          throw new WellPosedProblemException(s);
        }
        if(fbcy.length != 2) {
          String s = "The y-flux boundary array must have two parts";
          throw new WellPosedProblemException(s);
        }
        if(fbcz.length != 2) {
          String s = "The z-flux boundary array must have two parts";
          throw new WellPosedProblemException(s);
        }

        // check the size of the sample
        if(fbcx[0].length != nynz) {
          String s = "x-boundary array does not match mesh dimensions";
          throw new WellPosedProblemException(s);
        }
        if(fbcy[0].length != nxnz) {
          String s = "y-boundary array does not match mesh dimensions";
          throw new WellPosedProblemException(s);
        }
        if(fbcz[0].length != nxny) {
          String s = "z-boundary array does not match mesh dimensions";
          throw new WellPosedProblemException(s);
        }

        // pull off the boundary condition information
        xBoundaryFlux = new double[2][nynz];
        yBoundaryFlux = new double[2][nxnz];
        zBoundaryFlux = new double[2][nxny];
        // x-left
        for(int i=0; i<nynz; i++) xBoundaryFlux[0][i] = fbcx[0][i];
        // x-right
        for(int i=0; i<nynz; i++) xBoundaryFlux[1][i] = fbcx[1][i];
        // y-left
        for(int j=0; j<nxnz; j++) yBoundaryFlux[0][j] = fbcy[0][j];
        // y-right
        for(int j=0; j<nxnz; j++) yBoundaryFlux[1][j] = fbcy[1][j];
        // z-left
        for(int k=0; k<nxny; k++) zBoundaryFlux[0][k] = fbcz[0][k];
        // z-right
        for(int k=0; k<nxny; k++) zBoundaryFlux[1][k] = fbcz[1][k];

      }
      return;
    }

    // test for dirichlet or mixed conditions
    if(boundaryConditionType.equals(DIRICHLET_BOUNDARY) ||
                    boundaryConditionType.equals(MIXED_BOUNDARY) ) {

      if(boundaryEvaluationType.equals(SYMBOLIC_FUNCTION_DEFINITION)) {

        prsBoundaryFunctionXL = pbfxl;
        if(prsBoundaryFunctionXL == null) {
          String s = "The dirichlet boundary condition function is null";
          throw new WellPosedProblemException(s);
        }
        prsBoundaryFunctionXR = pbfxr;
        if(prsBoundaryFunctionXR == null) {
          String s = "The dirichlet boundary condition function is null";
          throw new WellPosedProblemException(s);
        }
        prsBoundaryFunctionYL = pbfyl;
        if(prsBoundaryFunctionYL == null) {
          String s = "The dirichlet boundary condition function is null";
          throw new WellPosedProblemException(s);
        }
        prsBoundaryFunctionYR = pbfyr;
        if(prsBoundaryFunctionYR == null) {
          String s = "The dirichlet boundary condition function is null";
          throw new WellPosedProblemException(s);
        }
        prsBoundaryFunctionZL = pbfzl;
        if(prsBoundaryFunctionZL == null) {
          String s = "The dirichlet boundary condition function is null";
          throw new WellPosedProblemException(s);
        }
        prsBoundaryFunctionZR = pbfzr;
        if(prsBoundaryFunctionZR == null) {
          String s = "The dirichlet boundary condition function is null";
          throw new WellPosedProblemException(s);
        }

      } else if (boundaryEvaluationType.equals(NUMERIC_FUNCTION_DEFINITION)) {

        // see if the boundary condition arrays are well defined
        if(pbcx == null) {
          String s = "The x-pressure boundary array is null";
          throw new WellPosedProblemException(s);
        }
        if(pbcy == null) {
          String s = "The y-pressure boundary array is null";
          throw new WellPosedProblemException(s);
        }
        if(pbcz == null) {
          String s = "The z-pressure boundary array is null";
          throw new WellPosedProblemException(s);
        }

        // test the two dimensional deal
        if(pbcx.length != 2) {
          String s = "The x-pressure boundary array must have two parts";
          throw new WellPosedProblemException(s);
        }
        if(pbcy.length != 2) {
          String s = "The y-pressure boundary array must have two parts";
          throw new WellPosedProblemException(s);
        }
        if(pbcz.length != 2) {
          String s = "The z-pressure boundary array must have two parts";
          throw new WellPosedProblemException(s);
        }

        // check the size of the sample
        if(pbcx[0].length != nynz) {
          String s = "x-boundary array does not match mesh dimensions";
          throw new WellPosedProblemException(s);
        }
        if(pbcy[0].length != nxnz) {
          String s = "y-boundary array does not match mesh dimensions";
          throw new WellPosedProblemException(s);
        }
        if(pbcz[0].length != nxny) {
          String s = "z-boundary array does not match mesh dimensions";
          throw new WellPosedProblemException(s);
        }

        // pull off the boundary condition information
        xBoundaryPrs = new double[2][nynz];
        yBoundaryPrs = new double[2][nxnz];
        zBoundaryPrs = new double[2][nxny];
        // x-left
        for(int i=0; i<nynz; i++) xBoundaryPrs[0][i] = pbcx[0][i];
        // x-right
        for(int i=0; i<nynz; i++) xBoundaryPrs[1][i] = pbcx[1][i];
        // y-left
        for(int j=0; j<nxnz; j++) yBoundaryPrs[0][j] = pbcy[0][j];
        // y-right
        for(int j=0; j<nxnz; j++) yBoundaryPrs[1][j] = pbcy[1][j];
        // z-left
        for(int k=0; k<nxny; k++) zBoundaryPrs[0][k] = pbcz[0][k];
        // z-right
        for(int k=0; k<nxny; k++) zBoundaryPrs[1][k] = pbcz[1][k];
      }

    }

    // if not mixed, jump out of the routine
    if(boundaryConditionType.equals(DIRICHLET_BOUNDARY)) {
      for(int i=0; i<nynz; i++) xBoundaryMask[i][0] = 1;
      for(int i=0; i<nynz; i++) xBoundaryMask[i][1] = 1;
      for(int j=0; j<nxnz; j++) yBoundaryMask[j][0] = 1;
      for(int j=0; j<nxnz; j++) yBoundaryMask[j][1] = 1;
      for(int k=0; k<nxny; k++) zBoundaryMask[k][0] = 1;
      for(int k=0; k<nxny; k++) zBoundaryMask[k][1] = 1;
      return;
    }
    if(boundaryConditionType.equals(NEUMANN_BOUNDARY)) {
      for(int i=0; i<nynz; i++) xBoundaryMask[i][0] = 0;
      for(int i=0; i<nynz; i++) xBoundaryMask[i][1] = 0;
      for(int j=0; j<nxnz; j++) yBoundaryMask[j][0] = 0;
      for(int j=0; j<nxnz; j++) yBoundaryMask[j][1] = 0;
      for(int k=0; k<nxny; k++) zBoundaryMask[k][0] = 0;
      for(int k=0; k<nxny; k++) zBoundaryMask[k][1] = 0;
      return;
    }

    // if the boundary conditions are mixed, then the following defines the
    // mask for the boundary condition
    if(boundaryConditionType.equals(MIXED_BOUNDARY)) {

      // see if the boundary condition arrays are well defined
      if(mbcx == null) {
        String s = "The x boundary mask array is null";
        throw new WellPosedProblemException(s);
      }
      if(mbcy == null) {
        String s = "The y boundary mask array is null";
        throw new WellPosedProblemException(s);
      }
      if(mbcz == null) {
        String s = "The z boundary mask array is null";
        throw new WellPosedProblemException(s);
      }

      // test the two dimensional deal
      if(mbcx[0].length != 2) {
        String s = "The x boundary mask array must have two parts";
        throw new WellPosedProblemException(s);
      }
      if(mbcy[0].length != 2) {
        String s = "The y boundary mask array must have two parts";
        throw new WellPosedProblemException(s);
      }
      if(mbcz[0].length != 2) {
        String s = "The z.boundary mask array must have two parts";
        throw new WellPosedProblemException(s);
      }

      // check the size of the sample
      if(mbcx.length != nynz) {
        String s = "x-boundary mask array does not match mesh dimensions";
        throw new WellPosedProblemException(s);
      }
      if(mbcy.length != nxnz) {
        String s = "y-boundary mask array does not match mesh dimensions";
        throw new WellPosedProblemException(s);
      }
      if(mbcz.length != nxny) {
        String s = "z-boundary mask array does not match mesh dimensions";
        throw new WellPosedProblemException(s);
      }

      // pull off the boundary condition information
      xBoundaryMask = new int[nynz][2];
      yBoundaryMask = new int[nxnz][2];
      yBoundaryMask = new int[nxny][2];
      // x-left
      for(int i=0; i<nynz; i++) xBoundaryMask[i][0] = mbcx[i][0];
      // x-right
      for(int i=0; i<nynz; i++) xBoundaryMask[i][1] = mbcx[i][1];
      // y-left
      for(int j=0; j<nxnz; j++) yBoundaryMask[j][0] = mbcy[j][0];
      // y-right
      for(int j=0; j<nxnz; j++) yBoundaryMask[j][1] = mbcy[j][1];
      // z-left
      for(int k=0; k<nxny; k++) zBoundaryMask[k][0] = mbcz[k][k];
      // z-right
      for(int k=0; k<nxny; k++) zBoundaryMask[k][1] = mbcz[k][k];

    } else { 
      // if the code gets to here there is a problem
      String s
        = "The boundary condition evaluation is not set to numeric or symbolic";
      throw new WellPosedProblemException(s);
    }

  }

  public void setUpComputationalGrid(double xa, double xb, double ya,
                               double yb, double za, double zb,
                               int nxv, int nyv, int nzv,
                               String grtypx, String grtypy, String grtypz,
                               double [] xn, double [] yn, double [] zn,
                               String fnx, String fny, String fnz,
                               double [] xg, double [] yg, double [] zg,
                               String fgx, String fgy, String fgz) {

    // check the type of mesh to create - defaults to uniform
    gridtypex = 0;
    if(grtypx.equals("INPUT ARRAY NODE")) {
      gridtypex = 1;
      // check node array to see if it exists
      if(xn == null) {
        gridtypex = 0;
      } else {
        x = sort(xn);
        nx = x.length;
      }
    }
    if(grtypx.equals("FUNCTION NODE")) { 
      gridtypex = 2;
      if(fnx == null) {
        gridtypex = 0;
      } else {
        fnodex = fnx;
        nx = nxv;
        x = new double[nx];
        double delx = 1.0 / ((double) nx);
        for(int i=0; i<nx; i++) {
          double xval = (((double) i) + 0.5 ) * delx;
          x[i] = getValueFromString(fnodex, xval, "x");
        }
        xl = getValueFromString(fnodex, 0.0, "x");
        xr = getValueFromString(fnodex, 1.0, "x");
      }
    }
    if(grtypx.equals("INPUT ARRAY GRID")) {
      gridtypex = 3;
      // check node array to see if it exists
      if(xg == null) {
        gridtypex = 0;
      } else {
        if(xg.length <2) {
          gridtypex = 0;
        } else {
          xgr = sort(xg);
          nx = xgr.length - 1;
        }
      }
    }
    if(grtypx.equals("FUNCTION GRID")) { 
      gridtypex = 4;
      if(fgx == null) {
        gridtypex = 0;
      } else {
        fgridx = fgx;
        nx = nxv;
        xgr = new double[nx+1];
        double delx = 1.0 / ((double) nx);
        for(int i=1; i<nx; i++) {
          double xval = ((double) i) * delx;
          xgr[i] = getValueFromString(fgridx, xval, "x");
        }
        xl = getValueFromString(fgridx, 0.0, "x");
        xr = getValueFromString(fgridx, 1.0, "x");
        xgr[0] = xl;
        xgr[nx] = xr;
      }
    }

    // check the type of mesh to create - defaults to uniform
    gridtypey = 0;
    if(grtypy.equals("INPUT ARRAY NODE")) {
      gridtypey = 1;
      // check node array to see if it exists
      if(yn == null) {
        gridtypey = 0;
      } else {
        y = sort(yn);
        ny = y.length;
      }
    }
    if(grtypy.equals("FUNCTION NODE")) { 
      gridtypey = 2;
      if(fny == null) {
        gridtypey = 0;
      } else {
        fnodey = fny;
        ny = nyv;
        y = new double[ny];
        double dely = 1.0 / ((double) ny);
        for(int j=0; j<ny; j++) {
          double yval = (((double) j) + 0.5 ) * dely;
          y[j] = getValueFromString(fnodey, yval, "y");
        }
        yl = getValueFromString(fnodey, 0.0, "y");
        yr = getValueFromString(fnodey, 1.0, "y");
      }
    }
    if(grtypy.equals("INPUT ARRAY GRID")) {
      gridtypey = 3;
      // check node array to see if it exists
      if(yg == null) {
        gridtypey = 0;
      } else {
        if(yg.length <2) {
          gridtypey = 0;
        } else {
          ygr = sort(yg);
          ny = ygr.length - 1;
        }
      }
    }
    if(grtypy.equals("FUNCTION GRID")) { 
      gridtypey = 4;
      if(fgy == null) {
        gridtypey = 0;
      } else {
        fgridy = fgy;
        ny = nyv;
        ygr = new double[ny+1];
        double dely = 1.0 / ((double) ny);
        for(int j=1; j<ny; j++) {
          double yval = ((double) j) * dely;
          ygr[j] = getValueFromString(fgridy, yval, "y");
        }
        yl = getValueFromString(fgridy, 0.0, "y");
        yr = getValueFromString(fgridy, 1.0, "y");
        ygr[0] = yl;
        ygr[ny] = yr;
      }
    }

    // check the type of mesh to create - defaults to uniform
    gridtypez = 0;
    if(grtypz.equals("INPUT ARRAY NODE")) {
      gridtypez = 1;
      // check node array to see if it exists
      if(zn == null) {
        gridtypez = 0;
      } else {
        z = sort(zn);
        nz = z.length;
      }
    }
    if(grtypz.equals("FUNCTION NODE")) { 
      gridtypez = 2;
      if(fnz == null) {
        gridtypez = 0;
      } else {
        fnodez = fnz;
        nz = nzv;
        z = new double[nz];
        double delz = 1.0 / ((double) nz);
        for(int k=0; k<nz; k++) {
          double zval = (((double) k) + 0.5 ) * delz;
          z[k] = getValueFromString(fnodez, zval, "z");
        }
        zl = getValueFromString(fnodez, 0.0, "z");
        zr = getValueFromString(fnodez, 1.0, "z");
      }
    }
    if(grtypz.equals("INPUT ARRAY GRID")) {
      gridtypez = 3;
      // check node array to see if it exists
      if(zg == null) {
        gridtypez = 0;
      } else {
        if(zg.length <2) {
          gridtypez = 0;
        } else {
          zgr = sort(zg);
          nz = zgr.length - 1;
        }
      }
    }
    if(grtypz.equals("FUNCTION GRID")) { 
      gridtypez = 4;
      if(fgz == null) {
        gridtypez = 0;
      } else {
        fgridz = fgz;
        nz = nzv;
        zgr = new double[nz+1];
        double delz = 1.0 / ((double) nz);
        for(int k=1; k<nz; k++) {
          double zval = ((double) k) * delz;
          zgr[k] = getValueFromString(fgridz, zval, "z");
        }
        zl = getValueFromString(fgridz, 0.0, "z");
        zr = getValueFromString(fgridz, 1.0, "z");
        zgr[0] = zl;
        zgr[nz] = zr;
      }
    }

    // define the boundary of the region
    if(gridtypex == 0) {
      // make sure the domain is set the correct way
      nx = nxv;
      if(xl < xr) {
        xl = xa;
        xr = xb;
      } else {
        xl = xb;
        xr = xa;
      }
    }
    if(gridtypey == 0) {
      ny = nyv;
      if(yl < yr) {
        yl = ya;
        yr = yb;
      } else {
        yl = yb;
        yr = ya;
      }
    }
    if(gridtypez == 0) {
      nz = nzv;
      if(zl < zr) {
        zl = za;
        zr = zb;
      } else {
        zl = zb;
        zr = za;
      }
    }
    if(gridtypex == 1) {
      if(nx > 1) {
        xl = 0.5 * ( 3.0 * x[0] - x[1] );
        xr = 0.5 * ( 3.0 * x[nxm1] - x[nx-2] );
      } else {
        xr = 2.0 * x[0];
        xl = 2.0 * x[0] - xr;
      }
    }
    if(gridtypey == 1) {
      if(ny > 1) {
        yl = 0.5 * ( 3.0 * y[0] - y[1] );
        yr = 0.5 * ( 3.0 * y[nym1] - y[ny-2] );
      } else {
        yr = 2.0 * y[0];
        yl = 2.0 * y[0] - yr;
      }
    }
    if(gridtypez == 1) {
      if(nz > 1) {
        zl = 0.5 * ( 3.0 * z[0] - z[1] );
        zr = 0.5 * ( 3.0 * z[nzm1] - z[nz-2] );
      } else {
        zr = 2.0 * z[0];
        zl = 2.0 * z[0] - zr;
      }
    }
    if(gridtypex == 3) {
      xl = xgr[0];
      xr = xgr[nxm1];
    }
    if(gridtypey == 3) {
      yl = ygr[0];
      yr = ygr[nym1];
    }
    if(gridtypez == 3) {
      zl = zgr[0];
      zr = zgr[nzm1];
    }
    if(gridtypex == 4) {
    }
    if(gridtypey == 4) {
    }
    if(gridtypez == 4) {
    }

    // compute some mesh parameters
    nxp1 = nx + 1;
    nxm1 = nx - 1;
    nyp1 = ny + 1;
    nym1 = ny - 1;
    nzp1 = nz + 1;
    nzm1 = nz - 1;

  }

  private void setComputationalRefinement(int nxr, int nyr, int nzr,
                                          int nxv, int nyv, int nzv) {
    // make sure each of the refinement values is at least 1
    if(nxr > 0) {
      nxref = nxr;
    } else {
      nxref = 1;
    }
    if(nyr > 0) {
      nyref = nyr;
    } else {
      nyref = 1;
    }
    if(nzr > 0) {
      nzref = nzr;
    } else {
      nzref = 1;
    }

    // test for a refinement
    if((nxref*nyref*nzref) > 1) {
      refineCoefficient(nxv, nyv, nzv);
      refineForcingFunction(nxv, nyv, nzv);
      refineBoundaryConditions(nxv, nyv, nzv);
      refineGrid(nxv, nyv, nzv);
    }

  }

  public void refineCoefficient(int nxv, int nyv, int nzv) {

    // if the coefficient is defined symbolically, we don't need to do this
    if(coefType.equals(SYMBOLIC_FUNCTION_DEFINITION)) return;

    int nxnew = nxref * nxv;
    int nynew = nyref * nyv;
    int nznew = nzref * nzv;
    int nnpnew = nxnew * nynew * nznew;
    int nxnynew = nxref * nxv * nyref * nyv;

    // reset the sample values
    double [] asampxxnew = new double[nnpnew];
    double [] asampxynew = new double[nnpnew];
    double [] asampxznew = new double[nnpnew];
    double [] asampyxnew = new double[nnpnew];
    double [] asampyynew = new double[nnpnew];
    double [] asampyznew = new double[nnpnew];
    double [] asampzxnew = new double[nnpnew];
    double [] asampzynew = new double[nnpnew];
    double [] asampzznew = new double[nnpnew];

    // redefine the variable
    for(int k=0; k<nzv; k++) {
      int k2 = k * nxny;
      for(int kk=0; kk<nzref; kk++) {
        int kk2 = kk * nxnynew + k * nzref * nxnynew;
        for(int j=0; j<nyv; j++) {
          int j2 = j * nx + k2;
          for(int jj=0; jj<nyref; jj++) {
            int jj2 = jj * nxnew + j * nyref * nxnew + kk2;
            for(int i=0; i<nxv; i++) {
              int ind = i + j2;
              for(int ii=0; ii<nxref; ii++) {
                int iind = ii + i * nxref + jj2;
                asampxxnew[iind] = asampxx[ind];
                asampxynew[iind] = asampxy[ind];
                asampxznew[iind] = asampxz[ind];
                asampyxnew[iind] = asampyx[ind];
                asampyynew[iind] = asampyy[ind];
                asampyznew[iind] = asampyz[ind];
                asampzxnew[iind] = asampzx[ind];
                asampzynew[iind] = asampzy[ind];
                asampzznew[iind] = asampzz[ind];
              }
            }
          }
        }
      }
    }
    asampxx = asampxxnew;
    asampxy = asampxynew;
    asampxz = asampxznew;
    asampyx = asampyxnew;
    asampyy = asampyynew;
    asampyz = asampyznew;
    asampzx = asampzxnew;
    asampzy = asampzynew;
    asampzz = asampzznew;

  }

  public void refineForcingFunction(int nxv, int nyv, int nzv) {

    // if the forcing function is defined symbolically, we don't need to do this
    if(forcingType.equals(SYMBOLIC_FUNCTION_DEFINITION)) return;
    if(forcingType.equals(POINT_SOURCE_FUNCTION_DEFINITION)) return;

    int nxnew = nxref * nxv;
    int nynew = nyref * nyv;
    int nznew = nzref * nzv;
    int nnpnew = nxnew * nynew * nznew;
    int nxnynew = nxref * nxv * nyref * nyv;

    // reset the sample values
    double [] fsampnew = new double[nnpnew];

    // redefine the variable
    for(int k=0; k<nzv; k++) {
      int k2 = k * nxny;
      for(int kk=0; kk<nzref; kk++) {
        int kk2 = kk * nxnynew + k * nzref * nxnynew;
        for(int j=0; j<nyv; j++) {
          int j2 = j * nx + k2;
          for(int jj=0; jj<nyref; jj++) {
            int jj2 = jj * nxnew + j * nyref * nxnew + kk2;
            for(int i=0; i<nxv; i++) {
              int ind = i + j2;
              for(int ii=0; ii<nxref; ii++) {
                int iind = ii + i * nxref + jj2;
                fsampnew[iind] = fsamp[ind];
              }
            }
          }
        }
      }
    }
    fsamp = fsampnew;

  }

  public void refineBoundaryConditions(int nxv, int nyv, int nzv) {

    // if the boundary conditions are defined symbolically, we don't need to do
    // this
    if(boundaryEvaluationType.equals(SYMBOLIC_FUNCTION_DEFINITION)) return;

    int nxnew = nxref * nxv;
    int nynew = nyref * nyv;
    int nznew = nzref * nzv;
    int nnpnew = nxnew * nynew * nznew;
    int nxny = nxv * nyv;
    int nxnynew = nxref * nxv * nyref * nyv;
    int nynz = nyv * nzv;
    int nynznew = nyref * nyv * nzref * nzv;
    int nxnz = nzv * nxv;
    int nxnznew = nzref * nzv * nxref * nxv;

    // reset the sample values
    double [][] xBoundaryFluxnew = new double[2][nynznew];
    double [][] yBoundaryFluxnew = new double[2][nxnznew];
    double [][] zBoundaryFluxnew = new double[2][nxnynew];

    // redefine the variable
    for(int k=0; k<nzv; k++) {
      int k2 = k * nxny;
      for(int kk=0; kk<nzref; kk++) {
        int kk2 = kk * nxnynew;
        for(int j=0; j<nyv; j++) {
          int j2 = j * nx + k2;
          for(int jj=0; jj<nyref; jj++) {
            int jj2 = jj * nxnew + kk2;
                xBoundaryFluxnew[0][jj2] = xBoundaryFlux[0][j2];
                xBoundaryFluxnew[1][jj2] = xBoundaryFlux[1][j2];
          }
        }
      }
    }
    for(int i=0; i<nxv; i++) {
      int i2 = i * nynz;
      for(int ii=0; ii<nxref; ii++) {
        int ii2 = ii * nynznew;
        for(int k=0; k<nzv; k++) {
          int k2 = k * ny + i2;
          for(int kk=0; kk<nzref; kk++) {
            int kk2 = kk * nynew + ii2;
                yBoundaryFluxnew[0][kk2] = yBoundaryFlux[0][k2];
                yBoundaryFluxnew[1][kk2] = yBoundaryFlux[1][k2];
          }
        }
      }
    }
    for(int j=0; j<nyv; j++) {
      int j2 = j * nxnz;
      for(int jj=0; jj<nyref; jj++) {
        int jj2 = jj * nxnznew;
        for(int i=0; i<nxv; i++) {
          int i2 = i * nz + j2;
          for(int ii=0; ii<nxref; ii++) {
            int ii2 = ii * nznew + jj2;
                zBoundaryFluxnew[0][ii2] = zBoundaryFlux[0][i2];
                zBoundaryFluxnew[1][ii2] = zBoundaryFlux[1][i2];
          }
        }
      }
    }
    xBoundaryFlux = xBoundaryFluxnew;
    yBoundaryFlux = yBoundaryFluxnew;
    zBoundaryFlux = zBoundaryFluxnew;

  }

  public void refineGrid(int nxv, int nyv, int nzv) {

    if(gridtypex == 1) {
      int nxnew = nxref * nxv;
      double [] xnnew = new double[nxnew];
      double [] xtmp = new double[nxv+2];
      xtmp[0] = x[0] - ( x[1] - x[0] );
      for(int i=1; i<=nxv; i++) xtmp[i] = x[i-1];
      xtmp[nxv+1] = x[nxv-1] + ( x[nxv-1] - x[nxv-2] );
      for(int i=1; i<=nxv; i++) {
        double xl = 0.5 * ( xtmp[i] + xtmp[i-1] );
        double xr = 0.5 * ( xtmp[i+1] + xtmp[i] );
        double dx = ( xr - xl ) / ((double) nxref);
        for(int ii=0; ii<nxref; ii++) {
          xnnew[ii] = xl + ( ((double) ii) + 0.5 ) * dx;
        }
      }
      x = xnnew;
      nx = x.length;
    }

    if(gridtypex == 3) {
      int nxnew = nxref * nxv;
      double [] xgrnew = new double[nxnew+1];
      int cnt = 0;
      xgrnew[cnt] = xgr[0];
      for(int i=0; i<nxv; i++) {
        double dx = ( xgr[i+1] - xgr[i] ) / ((double) nxref);
        for(int ii=1; ii<=nxref; ii++) {
          cnt++;
          xgrnew[cnt] = xgr[i] + ((double) ii) * dx;
        }
      }
      xgr = xgrnew;
      nx = xgr.length-1;
    }

    if(gridtypey == 1) {
      int nynew = nyref * nyv;
      double [] ynnew = new double[nynew];
      double [] ytmp = new double[nyv+2];
      ytmp[0] = y[0] - ( y[1] - y[0] );
      for(int i=1; i<=nyv; i++) ytmp[i] = y[i-1];
      ytmp[nyv+1] = y[nyv-1] + ( y[nyv-1] - y[nyv-2] );
      for(int j=1; j<=nyv; j++) {
        double yl = 0.5 * ( ytmp[j] + ytmp[j-1] );
        double yr = 0.5 * ( ytmp[j+1] + ytmp[j] );
        double dy = ( yr - yl ) / ((double) nyref);
        for(int jj=0; jj<nyref; jj++) {
          ynnew[jj] = yl + ( ((double) jj) + 0.5 ) * dy;
        }
      }
      y = ynnew;
      ny = y.length;
    }

    if(gridtypey == 3) {
      int nynew = nyref * nyv;
      double [] ygrnew = new double[nynew+1];
      int cnt = 0;
      ygrnew[0] = ygr[0];
      for(int j=0; j<nyv; j++) {
        double dy = ( ygr[j+1] - ygr[j] ) / ((double) nyref);
        for(int jj=1; jj<=nyref; jj++) {
          cnt++;
          ygrnew[cnt] = ygr[j] + ((double) jj) * dy;
        }
      }
      ygr = ygrnew;
      ny = ygr.length-1;
    }

    if(gridtypez == 1) {
      int nznew = nzref * nzv;
      double [] znnew = new double[nznew];
      double [] ztmp = new double[nzv+2];
      ztmp[0] = z[0] - ( z[1] - z[0] );
      for(int i=1; i<=nzv; i++) ztmp[i] = z[i-1];
      ztmp[nzv+1] = z[nzv-1] + ( z[nzv-1] - z[nzv-2] );
      for(int k=1; k<=nzv; k++) {
        double zl = 0.5 * ( ztmp[k] + ztmp[k-1] );
        double zr = 0.5 * ( ztmp[k+1] + ztmp[k] );
        double dz = ( zr - zl ) / ((double) nzref);
        for(int kk=0; kk<nzref; kk++) {
          znnew[kk] = zl + ( ((double) kk) + 0.5 ) * dz;
        }
      }
      z = znnew;
      nz = z.length;
    }

    if(gridtypez == 3) {
      int nznew = nzref * nzv;
      double [] zgrnew = new double[nznew+1];
      int cnt = 0;
      zgrnew[0] = zgr[0];
      for(int k=0; k<nzv; k++) {
        double dz = ( zgr[k+1] - zgr[k] ) / ((double) nzref);
        for(int kk=1; kk<=nzref; kk++) {
          cnt++;
          zgrnew[cnt] = zgr[k] + ((double) kk) * dz;
        }
      }
      zgr = zgrnew;
      nz = zgr.length-1;
    }

    // compute some mesh parameters
    nx = nxref * nx;
    ny = nyref * ny;
    nz = nzref * nz;
    nxp1 = nx + 1;
    nxm1 = nx - 1;
    nyp1 = ny + 1;
    nym1 = ny - 1;
    nzp1 = nz + 1;
    nzm1 = nz - 1;

  }

  public void makeMesh() throws WellPosedProblemException {

    if(gridtypex < 0 || gridtypex > 4) throw
          new WellPosedProblemException("X Grid Is Not Defined In makeMesh()");
    if(gridtypey < 0 || gridtypey > 4) throw
          new WellPosedProblemException("Y Grid Is Not Defined In makeMesh()");
    if(gridtypez < 0 || gridtypez > 4) throw
          new WellPosedProblemException("Z Grid Is Not Defined In makeMesh()");

    // uniform mesh case
    if(gridtypex == 0) meshx = new Mesh1D(xl, xr, nx);
    if(gridtypey == 0) meshy = new Mesh1D(yl, yr, ny);
    if(gridtypey == 0) meshz = new Mesh1D(zl, zr, nz);
    // node based grid specification
    if(gridtypex == 1) meshx = new Mesh1D(xl, xr, x);
    if(gridtypey == 1) meshy = new Mesh1D(yl, yr, y);
    if(gridtypez == 1) meshz = new Mesh1D(zl, zr, z);
    // function defined node specification
    if(gridtypex == 2) meshx = new Mesh1D(xl, xr, x);
    if(gridtypey == 2) meshy = new Mesh1D(yl, yr, y);
    if(gridtypez == 2) meshz = new Mesh1D(zl, zr, z);
    // grid based grid specification
    if(gridtypex == 3) meshx = new Mesh1D(xgr);
    if(gridtypey == 3) meshy = new Mesh1D(ygr);
    if(gridtypez == 3) meshz = new Mesh1D(zgr);
    // grid based grid specification
    if(gridtypex == 4) meshx = new Mesh1D(xgr);
    if(gridtypey == 4) meshy = new Mesh1D(ygr);
    if(gridtypez == 4) meshz = new Mesh1D(zgr);

    // to make sure it is all consistent, get the info from the mesh objects
    nx = meshx.getNumberOfNodes();
    nxm1 = nx - 1;
    x = meshx.getNodes();
    dx = meshx.getGridBlockSizes();
    nxp1 = meshx.getNumberOfGridLines();
    xgr = meshx.getGridLines();
    xl = meshx.getLeftBoundary();
    xr = meshx.getRightBoundary();
    ny = meshy.getNumberOfNodes();
    nym1 = ny - 1;
    y = meshy.getNodes();
    dy = meshy.getGridBlockSizes();
    nyp1 = meshy.getNumberOfGridLines();
    ygr = meshy.getGridLines();
    yl = meshy.getLeftBoundary();
    yr = meshy.getRightBoundary();
    nz = meshz.getNumberOfNodes();
    nzm1 = nz - 1;
    z = meshz.getNodes();
    dz = meshz.getGridBlockSizes();
    nzp1 = meshz.getNumberOfGridLines();
    zgr = meshz.getGridLines();
    zl = meshz.getLeftBoundary();
    zr = meshz.getRightBoundary();

    nvelx = ( nx + 1 ) * ny * nz;
    nvely = nx * ( ny + 1 ) * nz;
    nvelz = nx * ny * ( nz + 1 );
    nnp = nx * ny * nz;
    nxny = nx * ny;

  }

  public void testFunctions() throws WellPosedProblemException {
    // test to see if the definition is symbolic
    if(!coefType.equals(SYMBOLIC_FUNCTION_DEFINITION)) {
    outputMessage("EllipticRectMFEObject -> boundary function test");
      if(!testBoundaryFunction()) throw
         new WellPosedProblemException("Boundary Function Not Well Defined");
      return;
    }
    outputMessage("EllipticRectMFEObject -> boundary function ok");
    // test to see if the definition is symbolic
    outputMessage("EllipticRectMFEObject -> coefficient function test");
    if(!forcingType.equals(SYMBOLIC_FUNCTION_DEFINITION)) {
      if(!testCoefficientFunction()) throw
        new WellPosedProblemException("Coefficient Function Not Well Defined");
      return;
    }
    outputMessage("EllipticRectMFEObject -> coefficient function ok");
    // test to see if the definition is symbolic
    outputMessage("EllipticRectMFEObject -> forcing function test");
    if(!boundaryEvaluationType.equals(SYMBOLIC_FUNCTION_DEFINITION)) {
      if(!testForcingFunction()) throw
         new WellPosedProblemException("Forcing Function Not Well Defined");
      return;
    }
    outputMessage("EllipticRectMFEObject -> forcing function ok");
  }

  public boolean testBoundaryFunction() {
    for(int k=0; k<nzp1; k++) {
      double zval = zgr[k];
      for(int j=0; j<nyp1; j++) {
        double yval = ygr[j];
          double xval = xgr[0];
          if(boundaryConditionType.equals(NEUMANN_BOUNDARY) 
                     || boundaryConditionType.equals(MIXED_BOUNDARY)) {
            if(getValueFromString(fluxBoundaryFunctionXL,xval,yval,zval)
                                                  == garbage) return false;
          }
          if(boundaryConditionType.equals(DIRICHLET_BOUNDARY) 
                     || boundaryConditionType.equals(MIXED_BOUNDARY)) {
            if(getValueFromString(prsBoundaryFunctionXL,xval,yval,zval)
                                                  == garbage) return false;
          }
          xval = xgr[nx];
          if(boundaryConditionType.equals(NEUMANN_BOUNDARY) 
                     || boundaryConditionType.equals(MIXED_BOUNDARY)) {
            if(getValueFromString(fluxBoundaryFunctionXR,xval,yval,zval)
                                                  == garbage) return false;
          }
          if(boundaryConditionType.equals(DIRICHLET_BOUNDARY) 
                     || boundaryConditionType.equals(MIXED_BOUNDARY)) {
            if(getValueFromString(prsBoundaryFunctionXR,xval,yval,zval)
                                                  == garbage) return false;
          }
      }
    }
    for(int k=0; k<nzp1; k++) {
      double zval = zgr[k];
      for(int i=0; i<nxp1; i++) {
        double xval = xgr[i];
          double yval = ygr[0];
          if(boundaryConditionType.equals(NEUMANN_BOUNDARY) 
                     || boundaryConditionType.equals(MIXED_BOUNDARY)) {
            if(getValueFromString(fluxBoundaryFunctionYL,xval,yval,zval)
                                                    == garbage) return false;
          }
          if(boundaryConditionType.equals(DIRICHLET_BOUNDARY) 
                     || boundaryConditionType.equals(MIXED_BOUNDARY)) {
            if(getValueFromString(prsBoundaryFunctionYL,xval,yval,zval)
                                                    == garbage) return false;
          }
          yval = ygr[ny];
          if(boundaryConditionType.equals(NEUMANN_BOUNDARY) 
                     || boundaryConditionType.equals(MIXED_BOUNDARY)) {
            if(getValueFromString(fluxBoundaryFunctionYR,xval,yval,zval)
                                                    == garbage) return false;
          }
          if(boundaryConditionType.equals(DIRICHLET_BOUNDARY) 
                     || boundaryConditionType.equals(MIXED_BOUNDARY)) {
            if(getValueFromString(prsBoundaryFunctionYR,xval,yval,zval)
                                                    == garbage) return false;
          }
      }
    }
    for(int j=0; j<nyp1; j++) {
      double yval = ygr[j];
      for(int i=0; i<nxp1; i++) {
        double xval = xgr[i];
          double zval = zgr[0];
          if(boundaryConditionType.equals(NEUMANN_BOUNDARY) 
                     || boundaryConditionType.equals(MIXED_BOUNDARY)) {
            if(getValueFromString(fluxBoundaryFunctionZL,xval,yval,zval)
                                                    == garbage) return false;
          }
          if(boundaryConditionType.equals(DIRICHLET_BOUNDARY) 
                     || boundaryConditionType.equals(MIXED_BOUNDARY)) {
            if(getValueFromString(prsBoundaryFunctionZL,xval,yval,zval)
                                                    == garbage) return false;
          }
          zval = zgr[nz];
          if(boundaryConditionType.equals(NEUMANN_BOUNDARY) 
                     || boundaryConditionType.equals(MIXED_BOUNDARY)) {
            if(getValueFromString(fluxBoundaryFunctionZR,xval,yval,zval)
                                                    == garbage) return false;
          }
          if(boundaryConditionType.equals(DIRICHLET_BOUNDARY) 
                     || boundaryConditionType.equals(MIXED_BOUNDARY)) {
            if(getValueFromString(prsBoundaryFunctionZR,xval,yval,zval)
                                                    == garbage) return false;
          }
      }
    }
    return true;
  }

  public boolean testCoefficientFunction() {
    // loop over all nodes to test the function
    for(int k=0; k<nz; k++) {
      int k2 = k * nx * ny;
      double zval = z[k];
      for(int j=0; j<ny; j++) {
        int j2 = j * nx + k2;
        double yval = y[j];
        for(int i=0; i<nx; i++) {
          double xval = x[i];
          for(int l=0; l<3; l++) {
            for(int m=0; m<3; m++)
              if(getValueFromString(a[l][m], xval, yval, zval) == garbage)
                           return false;
          }
        }
      }
    }
    return true;
  }

  public boolean testForcingFunction() {
    for(int k=0; k<nz; k++) {
      int k2 = k * nx * ny;
      double zval = z[k];
      for(int j=0; j<ny; j++) {
        int j2 = j * nx + k2;
        double yval = y[j];
        for(int i=0; i<nx; i++) {
          double xval = x[i];
          if(getValueFromString(f, xval, yval, zval) == garbage) return false;
        }
      }
    }
    return true;
  }

  public void initializeUnknowns() {
    // initialize the head variable
    h = new double[nnp];
    hold = new double[nnp];
    ibith = new double[nnp];
    for(int i=0; i<nnp; i++) {
      h[i] = 0.0;
      hold[i] = 0.0;
      ibith[i] = 1.0;
    }
    velx = new double[nvelx];
    ibitvx = new double[nvelx];
    for(int i=0; i<nvelx; i++) {
      velx[i] = 0.0;
      ibitvx[i] = 1.0;
    }
    vely = new double[nvely];
    ibitvy = new double[nvely];
    for(int j=0; j<nvely; j++) {
      vely[j] = 0.0;
      ibitvy[j] = 1.0;
    }
    velz = new double[nvelz];
    ibitvz = new double[nvelz];
    for(int k=0; k<nvelz; k++) {
      velz[k] = 0.0;
      ibitvz[k] = 1.0;
    }
  }

  public void setBoundaryValues() {
    if(boundaryConditionType.equals(NEUMANN_BOUNDARY)) {
      if(boundaryEvaluationType.equals(SYMBOLIC_FUNCTION_DEFINITION)) {
        // do the x planes
        for(int k=0; k<nz; k++) {
          int k1 = k * nxp1 * ny;
          for(int j=0; j<ny; j++) {
             int index = j * nxp1 + k1;
             velx[index]
               = getValueFromString(fluxBoundaryFunctionXL,xgr[0],y[j],z[k]);
             index = index + nx;
             velx[index]
               = getValueFromString(fluxBoundaryFunctionXR,xgr[nx],y[j],z[k]);
          }
        }
        // do the y planes
        for(int i=0; i<nx; i++) {
          int i1 = i * nyp1 * nz;
          for(int k=0; k<nz; k++) {
             int index = k * nyp1 + i1;
             vely[index]
               = getValueFromString(fluxBoundaryFunctionYL,x[i],ygr[0],z[k]);
             index = index + ny;
             vely[index]
               = getValueFromString(fluxBoundaryFunctionYR,x[i],ygr[ny],z[k]);
          }
        }
        // do the z planes
        for(int j=0; j<ny; j++) {
          int j1 = j * nx * nzp1;
          for(int i=0; i<nx; i++) {
             int index = i * nzp1 + j1;
             velz[index]
               = getValueFromString(fluxBoundaryFunctionZL,x[i],y[j],zgr[0]);
             index = index + nz;
             velz[index]
               = getValueFromString(fluxBoundaryFunctionZR,x[i],y[j],zgr[nz]);
          }
        }
      } else if(boundaryEvaluationType.equals(NUMERIC_FUNCTION_DEFINITION)) {
        // do the x planes
        for(int k=0; k<nz; k++) {
          int k1 = k * nxp1 * ny;
          int k2 = k * nx;
          for(int j=0; j<ny; j++) {
             int j2 = j + k2;
             int index = j * nxp1 + k1;
             velx[index] = xBoundaryFlux[0][j2];
             index = index + nx;
             velx[index] = xBoundaryFlux[1][j2];
          }
        }
        // do the y planes
        for(int i=0; i<nx; i++) {
          int i1 = i * nyp1 * nz;
          int i2 = i;
          for(int k=0; k<nz; k++) {
             int k2 = k * nx + i2;
             int index = k * nyp1 + i1;
             vely[index] = yBoundaryFlux[0][k2];
             index = index + ny;
             vely[index] = yBoundaryFlux[1][k2];
          }
        }
        // do the z planes
        for(int j=0; j<ny; j++) {
          int j1 = j * ny * nzp1;
          int j2 = j * nx;
          for(int i=0; i<nx; i++) {
             int i2 = i + j2;
             int index = i * nzp1 + j1;
             velz[index] = zBoundaryFlux[0][i2];
             index = index + nz;
             velz[index] = zBoundaryFlux[1][i2];
          }
        }
      }
    }
  }

  public void setBoundaryFluxValues(double [] vx, double [] vy, double [] vz) {
    if(boundaryEvaluationType.equals(SYMBOLIC_FUNCTION_DEFINITION)) {
      // do the x planes
      for(int k=0; k<nz; k++) {
        int k1 = k * nxp1 * ny;
        for(int j=0; j<ny; j++) {
           int index = j * nxp1 + k1;
           vx[index] = 
                getValueFromString(fluxBoundaryFunctionXL,xgr[0],y[j],z[k]);
           index = index + nx;
           vx[index] = 
                getValueFromString(fluxBoundaryFunctionXR,xgr[nx],y[j],z[k]);
        }
      }
      // do the y planes
      for(int i=0; i<nx; i++) {
        int i1 = i * nyp1 * nz;
        for(int k=0; k<nz; k++) {
           int index = k * nyp1 + i1;
           vy[index] = 
               getValueFromString(fluxBoundaryFunctionYL,x[i],ygr[0],z[k]);
           index = index + ny;
           vy[index] = 
               getValueFromString(fluxBoundaryFunctionYR,x[i],ygr[ny],z[k]);
        }
      }
      // do the z planes
      for(int j=0; j<ny; j++) {
        int j1 = j * nx * nzp1;
        for(int i=0; i<nx; i++) {
           int index = i * nzp1 + j1;
           vz[index] = 
               getValueFromString(fluxBoundaryFunctionZL,x[i],y[j],zgr[0]);
           index = index + nz;
           vz[index] = 
               getValueFromString(fluxBoundaryFunctionZR,x[i],y[j],zgr[nz]);
        }
      }
    } else if(boundaryEvaluationType.equals(NUMERIC_FUNCTION_DEFINITION)) {
      // do the x planes
      for(int k=0; k<nz; k++) {
        int k1 = k * nxp1 * ny;
        int k2 = k * nx;
        for(int j=0; j<ny; j++) {
           int j2 = j + k2;
           int index = j * nxp1 + k1;
           vx[index] = xBoundaryFlux[0][j2];
           index = index + nx;
           vx[index] = xBoundaryFlux[1][j2];
        }
      }
      // do the y planes
      for(int i=0; i<nx; i++) {
        int i1 = i * nyp1 * nz;
        int i2 = i;
        for(int k=0; k<nz; k++) {
           int k2 = k * nx + i2;
           int index = k * nyp1 + i1;
           vy[index] = yBoundaryFlux[0][k2];
           index = index + ny;
           vy[index] = yBoundaryFlux[1][k2];
        }
      }
      // do the z planes
      for(int j=0; j<ny; j++) {
        int j1 = j * ny * nzp1;
        int j2 = j * nx;
        for(int i=0; i<nx; i++) {
           int i2 = i + j2;
           int index = i * nzp1 + j1;
           vz[index] = zBoundaryFlux[0][i2];
           index = index + nz;
           vz[index] = zBoundaryFlux[1][i2];
        }
      }
    }
  }

  public void zeroFlowX(double [] vx) {
    for(int k=0; k<nz; k++) {
      int k1 = k * nxp1 * ny;
      for(int j=0; j<ny; j++) {
         int index = j * nxp1 + k1;
         vx[index] = 0.0;
         index = index + nx;
         vx[index] = 0.0;
      }
    }
  }

  public void zeroFlowY(double [] vy) {
    // do the y planes
    for(int i=0; i<nx; i++) {
      int i1 = i * nyp1 * nz;
      for(int k=0; k<nz; k++) {
         int index = k * nyp1 + i1;
         vy[index] = 0.0;
         index = index + ny;
         vy[index] = 0.0;
      }
    }
  }
  public void zeroFlowZ(double [] vz) {
    // do the z planes
    for(int j=0; j<ny; j++) {
      int j1 = j * nx * nzp1;
      for(int i=0; i<nx; i++) {
         int index = i * nzp1 + j1;
         vz[index] = 0.0;
         index = index + nz;
         vz[index] = 0.0;
      }
    }
  }

  public void initializeMatrixStorage() {
    rhsh = new double[nnp];
    diagh = new double[nnp];
    for(int i=0; i<nnp; i++) rhsh[i] = 0.0;
    for(int i=0; i<nnp; i++) diagh[i] = 0.0;
    rhsvx = new double[nvelx];
    for(int i=0; i<nvelx; i++) rhsvx[i] = 0.0;
    rhsvy = new double[nvely];
    for(int j=0; j<nvely; j++) rhsvy[j] = 0.0;
    rhsvz = new double[nvelz];
    for(int k=0; k<nvelz; k++) rhsvz[k] = 0.0;
  }

  public void buildBasis() {
    // the lowest order raviart-thomas spaces
    if(order == 1) {
      noOfPresFunc = 1;
      pressureFunction = new TensorProduct3D [1];
      xVelocityFunction = new TensorProduct3D [2];
      yVelocityFunction = new TensorProduct3D [2];
      zVelocityFunction = new TensorProduct3D [2];
      pressureFunction[0]
          = new TensorProduct3D(new PiecewiseConstant1D(),
                                new PiecewiseConstant1D(),
                                new PiecewiseConstant1D());
      noOfXVelFunc = 2;
      xVelocityFunction[0]
          = new TensorProduct3D(new PiecewiseLinear1D(),
                                new PiecewiseConstant1D(),
                                new PiecewiseConstant1D());
      xVelocityFunction[1]
          = new TensorProduct3D(new PiecewiseLinear1D(),
                                new PiecewiseConstant1D(),
                                new PiecewiseConstant1D());
      noOfYVelFunc = 2;
      yVelocityFunction[0]
          = new TensorProduct3D(new PiecewiseConstant1D(),
                                new PiecewiseLinear1D(),
                                new PiecewiseConstant1D());
      yVelocityFunction[1]
          = new TensorProduct3D(new PiecewiseConstant1D(),
                                new PiecewiseLinear1D(),
                                new PiecewiseConstant1D());
      noOfZVelFunc = 2;
      zVelocityFunction[0]
          = new TensorProduct3D(new PiecewiseConstant1D(),
                                new PiecewiseConstant1D(),
                                new PiecewiseLinear1D());
      zVelocityFunction[1]
          = new TensorProduct3D(new PiecewiseConstant1D(),
                                new PiecewiseConstant1D(),
                                new PiecewiseLinear1D());
    }
    // the next to lowest order raviart-thomas subspace definition
    if(order == 2) {
      noOfPresFunc = 1;
      pressureFunction = new TensorProduct3D [8];
      xVelocityFunction = new TensorProduct3D [8];
      yVelocityFunction = new TensorProduct3D [8];
      zVelocityFunction = new TensorProduct3D [8];
    }
  }

  private void buildVelMatrices() {
    // test for the type of coefficient definition in the operator
    if(coefType.equals(SYMBOLIC_FUNCTION_DEFINITION)) {
      buildVelMatricesSymbolic();
    } else {
      buildVelMatricesSamples();
    }
  }

  private void buildVelMatricesSymbolic() {
    // get the grid lines in the three dimensions
    xgr = meshx.getGridLines();
    ygr = meshy.getGridLines();
    zgr = meshz.getGridLines();

    // set up the storage for the matrices
    outputMessage("                      -> building x-velocity matrix");
    adx = new double[nvelx];
    asx = new double[nvelx];
    an = new double[nnp];
    for(int i=0; i<nvelx; i++) {
      adx[i] = 0.0; asx[i] = 0.0;
    }
    // do the integration for the x component
    for(int k=0; k<nz; k++) {
      int k2 = k * nx * ny;
      int k22 = k * nxp1 * ny;
      for(int j=0; j<ny; j++) {
        int j2 = j * nx + k2;
        int j22 = j * nxp1 + k22;
        // do the first half of the first basis function
        if(nx>1) {
          xVelocityFunction[0].setDomain(xgr[0], xgr[2],
                                         ygr[j], ygr[j+1],
                                         zgr[k], zgr[k+1]);
        } else {
          xVelocityFunction[0].setDomain(xgr[0], xgr[1]+dx[0],
                                         ygr[j], ygr[j+1],
                                         zgr[k], zgr[k+1]);
        }
        adx[1+j22] = adx[1+j22]
                   + inrprd("1.0/(" + a[0][0] + ")",
                            xVelocityFunction[0], xVelocityFunction[0],
                            xgr[0], xgr[1],
                            ygr[j], ygr[j+1],
                            zgr[k], zgr[k+1]);
        for(int i=1; i<nxm1; i++) {
          // set up the basis functions for evaluation on the appropriate
          // domain
          xVelocityFunction[0].setDomain(xgr[i]-dx[i], xgr[i+1],
                                         ygr[j], ygr[j+1],
                                         zgr[k], zgr[k+1]);
          xVelocityFunction[1].setDomain(xgr[i], xgr[i+1]+dx[i+1],
                                         ygr[j], ygr[j+1],
                                         zgr[k], zgr[k+1]);
          // get the diagonal value in the tensor in the left cell
          // add to the diagonal of the matrix
          adx[i+j22] = adx[i+j22]
                    + inrprd("1.0/(" + a[0][0] + ")",
                             xVelocityFunction[0], xVelocityFunction[0],
                             xgr[i], xgr[i+1],
                             ygr[j], ygr[j+1],
                             zgr[k], zgr[k+1]);
          // add to the diagonal of the matrix
          adx[i+1+j22] = adx[i+1+j22]
                    + inrprd("1.0/(" + a[0][0] + ")",
                             xVelocityFunction[1], xVelocityFunction[1],
                             xgr[i], xgr[i+1],
                             ygr[j], ygr[j+1],
                             zgr[k], zgr[k+1]);
          // add to the super/subdiagonal of the matrix
          asx[i+j22] = asx[i+j22]
                    + inrprd("1.0/(" + a[0][0] + ")",
                             xVelocityFunction[0], xVelocityFunction[1],
                             xgr[i], xgr[i+1],
                             ygr[j], ygr[j+1],
                             zgr[k], zgr[k+1]);
        }
        // do the second half of the last basis function
        if(nx>1) {
          xVelocityFunction[1].setDomain(xgr[nx-2], xgr[nx],
                                         ygr[j], ygr[j+1],
                                         zgr[k], zgr[k+1]);
        } else {
          xVelocityFunction[1].setDomain(xgr[nxm1]-dx[nxm1], xgr[nx],
                                         ygr[j], ygr[j+1],
                                         zgr[k], zgr[k+1]);
        }
        adx[nxm1+j22] = adx[nxm1+j22]
                     + inrprd("1.0/(" + a[0][0] + ")",
                                xVelocityFunction[1], xVelocityFunction[1],
                                xgr[nxm1], xgr[nx],
                                ygr[j], ygr[j+1],
                                zgr[k], zgr[k+1]);
      }

    }

    // set up the storage for the matrices
    outputMessage("                      -> building y-velocity matrix");
    ady = new double[nvely];
    asy = new double[nvely];
    for(int i=0; i<nvely; i++) {
      ady[i] = 0.0; asy[i] = 0.0;
    }
    // do the integration for the y component
    for(int i=0; i<nx; i++) {
      int i2 = i;
      int i22 = i * nyp1 * nz;
      for(int k=0; k<nz; k++) {
        int k2 = k * nx * ny + i2;
        int k22 = k * nyp1 + i22;
        // do the first half of the first basis function
        if(ny>1) {
          yVelocityFunction[0].setDomain(xgr[i], xgr[i+1],
                                         ygr[0], ygr[2],
                                         zgr[k], zgr[k+1]);
        } else {
          yVelocityFunction[0].setDomain(xgr[i], xgr[i+1],
                                         ygr[0], ygr[1]+dy[0],
                                         zgr[k], zgr[k+1]);
        }
        ady[1+k22] = ady[1+k22]
                   + inrprd("1.0/(" + a[1][1] + ")",
                            yVelocityFunction[0], yVelocityFunction[0],
                            xgr[i], xgr[i+1],
                            ygr[0], ygr[1],
                            zgr[k], zgr[k+1]);
        for(int j=1; j<nym1; j++) {
          yVelocityFunction[0].setDomain(xgr[i], xgr[i+1],
                                         ygr[j]-dy[j], ygr[j+1],
                                         zgr[k], zgr[k+1]);
          yVelocityFunction[1].setDomain(xgr[i], xgr[i+1],
                                         ygr[j], ygr[j+1]+dy[j+1],
                                         zgr[k], zgr[k+1]);
          ady[j+k22] = ady[j+k22]
                    + inrprd("1.0/(" + a[1][1] + ")",
                             yVelocityFunction[0], yVelocityFunction[0],
                             xgr[i], xgr[i+1],
                             ygr[j], ygr[j+1],
                             zgr[k], zgr[k+1]);
          ady[j+1+k22] = ady[j+1+k22]
                    + inrprd("1.0/(" + a[1][1] + ")",
                             yVelocityFunction[1], yVelocityFunction[1],
                             xgr[i], xgr[i+1],
                             ygr[j], ygr[j+1],
                             zgr[k], zgr[k+1]);
          asy[j+k22] = asy[j+k22]
                    + inrprd("1.0/(" + a[1][1] + ")",
                             yVelocityFunction[0], yVelocityFunction[1],
                             xgr[i], xgr[i+1],
                             ygr[j], ygr[j+1],
                             zgr[k], zgr[k+1]);
        }
        // do the second half of the last basis function
        if(ny>1) {
          yVelocityFunction[1].setDomain(xgr[i], xgr[i+1],
                                         ygr[ny-2], ygr[ny],
                                         zgr[k], zgr[k+1]);
        } else {
          yVelocityFunction[1].setDomain(xgr[i], xgr[i+1],
                                         ygr[nym1]-dy[nym1], ygr[ny],
                                         zgr[k], zgr[k+1]);
        }
        ady[nym1+k22] = ady[nym1+k22]
                     + inrprd("1.0/(" + a[1][1] + ")",
                                yVelocityFunction[1], yVelocityFunction[1],
                                xgr[i], xgr[i+1],
                                ygr[nym1], ygr[ny],
                                zgr[k], zgr[k+1]);
      }

    }

    // set up the storage for the matrices
    outputMessage("                      -> building z-velocity matrix");
    adz = new double[nvelz];
    asz = new double[nvelz];
    for(int i=0; i<nvelz; i++) {
      adz[i] = 0.0; asz[i] = 0.0;
    }
    // do the integration for the y component
    for(int j=0; j<ny; j++) {
      int j2 = j * nx;
      int j22 = j * nx * nzp1;
      for(int i=0; i<nx; i++) {
        int i2 = i + j2;
        int i22 = i * nzp1 + j22;
        // do the first half of the first basis function
        if(nz>1) {
          zVelocityFunction[0].setDomain(xgr[i], xgr[i+1],
                                         ygr[j], ygr[j+1],
                                         zgr[0], zgr[2]);
        } else {
          zVelocityFunction[0].setDomain(xgr[i], xgr[i+1],
                                         ygr[j], ygr[j+1],
                                         zgr[0], zgr[1]+dz[0]);
        }
        adz[1+i22] = adz[1+i22]
                   + inrprd("1.0/(" + a[2][2] + ")",
                            zVelocityFunction[0], zVelocityFunction[0],
                            xgr[i], xgr[i+1],
                            ygr[j], ygr[j+1],
                            zgr[0], zgr[1]);
        for(int k=1; k<nzm1; k++) {
          zVelocityFunction[0].setDomain(xgr[i], xgr[i+1],
                                         ygr[j], ygr[j+1],
                                         zgr[k]-dz[k], zgr[k+1]);
          zVelocityFunction[1].setDomain(xgr[i], xgr[i+1],
                                         ygr[j], ygr[j+1],
                                         zgr[k], zgr[k+1]+dz[k+1]);
          adz[k+i22] = adz[k+i22]
                    + inrprd("1.0/(" + a[2][2] + ")",
                             zVelocityFunction[0], zVelocityFunction[0],
                             xgr[i], xgr[i+1],
                             ygr[j], ygr[j+1],
                             zgr[k], zgr[k+1]);
          adz[k+1+i22] = adz[k+1+i22]
                    + inrprd("1.0/(" + a[2][2] + ")",
                             zVelocityFunction[1], zVelocityFunction[1],
                             xgr[i], xgr[i+1],
                             ygr[j], ygr[j+1],
                             zgr[k], zgr[k+1]);
          asz[k+i22] = asz[k+i22]
                    + inrprd("1.0/(" + a[2][2] + ")",
                             zVelocityFunction[0], zVelocityFunction[1],
                             xgr[i], xgr[i+1],
                             ygr[j], ygr[j+1],
                             zgr[k], zgr[k+1]);
        }
        // do the second half of the last basis function
        if(nz>1) {
          zVelocityFunction[1].setDomain(xgr[i], xgr[i+1],
                                         ygr[j], ygr[j+1],
                                         zgr[nz-2], zgr[nz]);
        } else {
          zVelocityFunction[1].setDomain(xgr[i], xgr[i+1],
                                         ygr[j], ygr[j+1],
                                         zgr[nzm1]-dz[nzm1], zgr[nz]);
        }
        adz[nzm1+i22] = adz[nzm1+i22]
                     + inrprd("1.0/(" + a[2][2] + ")",
                                zVelocityFunction[1], zVelocityFunction[1],
                                xgr[i], xgr[i+1],
                                ygr[j], ygr[j+1],
                                zgr[nzm1], zgr[nz]);
      }
    }

  }

  private void buildVelMatricesSamples() {
    // get the grid lines in the three dimensions
    xgr = meshx.getGridLines();
    ygr = meshy.getGridLines();
    zgr = meshz.getGridLines();

    // set up the storage for the matrices
    outputMessage("                      -> building x-velocity matrix");
    adx = new double[nvelx];
    asx = new double[nvelx];
    an = new double[nnp];
    for(int i=0; i<nvelx; i++) {
      adx[i] = 0.0; asx[i] = 0.0;
    }
    // do the integration for the x component
    for(int k=0; k<nz; k++) {
      int k2 = k * nx * ny;
      int k22 = k * nxp1 * ny;
      for(int j=0; j<ny; j++) {
        int j2 = j * nx + k2;
        int j22 = j * nxp1 + k22;
        // do the first half of the first basis function
        if(nx>1) {
          xVelocityFunction[0].setDomain(xgr[0], xgr[2],
                                         ygr[j], ygr[j+1],
                                         zgr[k], zgr[k+1]);
        } else {
          xVelocityFunction[0].setDomain(xgr[0], xgr[1]+dx[0],
                                         ygr[j], ygr[j+1],
                                         zgr[k], zgr[k+1]);
        }
        adx[1+j22] = adx[1+j22]
                   + inrprd(1.0/asampxx[j2],
                              xVelocityFunction[0], xVelocityFunction[0],
                              xgr[0], xgr[1],
                              ygr[j], ygr[j+1],
                              zgr[k], zgr[k+1]);
        for(int i=1; i<nxm1; i++) {
          double val = asampyy[i+j2];
          // set up the basis functions for evaluation on the appropriate
          // domain
          xVelocityFunction[0].setDomain(xgr[i]-dx[i], xgr[i+1],
                                         ygr[j], ygr[j+1],
                                         zgr[k], zgr[k+1]);
          xVelocityFunction[1].setDomain(xgr[i], xgr[i+1]+dx[i+1],
                                         ygr[j], ygr[j+1],
                                         zgr[k], zgr[k+1]);
          // get the diagonal value in the tensor in the left cell
          // add to the diagonal of the matrix
          adx[i+j22] = adx[i+j22]
                    + inrprd(1.0/val,
                             xVelocityFunction[0], xVelocityFunction[0],
                             xgr[i], xgr[i+1],
                             ygr[j], ygr[j+1],
                             zgr[k], zgr[k+1]);
          // add to the diagonal of the matrix
          adx[i+1+j22] = adx[i+1+j22]
                    + inrprd(1.0/val,
                             xVelocityFunction[1], xVelocityFunction[1],
                             xgr[i], xgr[i+1],
                             ygr[j], ygr[j+1],
                             zgr[k], zgr[k+1]);
          // add to the super/subdiagonal of the matrix
          asx[i+j22] = asx[i+j22]
                    + inrprd(1.0/val,
                             xVelocityFunction[0], xVelocityFunction[1],
                             xgr[i], xgr[i+1],
                             ygr[j], ygr[j+1],
                             zgr[k], zgr[k+1]);
        }
        // do the second half of the last basis function
        if(nx>1) {
          xVelocityFunction[1].setDomain(xgr[nx-2], xgr[nx],
                                         ygr[j], ygr[j+1],
                                         zgr[k], zgr[k+1]);
        } else {
          xVelocityFunction[1].setDomain(xgr[nxm1]-dx[nxm1], xgr[nx],
                                         ygr[j], ygr[j+1],
                                         zgr[k], zgr[k+1]);
        }
        adx[nxm1+j22] = adx[nxm1+j22]
                     + inrprd(1.0/asampxx[nxm1+j2],
                                xVelocityFunction[1], xVelocityFunction[1],
                                xgr[nxm1], xgr[nx],
                                ygr[j], ygr[j+1],
                                zgr[k], zgr[k+1]);
      }
    }
    // set up the storage for the matrices
    outputMessage("                      -> building y-velocity matrix");
    ady = new double[nvely];
    asy = new double[nvely];
    for(int i=0; i<nvely; i++) {
      ady[i] = 0.0; asy[i] = 0.0;
    }
    // do the integration for the y component
    for(int i=0; i<nx; i++) {
      int i2 = i;
      int i22 = i * nyp1 * nz;
      for(int k=0; k<nz; k++) {
        int k2 = k * nx * ny + i2;
        int k22 = k * nyp1 + i22;
        // do the first half of the first basis function
        if(ny>1) {
          yVelocityFunction[0].setDomain(xgr[i], xgr[i+1],
                                         ygr[0], ygr[2],
                                         zgr[k], zgr[k+1]);
        } else {
          yVelocityFunction[0].setDomain(xgr[i], xgr[i+1],
                                         ygr[0], ygr[1]+dy[0],
                                         zgr[k], zgr[k+1]);
        }
        ady[1+k22] = ady[1+k22]
                   + inrprd(1.0/asampyy[k2],
                            yVelocityFunction[0], yVelocityFunction[0],
                            xgr[i], xgr[i+1],
                            ygr[0], ygr[1],
                            zgr[k], zgr[k+1]);
        for(int j=1; j<nym1; j++) {
          double val = asampyy[j*nx+k2];
          yVelocityFunction[0].setDomain(xgr[i], xgr[i+1],
                                         ygr[j]-dy[j], ygr[j+1],
                                         zgr[k], zgr[k+1]);
          yVelocityFunction[1].setDomain(xgr[i], xgr[i+1],
                                         ygr[j], ygr[j+1]+dy[j+1],
                                         zgr[k], zgr[k+1]);
          ady[j+k22] = ady[j+k22]
                    + inrprd(1.0/val,
                             yVelocityFunction[0], yVelocityFunction[0],
                             xgr[i], xgr[i+1],
                             ygr[j], ygr[j+1],
                             zgr[k], zgr[k+1]);
          ady[j+1+k22] = ady[j+1+k22]
                    + inrprd(1.0/val,
                             yVelocityFunction[1], yVelocityFunction[1],
                             xgr[i], xgr[i+1],
                             ygr[j], ygr[j+1],
                             zgr[k], zgr[k+1]);
          asy[j+k22] = asy[j+k22]
                    + inrprd(1.0/val,
                             yVelocityFunction[0], yVelocityFunction[1],
                             xgr[i], xgr[i+1],
                             ygr[j], ygr[j+1],
                             zgr[k], zgr[k+1]);
        }
        // do the second half of the last basis function
        if(ny>1) {
          yVelocityFunction[0].setDomain(xgr[i], xgr[i+1],
                                         ygr[ny-2], ygr[ny],
                                         zgr[k], zgr[k+1]);
        } else {
          yVelocityFunction[0].setDomain(xgr[i], xgr[i+1],
                                         ygr[nym1]-dy[nym1], ygr[ny],
                                         zgr[k], zgr[k+1]);
        }
        ady[nym1+k22] = ady[nym1+k22]
                   + inrprd(1.0/asampyy[(nym1)*nx+k2],
                            yVelocityFunction[0], yVelocityFunction[0],
                            xgr[i], xgr[i+1],
                            ygr[nym1], ygr[ny],
                            zgr[k], zgr[k+1]);
      }
    }
    // set up the storage for the matrices
    outputMessage("                      -> building z-velocity matrix");
    adz = new double[nvelz];
    asz = new double[nvelz];
    for(int i=0; i<nvelz; i++) {
      adz[i] = 0.0; asz[i] = 0.0;
    }
    // do the integration for the y component
    for(int j=0; j<ny; j++) {
      int j2 = j * nx;
      int j22 = j * nx * nzp1;
      for(int i=0; i<nx; i++) {
        int i2 = i + j2;
        int i22 = i * nzp1 + j22;
        // do the first half of the first basis function
        if(nz>1) {
          zVelocityFunction[0].setDomain(xgr[i], xgr[i+1],
                                         ygr[j], ygr[j+1],
                                         zgr[0], zgr[2]);
        } else {
          zVelocityFunction[0].setDomain(xgr[i], xgr[i+1],
                                         ygr[j], ygr[j+1],
                                         zgr[0], zgr[1]+dz[0]);
        }
        adz[1+i22] = adz[1+i22]
                   + inrprd(1.0/asampzz[i2],
                            zVelocityFunction[0], zVelocityFunction[0],
                            xgr[i], xgr[i+1],
                            ygr[j], ygr[j+1],
                            zgr[0], zgr[1]);
        for(int k=1; k<nzm1; k++) {
          double val = asampyy[k*nx*ny+i2];
          zVelocityFunction[0].setDomain(xgr[i], xgr[i+1],
                                         ygr[j], ygr[j+1],
                                         zgr[k]-dz[k], zgr[k+1]);
          zVelocityFunction[1].setDomain(xgr[i], xgr[i+1],
                                         ygr[j], ygr[j+1],
                                         zgr[k], zgr[k+1]+dz[k+1]);
          adz[k+i22] = adz[k+i22]
                    + inrprd(1.0/val,
                             zVelocityFunction[0], zVelocityFunction[0],
                             xgr[i], xgr[i+1],
                             ygr[j], ygr[j+1],
                             zgr[k], zgr[k+1]);
          adz[k+1+i22] = adz[k+1+i22]
                    + inrprd(1.0/val,
                             zVelocityFunction[1], zVelocityFunction[1],
                             xgr[i], xgr[i+1],
                             ygr[j], ygr[j+1],
                             zgr[k], zgr[k+1]);
          asz[k+i22] = asz[k+i22]
                    + inrprd(1.0/val,
                             zVelocityFunction[0], zVelocityFunction[1],
                             xgr[i], xgr[i+1],
                             ygr[j], ygr[j+1],
                             zgr[k], zgr[k+1]);
        }
        // do the second half of the last basis function
        if(nz>1) {
          zVelocityFunction[0].setDomain(xgr[i], xgr[i+1],
                                         ygr[j], ygr[j+1],
                                         zgr[nz-2], zgr[nz]);
        } else {
          zVelocityFunction[0].setDomain(xgr[i], xgr[i+1],
                                         ygr[j], ygr[j+1],
                                         zgr[nzm1]-dz[nzm1], zgr[nz]);
        }
        adz[nzm1+i22] = adz[nzm1+i22]
                   + inrprd(1.0/asampzz[(nzm1)*nx*ny+i2],
                            zVelocityFunction[0], zVelocityFunction[0],
                            xgr[i], xgr[i+1],
                            ygr[j], ygr[j+1],
                            zgr[nzm1], zgr[nz]);
      }
    }
  }

  public void buildDiagH() {
    double factor = 1.0;
    if(tstep != 0.0) factor = 1 / tstep;
    for(int k=0; k<nz; k++) {
      int k2 = k * nxny;
      double factork = factor * dz[k];
      for(int j=0; j<ny; j++) {
        int j2 = j * nx + k2;
        double factorj = factork * dy[j];
        for(int i=0; i<nx; i++) {
          int indp = i + j2;
          double cnst = tcoeff * factorj * dx[i];
          diagh[indp] = diagh[indp] - cnst;
          rhsh[indp] = rhsh[indp] - cnst;
        }
      }
    }
  }

  public void buildPRSRHS() {
    if(forcingType.equals(SYMBOLIC_FUNCTION_DEFINITION)) {
      buildPRSRHSSymbolic();
    } else if(forcingType.equals(POINT_SOURCE_FUNCTION_DEFINITION)) {
      buildPRSRHSPointSource();
    } else {
      buildPRSRHSSamples();
    }
  }

  public void buildPRSRHSSymbolic() {
    // do the integration for the y component
    for(int k=0; k<nz; k++) {
      int k2 = k * nx * ny;
      for(int j=0; j<ny; j++) {
        int j2 = j * nx + k2;
        for(int i=0; i<nx; i++) {
          int ind = i + j2;
          pressureFunction[0].setDomain(xgr[i], xgr[i+1],
                                        ygr[j], ygr[j+1],
                                        zgr[k], zgr[k+1]);
          rhsh[ind] = rhsh[ind]
                    + inrprd(f,
                             pressureFunction[0], pressureFunction[0],
                             xgr[i], xgr[i+1],
                             ygr[j], ygr[j+1],
                             zgr[k], zgr[k+1]);
        }
      }
    }
  }

  public void buildPRSRHSSamples() {
    // do the integration for the y component
    for(int k=0; k<nz; k++) {
      int k2 = k * nx * ny;
      for(int j=0; j<ny; j++) {
        int j2 = j * nx + k2;
        for(int i=0; i<nx; i++) {
          int ind = i + j2;
          pressureFunction[0].setDomain(xgr[i], xgr[i+1],
                                        ygr[j], ygr[j+1],
                                        zgr[k], zgr[k+1]);
          rhsh[ind] = rhsh[ind]
                    + inrprd(fsamp[ind],
                             pressureFunction[0], pressureFunction[0],
                             xgr[i], xgr[i+1],
                             ygr[j], ygr[j+1],
                             zgr[k], zgr[k+1]);
        }
      }
    }
  }

  public void buildPRSRHSPointSource() {

    // if the point sources do not exist, jump out
    if(pointSources == null) return;

    // get the number of point sources and loop over the sources adding in the
    // contribution
    int n = pointSources.getNumberOfSources();
    if(n < 1) return;
    Vector sv = pointSources.getSources();
    for(int l=0; l<n; l++) {
      double [] d = (double []) sv.get(l);
      double xloc = d[0];
      double yloc = d[1];
      double zloc = d[2];
      double val = d[3];

      // set the multiplicative factor based on the location
      double xfac = 1.0;
      if(xloc <= xgr[0]) xfac = 0.5;
      if(xloc >= xgr[nx]) xfac = 0.5;
      double yfac = 1.0;
      if(yloc <= ygr[0]) yfac = 0.5;
      if(yloc >= ygr[ny]) yfac = 0.5;
      double zfac = 1.0;
      if(zloc <= zgr[0]) zfac = 0.5;
      if(zloc >= zgr[nz]) zfac = 0.5;
      val = xfac * yfac * zfac * val;

      // find the location of the source
      int indx = 0;
      for(int i=0; i<nx; i++) if(xloc >= xgr[i] && xloc <= xgr[i+1]) indx = i;
      int indy = 0;
      for(int j=0; j<ny; j++) if(yloc >= ygr[j] && yloc <= ygr[j+1]) indy = j;
      int indz = 0;
      for(int k=0; k<nz; k++) if(zloc >= zgr[k] && zloc <= zgr[k+1]) indz = k;
      int index = indx + indy * nx + indz * nx * ny;

      // evaluate the basis function at the point
      pressureFunction[0].setDomain(xgr[indx], xgr[indx+1],
                                    ygr[indy], ygr[indy+1],
                                    zgr[indz], zgr[indz+1]);
      double bval = pressureFunction[0].evaluate(xloc, xgr[indx], xgr[indx+1],
                                                 yloc, ygr[indy], ygr[indy+1],
                                                 zloc, zgr[indz], zgr[indz+1]);
      rhsh[index] = rhsh[index] + val * bval;

    }

  }

  public void buildFluxBoundaryConditionRHS() throws WellPosedProblemException {
    if(coefType.equals(SYMBOLIC_FUNCTION_DEFINITION)) {
      buildFluxBoundaryConditionRHSSymbolic();
    } else {
      buildFluxBoundaryConditionRHSSamples();
    }
  }

  public void buildFluxBoundaryConditionRHSSymbolic()
                                          throws WellPosedProblemException {
    double [] sx = new double[nvelx];
    for(int i=0; i<nvelx; i++) sx[i] = 0.0;
    double [] sy = new double[nvely];
    for(int j=0; j<nvely; j++) sy[j] = 0.0;
    double [] sz = new double[nvelz];
    for(int k=0; k<nvelz; k++) sz[k] = 0.0;

    // check for a contribution in the first component
    if(nx > 1) {
      outputMessage("                      -> building x-neumann rhs");

      // set a grid parameter
      int nvxtemp = nxp1 * ny;

      // do the integration for the x component
      for(int k=0; k<nz; k++) {
        int k1 = k * nvxtemp;
        int k2 = k * nx * ny;
        for(int j=0; j<ny; j++) {
          int j1 = j * nxp1 + k1;
          int j2 = j * nx + k2;

          // do the left boundary
          xVelocityFunction[0].setDomain(xgr[0]-dx[0], xgr[1],
                                         ygr[j], ygr[j+1],
                                         zgr[k], zgr[k+1]);

          xVelocityFunction[1].setDomain(xgr[0], xgr[2],
                                         ygr[j], ygr[j+1],
                                         zgr[k], zgr[k+1]);
          sx[1+j1] = sx[1+j1]
                     + velx[j1] * inrprd("1.0/(" + a[0][0] + ")",
                              xVelocityFunction[0], xVelocityFunction[1],
                              xgr[0], xgr[1],
                              ygr[j], ygr[j+1],
                              zgr[k], zgr[k+1]);

          // do the right boundary
          xVelocityFunction[0].setDomain(xgr[nx-2], xgr[nx],
                                         ygr[j], ygr[j+1],
                                         zgr[k], zgr[k+1]);
          xVelocityFunction[1].setDomain(xgr[nxm1], xgr[nx]+dx[nxm1],
                                         ygr[j], ygr[j+1],
                                         zgr[k], zgr[k+1]);
          sx[nxm1+j1] = sx[nxm1+j1]
                     + velx[nx+j1] * inrprd("1.0/(" + a[0][0] + ")",
                              xVelocityFunction[0], xVelocityFunction[1],
                              xgr[nxm1], xgr[nx],
                              ygr[j], ygr[j+1],
                              zgr[k], zgr[k+1]);
        }
      }
    }

    // check for a contribution in the second component
    if(ny > 1) {
      outputMessage("                      -> building y-neumann rhs");

      // set a grid parameter
      int nvytemp = nyp1 * nz;

      // do the integration for the y component
      for(int i=0; i<nx; i++) {
        int i1 = i * nvytemp;
        int i2 = i;
        for(int k=0; k<nz; k++) {
          int k1 = k * nyp1 + i1;
          int k2 = k * nx * ny + i2;

          // do the left boundary
          yVelocityFunction[0].setDomain(xgr[i], xgr[i+1],
                                         ygr[0]-dy[0], ygr[1],
                                         zgr[k], zgr[k+1]);

          yVelocityFunction[1].setDomain(xgr[i], xgr[i+1],
                                         ygr[0], ygr[2],
                                         zgr[k], zgr[k+1]);
          sy[1+k1] = sy[1+k1]
                     + vely[k1] * inrprd("1.0/(" + a[1][1] + ")",
                              yVelocityFunction[0], yVelocityFunction[1],
                              xgr[i], xgr[i+1],
                              ygr[0], ygr[1],
                              zgr[k], zgr[k+1]);

          // do the right boundary
          yVelocityFunction[0].setDomain(xgr[i], xgr[i+1],
                                         ygr[ny-2], ygr[ny],
                                         zgr[k], zgr[k+1]);
          yVelocityFunction[1].setDomain(xgr[i], xgr[i+1],
                                         ygr[nym1], ygr[ny]+dy[nym1],
                                         zgr[k], zgr[k+1]);
          sy[nym1+k1] = sy[nym1+k1]
                     + vely[ny+k1] * inrprd("1.0/(" + a[1][1] + ")",
                              yVelocityFunction[0], yVelocityFunction[1],
                              xgr[i], xgr[i+1],
                              ygr[nym1], ygr[ny],
                              zgr[k], zgr[k+1]);
        }
      }
    }

    // check for a contribution in the third component
    if(nz > 1) {
      outputMessage("                      -> building z-neumann rhs");

      // set some parameters
      int nvztemp = nzp1 * nx;

      // do the integration for the y component
      for(int j=0; j<ny; j++) {
        int j1 = j * nvztemp;
        int j2 = j * nx;
        for(int i=0; i<nx; i++) {
          int i1 = i * nzp1 + j1;
          int i2 = i + j2;

          // do the left boundary
          zVelocityFunction[0].setDomain(xgr[i], xgr[i+1],
                                         ygr[j], ygr[j+1],
                                         zgr[0]-dz[0], zgr[1]);
          zVelocityFunction[1].setDomain(xgr[i], xgr[i+1],
                                         ygr[j], ygr[j+1],
                                         zgr[0], zgr[2]);
          sz[1+i1] = sz[1+i1]
                     + velz[i1] * inrprd("1.0/(" + a[2][2] + ")",
                              zVelocityFunction[0], zVelocityFunction[1],
                              xgr[i], xgr[i+1],
                              ygr[j], ygr[j+1],
                              zgr[0], zgr[1]);

          // do the right boundary
          zVelocityFunction[0].setDomain(xgr[i], xgr[i+1],
                                         ygr[j], ygr[j+1],
                                         zgr[nz-2], zgr[nz]);
          zVelocityFunction[1].setDomain(xgr[i], xgr[i+1],
                                         ygr[j], ygr[j+1],
                                         zgr[nzm1], zgr[nz]+dz[nzm1]);
          sz[nzm1+i1] = sz[nzm1+i1]
                     + velz[nz+i1] * inrprd("1.0/(" + a[2][2] + ")",
                              zVelocityFunction[0], zVelocityFunction[1],
                              xgr[i], xgr[i+1],
                              ygr[j], ygr[j+1],
                              zgr[nzm1], zgr[nz]);
        }
      }
    }

    // need to add in the right hand side terms to the pressure equation
    sx = lusol(adx, asx, nx, ny, nz, sx);
    sy = lusol(ady, asy, ny, nz, nx, sy);
    sz = lusol(adz, asz, nz, nx, ny, sz);
    for(int i=0; i<nvelx; i++) rhsvx[i] = rhsvx[i] - sx[i];
    for(int j=0; j<nvely; j++) rhsvy[j] = rhsvy[j] - sy[j];
    for(int k=0; k<nvelz; k++) rhsvz[k] = rhsvz[k] - sz[k];
    double [] rval = ntmlt(rhsvx, rhsvy, rhsvz);
    for(int i=0; i<nnp; i++) rhsh[i] = rhsh[i] - rval[i];

  }

  public void buildFluxBoundaryConditionRHSSamples()
                                          throws WellPosedProblemException {
    double [] sx = new double[nvelx];
    for(int i=0; i<nvelx; i++) sx[i] = 0.0;
    double [] sy = new double[nvely];
    for(int j=0; j<nvely; j++) sy[j] = 0.0;
    double [] sz = new double[nvelz];
    for(int k=0; k<nvelz; k++) sz[k] = 0.0;

    // check for a contribution in the first component
    if(nx > 1) {
      outputMessage("                      -> building x-neumann rhs");

      // set a grid parameter
      int nvxtemp = nxp1 * ny;

      // do the integration for the x component
      for(int k=0; k<nz; k++) {
        int k1 = k * nvxtemp;
        int k2 = k * nx * ny;
        for(int j=0; j<ny; j++) {
          int j1 = j * nxp1 + k1;
          int j2 = j * nx + k2;

          // do the left boundary
          xVelocityFunction[0].setDomain(xgr[0]-dx[0], xgr[1],
                                         ygr[j], ygr[j+1],
                                         zgr[k], zgr[k+1]);

          xVelocityFunction[1].setDomain(xgr[0], xgr[2],
                                         ygr[j], ygr[j+1],
                                         zgr[k], zgr[k+1]);
          if(asampxx[j2] == 0.0) {
            String s = "The coeff. is zero in buildFluxBoundaryRHS - asampxx";
            throw new WellPosedProblemException(s);
          }
          sx[1+j1] = sx[1+j1]
                     + inrprd(velx[j1]/asampxx[j2],
                              xVelocityFunction[0], xVelocityFunction[1],
                              xgr[0], xgr[1],
                              ygr[j], ygr[j+1],
                              zgr[k], zgr[k+1]);

          // do the right boundary
          xVelocityFunction[0].setDomain(xgr[nx-2], xgr[nx],
                                         ygr[j], ygr[j+1],
                                         zgr[k], zgr[k+1]);
          xVelocityFunction[1].setDomain(xgr[nxm1], xgr[nx]+dx[nxm1],
                                         ygr[j], ygr[j+1],
                                         zgr[k], zgr[k+1]);

          if(asampxx[nxm1+j2] == 0.0) {
            String s = "The coeff. is zero in buildFluxBoundaryRHS - asampxx";
            throw new WellPosedProblemException(s);
          }
          sx[nxm1+j1] = sx[nxm1+j1]
                     + inrprd(velx[nx+j1]/asampxx[nxm1+j2],
                              xVelocityFunction[0], xVelocityFunction[1],
                              xgr[nxm1], xgr[nx],
                              ygr[j], ygr[j+1],
                              zgr[k], zgr[k+1]);
        }
      }
    }

    // check for a contribution in the second component
    if(ny > 1) {
      outputMessage("                      -> building y-neumann rhs");

      // set a grid parameter
      int nvytemp = nyp1 * nz;

      // do the integration for the y component
      for(int i=0; i<nx; i++) {
        int i1 = i * nvytemp;
        int i2 = i;
        for(int k=0; k<nz; k++) {
          int k1 = k * nyp1 + i1;
          int k2 = k * nx * ny + i2;

          // do the left boundary
          yVelocityFunction[0].setDomain(xgr[i], xgr[i+1],
                                         ygr[0]-dy[0], ygr[1],
                                         zgr[k], zgr[k+1]);

          yVelocityFunction[1].setDomain(xgr[i], xgr[i+1],
                                         ygr[0], ygr[2],
                                         zgr[k], zgr[k+1]);
          if(asampyy[k2] == 0.0) {
            String s = "The coeff. is zero in buildFluxBoundaryRHS - asampyy";
            throw new WellPosedProblemException(s);
          }
          sy[1+k1] = sy[1+k1]
                     + inrprd(vely[k1]/asampyy[k2],
                              yVelocityFunction[0], yVelocityFunction[1],
                              xgr[i], xgr[i+1],
                              ygr[0], ygr[1],
                              zgr[k], zgr[k+1]);

          // do the right boundary
          yVelocityFunction[0].setDomain(xgr[i], xgr[i+1],
                                         ygr[ny-2], ygr[ny],
                                         zgr[k], zgr[k+1]);
          yVelocityFunction[1].setDomain(xgr[i], xgr[i+1],
                                         ygr[nym1], ygr[ny]+dy[nym1],
                                         zgr[k], zgr[k+1]);
          if(asampyy[nym1+k2] == 0.0) {
            String s = "The coeff. is zero in buildFluxBoundaryRHS - asampyy";
            throw new WellPosedProblemException(s);
          }
          sy[nym1+k1] = sy[nym1+k1]
                     + inrprd(vely[ny+k1]/asampyy[nym1+k2],
                              yVelocityFunction[0], yVelocityFunction[1],
                              xgr[i], xgr[i+1],
                              ygr[nym1], ygr[ny],
                              zgr[k], zgr[k+1]);
        }
      }
    }

    // check for a contribution in the third component
    if(nz > 1) {
      outputMessage("                      -> building z-neumann rhs");

      // set some parameters
      int nvztemp = nzp1 * nx;

      // do the integration for the y component
      for(int j=0; j<ny; j++) {
        int j1 = j * nvztemp;
        int j2 = j * nx;
        for(int i=0; i<nx; i++) {
          int i1 = i * nzp1 + j1;
          int i2 = i + j2;

          // do the left boundary
          zVelocityFunction[0].setDomain(xgr[i], xgr[i+1],
                                         ygr[j], ygr[j+1],
                                         zgr[0]-dz[0], zgr[1]);
          zVelocityFunction[1].setDomain(xgr[i], xgr[i+1],
                                         ygr[j], ygr[j+1],
                                         zgr[0], zgr[2]);
          if(asampzz[i2] == 0.0) {
            String s = "The coeff. is zero in buildFluxBoundaryRHS - asampzz";
            throw new WellPosedProblemException(s);
          }
          sz[1+i1] = sz[1+i1]
                     + inrprd(velz[i1]/asampzz[i2],
                              zVelocityFunction[0], zVelocityFunction[1],
                              xgr[i], xgr[i+1],
                              ygr[j], ygr[j+1],
                              zgr[0], zgr[1]);

          // do the right boundary
          zVelocityFunction[0].setDomain(xgr[i], xgr[i+1],
                                         ygr[j], ygr[j+1],
                                         zgr[nz-2], zgr[nz]);
          zVelocityFunction[1].setDomain(xgr[i], xgr[i+1],
                                         ygr[j], ygr[j+1],
                                         zgr[nzm1], zgr[nz]+dz[nzm1]);

          if(asampzz[nzm1+i2] == 0.0) {
            String s = "The coeff. is zero in buildFluxBoundaryRHS - asampzz";
            throw new WellPosedProblemException(s);
          }
          sz[nzm1+i1] = sz[nzm1+i1]
                     + inrprd(velz[nz+i1]/asampzz[nzm1+i2],
                              zVelocityFunction[0], zVelocityFunction[1],
                              xgr[i], xgr[i+1],
                              ygr[j], ygr[j+1],
                              zgr[nzm1], zgr[nz]);
        }
      }
    }

    // need to add in the right hand side terms to the pressure equation
    sx = lusol(adx, asx, nx, ny, nz, sx);
    sy = lusol(ady, asy, ny, nz, nx, sy);
    sz = lusol(adz, asz, nz, nx, ny, sz);
    for(int i=0; i<nvelx; i++) rhsvx[i] = rhsvx[i] - sx[i];
    for(int j=0; j<nvely; j++) rhsvy[j] = rhsvy[j] - sy[j];
    for(int k=0; k<nvelz; k++) rhsvz[k] = rhsvz[k] - sz[k];
    double [] rval = ntmlt(rhsvx, rhsvy, rhsvz);
    for(int i=0; i<nnp; i++) rhsh[i] = rhsh[i] - rval[i];

  }

  public void buildPRSBoundaryConditionRHS() {
    double [] sx = new double[nvelx];
    for(int i=0; i<nvelx; i++) sx[i] = 0.0;
    double [] sy = new double[nvely];
    for(int j=0; j<nvely; j++) sy[j] = 0.0;
    double [] sz = new double[nvelz];
    for(int k=0; k<nvelz; k++) sz[k] = 0.0;
    setBoundaryFluxValues(sx, sy, sz);
    double [] rval = ntmlt(sx, sy, sz);
    for(int i=0; i<nnp; i++) rhsh[i] = rhsh[i] - rval[i];
  }

  private double inrprd(String fs,
                        TensorProduct3D tp1, TensorProduct3D tp2,
                        double xa, double xb,
                        double ya, double yb,
                        double za, double zb) {
    double [] sig = {0.2113248654052, 0.7886751345948};
    double delx = xb - xa;
    double dely = yb - ya;
    double delz = zb - za;
    double dv = delx * dely * delz;
    double sum = 0.0;
    for(int k=0; k<2; k++) {
      double zv = za + delz * sig[k];
      for(int j=0; j<2; j++) {
        double yv = ya + dely * sig[j];
        for(int i=0; i<2; i++) {
          double xv = xa + delx * sig[i];
          double c = 0.125 * getValueFromString(fs, xv, yv, zv);
          double v1 = tp1.evaluate(xv, xa, xb, yv, ya, yb, zv, za, zb);
          double v2 = tp2.evaluate(xv, xa, xb, yv, ya, yb, zv, za, zb);
          sum = sum + c * v1 * v2 * dv;
        }
      }
    }
    return sum;
  }

  private double inrprd(double constant,
                        TensorProduct3D tp1, TensorProduct3D tp2,
                        double xa, double xb,
                        double ya, double yb,
                        double za, double zb) {
    double [] sig = {0.2113248654052, 0.7886751345948};
    double c = 0.125 * constant;
    double delx = xb - xa;
    double dely = yb - ya;
    double delz = zb - za;
    double dv = delx * dely * delz;
    double sum = 0.0;
    for(int k=0; k<2; k++) {
      double zv = za + delz * sig[k];
      for(int j=0; j<2; j++) {
        double yv = ya + dely * sig[j];
        for(int i=0; i<2; i++) {
          double xv = xa + delx * sig[i];
          double v1 = tp1.evaluate(xv, xa, xb, yv, ya, yb, zv, za, zb);
          double v2 = tp2.evaluate(xv, xa, xb, yv, ya, yb, zv, za, zb);
          sum = sum + c * v1 * v2 * dv;
        }
      }
    }
    return sum;
  }

  private double inrprddx(String fs,
                          TensorProduct3D tp1, TensorProduct3D tp2,
                          double xa, double xb,
                          double ya, double yb,
                          double za, double zb) {
    double [] sig = {0.2113248654052, 0.7886751345948};
    double delx = xb - xa;
    double dely = yb - ya;
    double delz = zb - za;
    double dv = delx * dely * delz;
    double sum = 0.0;
    for(int k=0; k<2; k++) {
      double zv = za + delz * sig[k];
      for(int j=0; j<2; j++) {
        double yv = ya + dely * sig[j];
        for(int i=0; i<2; i++) {
          double xv = xa + delx * sig[i];
          double c = 0.125 * getValueFromString(fs, xv, yv, zv);
          double v1 = tp1.dbydx(xv, xa, xb, yv, ya, yb, zv, za, zb);
          double v2 = tp2.dbydx(xv, xa, xb, yv, ya, yb, zv, za, zb);
          sum = sum + c * v1 * v2 * dv;
        }
      }
    }
    return sum;
  }

  private double inrprddx(double constant,
                          TensorProduct3D tp1, TensorProduct3D tp2,
                          double xa, double xb,
                          double ya, double yb,
                          double za, double zb) {
    double [] sig = {0.2113248654052, 0.7886751345948};
    double c = 0.125 * constant;
    double delx = xb - xa;
    double dely = yb - ya;
    double delz = zb - za;
    double dv = delx * dely * delz;
    double sum = 0.0;
    for(int k=0; k<2; k++) {
      double zv = za + delz * sig[k];
      for(int j=0; j<2; j++) {
        double yv = ya + dely * sig[j];
        for(int i=0; i<2; i++) {
          double xv = xa + delx * sig[i];
          double v1 = tp1.dbydx(xv, xa, xb, yv, ya, yb, zv, za, zb);
          double v2 = tp2.dbydx(xv, xa, xb, yv, ya, yb, zv, za, zb);
          sum = sum + c * v1 * v2 * dv;
        }
      }
    }
    return sum;
  }

  private double inrprddy(String fs,
                          TensorProduct3D tp1, TensorProduct3D tp2,
                          double xa, double xb,
                          double ya, double yb,
                          double za, double zb) {
    double [] sig = {0.2113248654052, 0.7886751345948};
    double delx = xb - xa;
    double dely = yb - ya;
    double delz = zb - za;
    double dv = delx * dely * delz;
    double sum = 0.0;
    for(int k=0; k<2; k++) {
      double zv = za + delz * sig[k];
      for(int j=0; j<2; j++) {
        double yv = ya + dely * sig[j];
        for(int i=0; i<2; i++) {
          double xv = xa + delx * sig[i];
          double c = 0.125 * getValueFromString(fs, xv, yv, zv);
          double v1 = tp1.dbydy(xv, xa, xb, yv, ya, yb, zv, za, zb);
          double v2 = tp2.dbydy(xv, xa, xb, yv, ya, yb, zv, za, zb);
          sum = sum + c * v1 * v2 * dv;
        }
      }
    }
    return sum;
  }

  private double inrprddy(double constant,
                          TensorProduct3D tp1, TensorProduct3D tp2,
                          double xa, double xb,
                          double ya, double yb,
                          double za, double zb) {
    double [] sig = {0.2113248654052, 0.7886751345948};
    double c = 0.125 * constant;
    double delx = xb - xa;
    double dely = yb - ya;
    double delz = zb - za;
    double dv = delx * dely * delz;
    double sum = 0.0;
    for(int k=0; k<2; k++) {
      double zv = za + delz * sig[k];
      for(int j=0; j<2; j++) {
        double yv = ya + dely * sig[j];
        for(int i=0; i<2; i++) {
          double xv = xa + delx * sig[i];
          double v1 = tp1.dbydy(xv, xa, xb, yv, ya, yb, zv, za, zb);
          double v2 = tp2.dbydy(xv, xa, xb, yv, ya, yb, zv, za, zb);
          sum = sum + c * v1 * v2 * dv;
        }
      }
    }
    return sum;
  }

  private double inrprddz(String fs,
                          TensorProduct3D tp1, TensorProduct3D tp2,
                          double xa, double xb,
                          double ya, double yb,
                          double za, double zb) {
    double [] sig = {0.2113248654052, 0.7886751345948};
    double delx = xb - xa;
    double dely = yb - ya;
    double delz = zb - za;
    double dv = delx * dely * delz;
    double sum = 0.0;
    for(int k=0; k<2; k++) {
      double zv = za + delz * sig[k];
      for(int j=0; j<2; j++) {
        double yv = ya + dely * sig[j];
        for(int i=0; i<2; i++) {
          double xv = xa + delx * sig[i];
          double c = 0.125 * getValueFromString(fs, xv, yv, zv);
          double v1 = tp1.dbydz(xv, xa, xb, yv, ya, yb, zv, za, zb);
          double v2 = tp2.dbydz(xv, xa, xb, yv, ya, yb, zv, za, zb);
          sum = sum + c * v1 * v2 * dv;
        }
      }
    }
    return sum;
  }

  private double inrprddz(double constant,
                          TensorProduct3D tp1, TensorProduct3D tp2,
                          double xa, double xb,
                          double ya, double yb,
                          double za, double zb) {
    double [] sig = {0.2113248654052, 0.7886751345948};
    double c = 0.125 * constant;
    double delx = xb - xa;
    double dely = yb - ya;
    double delz = zb - za;
    double dv = delx * dely * delz;
    double sum = 0.0;
    for(int k=0; k<2; k++) {
      double zv = za + delz * sig[k];
      for(int j=0; j<2; j++) {
        double yv = ya + dely * sig[j];
        for(int i=0; i<2; i++) {
          double xv = xa + delx * sig[i];
          double v1 = tp1.dbydz(xv, xa, xb, yv, ya, yb, zv, za, zb);
          double v2 = tp2.dbydz(xv, xa, xb, yv, ya, yb, zv, za, zb);
          sum = sum + c * v1 * v2 * dv;
        }
      }
    }
    return sum;
  }

  public double getCoeffValue(int i, int j, double x, double y, double z) {
    if(!coefType.equals(SYMBOLIC_FUNCTION_DEFINITION)) return garbage; 
    return getValueFromString(a[i-1][j-1], x, y, z);
  }

  public double getForcingValue(double x, double y, double z) {
    if(!forcingType.equals(SYMBOLIC_FUNCTION_DEFINITION)) return garbage; 
    return getValueFromString(f, x, y, z);
  }

  public double getFluxBoundaryXValue(double x, double y, double z) {
    if(!boundaryEvaluationType.equals(SYMBOLIC_FUNCTION_DEFINITION))
                                                          return garbage; 
    // first test to make sure the point is on the boundary
    if(x == xgr[0]) return getValueFromString(fluxBoundaryFunctionXL, x, y, z);
    if(x == xgr[nx]) return getValueFromString(fluxBoundaryFunctionXR, x, y, z);
    return garbage;
  }

  public double getFluxBoundaryYValue(double x, double y, double z) {
    if(!boundaryEvaluationType.equals(SYMBOLIC_FUNCTION_DEFINITION))
                                                          return garbage; 
    // first test to make sure the point is on the boundary
    if(y == ygr[0]) return getValueFromString(fluxBoundaryFunctionYL, x, y, z);
    if(y == ygr[ny]) return getValueFromString(fluxBoundaryFunctionYR, x, y, z);
    return garbage;
  }

  public double getFluxBoundaryZValue(double x, double y, double z) {
    if(!boundaryEvaluationType.equals(SYMBOLIC_FUNCTION_DEFINITION))
                                                          return garbage; 
    // first test to make sure the point is on the boundary
    if(z == zgr[0]) return getValueFromString(fluxBoundaryFunctionZL, x, y, z);
    if(z == zgr[nz]) return getValueFromString(fluxBoundaryFunctionZR, x, y, z);
    return garbage;
  }

  public double getPrsBoundaryXValue(double x, double y, double z) {
    if(!boundaryEvaluationType.equals(SYMBOLIC_FUNCTION_DEFINITION))
                                                          return garbage; 
    // first test to make sure the point is on the boundary
    if(x != xgr[0]) return getValueFromString(prsBoundaryFunctionXL, x, y, z);
    if(x != xgr[nx]) return getValueFromString(prsBoundaryFunctionXR, x, y, z);
    return garbage;
  }

  public double getPrsBoundaryYValue(double x, double y, double z) {
    if(!boundaryEvaluationType.equals(SYMBOLIC_FUNCTION_DEFINITION))
                                                          return garbage; 
    // first test to make sure the point is on the boundary
    if(y != ygr[0]) return getValueFromString(prsBoundaryFunctionYL, x, y, z);
    if(y != ygr[ny]) return getValueFromString(prsBoundaryFunctionYR, x, y, z);
    return garbage;
  }

  public double getPrsBoundaryZValue(double x, double y, double z) {
    if(!boundaryEvaluationType.equals(SYMBOLIC_FUNCTION_DEFINITION))
                                                          return garbage; 
    // first test to make sure the point is on the boundary
    if(z != zgr[0]) return getValueFromString(prsBoundaryFunctionZL, x, y, z);
    if(z != zgr[nz]) return getValueFromString(prsBoundaryFunctionZR, x, y, z);
    return garbage;
  }

  private void computeSolution(double [] ibit) {

    // check to see if it is possible to compute the solution
    if(problem) return;

    // update the old value to contain the new value
    for(int i=0; i<nnp; i++) hold[i] = h[i];

    // overall tolerance for the solution
    double cgeps = 0.000001;
    // vector length tolerance
    double veceps = 0.0000000001;
    // scale tolerance
    double scleps = 0.0000000001;
    // threshold tolerances for upper and lower approximations
    double cgthrshl = 0.0001;
    double cgthrshu = 0.01;

    // compute a maximum number of iterations allowed based on the size of the
    // problem
    int maxiter = 5 * ( nnp + nvelx + nvely + nvelz );

    // the error variables - current and best so far
    double error = cgeps + 1.0;
    double errbst = 100000.0 * cgeps;
    // upper and lower thresholds
    boolean lower = false;
    boolean upper = false;

    // temporary storage
    double [] r = new double [nnp];
    double [] c = new double [nnp];
    double [] q = new double [nnp];
    double [] htmp = new double [nnp];
    for(int i=0; i<nnp; i++) htmp[i] = hold[i];
    double [] hnew = new double [nnp];
    double [] hbst = new double [nnp];

    // now do the cg thing

    // initialize the residual and scale of the problem
    double scale = 0.0;
    for(int i=0; i<nnp; i++) {
      r[i] = 0.0;
      c[i] = 0.0;
      scale = scale + htmp[i] * htmp[i];
    }
    scale = Math.sqrt(scale);
    if(scale < scleps) scale = 1.0;

    // compute the initial residual
    r = femmm(htmp);

    // initialize some things
    double rw = 0.0;
    for(int i=0; i<nnp; i++) {
      // compute the components of the initial residual vector
      r[i] = ibit[i] * ( rhsh[i] - r[i] );
      // set the components of the initial search direction to the residual
      // vector components
      q[i] = r[i];
      // sum up the initial residual (squared)
      rw = rw + r[i] * r[i];
      // initialize the new value to be the original guess
      hnew[i] = htmp[i];
    }

    // check the residual - if zero we cannot do any better
    if(rw == 0.0) return;

    // initialize the iteration counter
    int iter = 0;

    // start the iteration loop
    while(error > cgeps && iter < maxiter) { 

      // initialize the old residual value for later
      double rwold = rw;

      // update the iteration counter
      iter++;

      // compute the action of the matrix on the direction vector
      c = femmm(q);

      // compute the conjugate product q-transpose*A*q
      double qw = 0.0;
      for(int i=0; i<nnp; i++) qw = qw + q[i] * c[i];

      // test the length of the direction vector
      double a = rwold;
      if(qw != 0.0) a = rwold / qw;

      // compute the new iterate and residual vector
      for(int i=0; i<nnp; i++) {
        hnew[i] = htmp[i] + a * q[i];
        r[i] = r[i] - a * c[i];
      }

      // compute the residual
      rw = 0.0;
      error = 0.0;
      for(int i=0; i<nnp; i++) {
        rw = rw + r[i] * r[i];
        double temp = hnew[i] - htmp[i];
        error = error + temp * temp;
      }
      error = Math.sqrt(error) / scale;

      // check for the threshold trips
      if(error < cgthrshl) lower = true;
      if(lower && error > cgthrshu) upper = true;

      // keep the best solution
      if(error < errbst) {
        errbst = error; 
        for(int i=0; i<nnp; i++) hbst[i] = hnew[i];
      }

      // compute the multiplier for the direction 
      double b = rw / rwold;
      for(int i=0; i<nnp; i++) {
        q[i] = r[i] + b * q[i];
        htmp[i] = hnew[i];
      }

      // compute the length of the new search vector
      qw = 0.0;
      for(int i=0; i<nnp; i++) qw = qw + q[i] * q[i];
      qw = Math.sqrt(qw);
      if(qw < .00000000001) upper = true;

    }

    // output a message if the maximum number of iterations is reached
    if(iter >= maxiter)
      outputMessage("maximum number of iterations exceeded in cg routine");
    // output a message if the tolerance is not attaned
    if(error >= cgeps)
      outputMessage("error tolerance is not attained in cg routine");

    // set the solution after the simulation is complete
    for(int i=0; i<nnp; i++) h[i] = hbst[i];
 
  }

  public boolean testSolution() {
    if(h == null) return false;
    return true;
  }

  private void computeVelocity() {
    // set up some temporary storage
    double [] sx;
    double [] sy;
    double [] sz;
    // compute the differencing operation
    sx = nmltx(h);
    sy = nmlty(h);
    sz = nmltz(h);
    sx = lusol(adx, asx, nx, ny, nz, sx);
    sy = lusol(ady, asy, ny, nz, nx, sy);
    sz = lusol(adz, asz, nz, nx, ny, sz);
    for(int i=0; i<nvelx; i++) sx[i] = sx[i] + rhsvx[i];
    for(int j=0; j<nvely; j++) sy[j] = sy[j] + rhsvy[j];
    for(int k=0; k<nvelz; k++) sz[k] = sz[k] + rhsvz[k];
    velx = ndmmodx(sx, sy, sz);
    vely = ndmmody(sx, sy, sz);
    velz = ndmmodz(sx, sy, sz);
    setBoundaryValues();
  }

  private double [] nmltx(double [] p) {
    double [] vx = new double[nvelx];
    int nt1 = nxp1 * ny;
    int nt2 = nx * ny;
    for(int k=0; k<nz; k++) {
      int k1 = k * nt1;
      int k2 = k * nt2;
      for(int j=0; j<ny; j++) {
        int indvt = k1 + j * nxp1;
        int indpt = k2 + j * nx;
        double factor = dz[k] * dy[j];
        for(int i=1; i<nx; i++) {
          int indp = i + indpt;
          int indv = i + indvt;
          vx[indv] = ( p[indp-1] - p[indp] ) * factor;
        }
      }
    }
    return vx;
  }

  private double [] nmlty(double [] p) {
    double [] vy = new double[nvely];
    int nt1 = nyp1 * nz;
    int nt2 = nx * ny;
    for(int i=0; i<nx; i++) {
      int i1 = i * nt1;
      for(int k=0; k<nz; k++) {
        int indvt = i1 + k * nyp1;
        int indpt = i + k * nt2;
        double factor = dx[i] * dz[k];
        for(int j=1; j<ny; j++) {
          int indp = indpt + j * nx;
          int indv = j + indvt;
          vy[indv] = ( p[indp-nx] - p[indp] ) * factor;
        }
      }
    }
    return vy;
  }

  private double [] nmltz(double [] p) {
    double [] vz = new double[nvelz];
    int nt1 = nzp1 * nx;
    int nt2 = nx * ny;
    for(int j=0; j<ny; j++) {
      int j1 = j * nt1;
      int j2 = j * nx;
      for(int i=0; i<nx; i++) {
        int indvt = j1 + i * nzp1;
        int indpt = i + j2;
        double factor = dy[j] * dx[i];
        for(int k=1; k<nz; k++) {
          int indp = indpt + k * nt2;
          int indv = k + indvt;
          vz[indv] = ( p[indp-nt2] - p[indp] ) * factor;
        }
      }
    }
    return vz;
  }

  private double [] nmltxf(double [] p) {
    int nt2 = nx * ny;
    // output vector
    double [] vx = new double[nvelx];
    // do the x components of the velocity
    int nt1 = nxp1 * ny;
    for(int k=0; k<nz; k++) {
      int k1 = k * nt1;
      int k2 = k * nt2;
      for(int j=0; j<ny; j++) {
        int indvt = k1 + j * nxp1;
        int indpt = k2 + j * nx;
        double factor = dy[j] * dz[k];
        for(int i=1; i<nx; i++) {
          int indp = i + indpt;
          int indv = i + indvt;
          vx[indv] = ibitvx[indv] * ( p[indp-1] - p[indp] ) * factor;
        }
      }
    }
    return vx;
  }

  private double [] nmltyf(double [] p) {
    int nt2 = nx * ny;
    // output vector
    double [] vy = new double[nvely];
    // do the y componenets of the velocity
    int nt1 = nyp1 * nz;
    for(int i=0; i<nx; i++) {
      int i1 = i * nt1;
      for(int k=0; k<nz; k++) {
        int indvt = i1 + k * nyp1;
        int indpt = i + k * nt2;
        double factor = dx[i] * dz[k];
        for(int j=1; j<ny; j++) {
          int indp = indpt + j * nx;
          int indv = j + indvt;
          vy[indv] = ibitvy[indv] * ( p[indp-nx] - p[indp] ) * factor;
        }
      }
    }
    return vy;
  }

  private double [] nmltzf(double [] p) {
    int nt2 = nx * ny;
    // output vector
    double [] vz = new double[nvelz];
    // do the z componenets of the velocity
    int nt1 = nzp1 * nx;
    for(int j=0; j<ny; j++) {
      int j1 = j * nt1;
      int j2 = j * nx;
      for(int i=0; i<nx; i++) {
        int indvt = j1 + i * nzp1;
        int indpt = i + j2;
        double factor = dy[j] * dx[i];
        for(int k=1; k<nz; k++) {
          int indp = indpt + k * nt2;
          int indv = k + indvt;
          vz[indv] = ibitvz[indv] * ( p[indp-nt2] - p[indp] ) * factor;
        }
      }
    }
    return vz;
  }

  private double [] ndmmodx(double [] vx, double [] vy, double [] vz) {
    double [] wrk;
    wrk = vnprjctx(vy, vz);
    wrk = lusol(admssx, asmssx, nx, ny, nz, wrk);
    // keep in mind that the following is assuming that vx is a pointer to the
    // appropriate array
    for(int i=0; i<nvelx; i++) vx[i] = vx[i] + wrk[i];
    return vx;
  }

  private double [] ndmmody(double [] vx, double [] vy, double [] vz) {
    double [] wrk;
    wrk = vnprjcty(vx, vz);
    wrk = lusol(admssy, asmssy, ny, nz, nx, wrk);
    // keep in mind that the following is assuming that vx is a pointer to the
    // appropriate array
    for(int j=0; j<nvely; j++) vy[j] = vy[j] + wrk[j];
    return vy;
  }

  private double [] ndmmodz(double [] vx, double [] vy, double [] vz) {
    double [] wrk;
    wrk = vnprjctz(vx, vy);
    wrk = lusol(admssz, asmssz, nz, nx, ny, wrk);
    // keep in mind that the following is assuming that vx is a pointer to the
    // appropriate array
    for(int k=0; k<nvelz; k++) vz[k] = vz[k] + wrk[k];
    return vz;
  }

  private double [] vnprjctx(double [] vy, double [] vz) {
    double [] vxo = new double[nvelx];
    for(int i=0; i<nvelx; i++) vxo[i] = 0;

    // this does the x component
    int ntemp1 = nxp1 * ny;
    int ntemp2 = nx * ny;
    int ntemp3 = nyp1 * nz;
    for(int k=0; k<nz; k++) {
      int k1 = k * ntemp1;
      int k2 = k * ntemp2;
      int k11 = k * nyp1;
      double factor1 = 0.125 * dz[k];
      for(int j=0; j<ny; j++) {
        int j1 = j * nxp1 + k1;
        int j2 = j * nx + k2;
        int j11 = j + k11;
        double factor2 = factor1 * dy[j];
        for(int i=1; i<nx; i++) {
          int indp = i + j2;
          int indvo = i + j1;
          int indvi1 = j11 + ( i - 1 ) * ntemp3;
          int indvi2 = indvi1 + 1;
          int indvi3 = indvi1 + ntemp3;
          int indvi4 = indvi3 + 1;
          double valm = accxy[indp-1];
          double valp = accxy[indp];
          vxo[indvo] = ( valm * ( vy[indvi1] + vy[indvi2] ) * dx[i-1]
                   + valp * ( vy[indvi3] + vy[indvi4] ) * dx[i] ) * factor2;
        }
      }
    }
    ntemp3 = nzp1 * nx;
    for(int k=0; k<nz; k++) {
      int k1 = k * ntemp1;
      int k2 = k * ntemp2;
      int k11 = k;
      double factor1 = 0.125 * dz[k];
      for(int j=0; j<ny; j++) {
        int j1 = j * nxp1 + k1;
        int j2 = j * nx + k2;
        int j11 = k11 + j * ntemp3;
        double factor2 = factor1 * dy[j];
        for(int i=1; i<nx; i++) {
          int indp = i + j2;
          int indvo = i + j1;
          int indvi1 = ( i - 1 ) * nzp1 + j11;
          int indvi2 = indvi1 + 1;
          int indvi3 = indvi1 + nzp1;
          int indvi4 = indvi3 + 1;
          double valm = accxz[indp-1];
          double valp = accxz[indp];
          vxo[indvo] = vxo[indvo]
                     + ( valm * ( vz[indvi1] + vz[indvi2] ) * dx[i-1]
                     + valp * ( vz[indvi3] + vz[indvi4] ) * dx[i] ) * factor2;
        }
      }
    }
    return vxo;
  }

  private double [] vnprjcty(double [] vx, double [] vz) {
    double [] vyo = new double[nvely];
    for(int j=0; j<nvely; j++) vyo[j] = 0;

    // this does the x component
    int ntemp1 = nyp1 * nz;
    int ntemp2 = nx * ny;
    int ntemp3 = nxp1 * ny;

    for(int k=0; k<nz; k++) {
      int k1 = k * nyp1;
      int k2 = k * nx * ny;
      int k11 = k * ntemp3;
      double factor1 = 0.125 * dz[k];
      for(int j=1; j<ny; j++) {
        int j1 = j + k1;
        int j2 = j * nx + k2;
        int j11 = ( j - 1 ) * nxp1 + k11;
        double factor21 = factor1 * dy[j-1];
        double factor22 = factor1 * dy[j];
        for(int i=0; i<nx; i++) {
          int indp = i + j2;
          int indvo = i * ntemp1 + j1;
          int indvi1 = i + j11;
          int indvi2 = indvi1 + 1;
          int indvi3 = indvi1 + nxp1;
          int indvi4 = indvi3 + 1;
          double valm = accyx[indp-nx];
          double valp = accyx[indp];
          vyo[indvo] = ( valm * ( vx[indvi1] + vx[indvi2] ) * factor21
                     + valp * ( vx[indvi3] + vx[indvi4] ) * factor22 ) * dx[i];
        }
      }
    }
    ntemp3 = nzp1 * nx;
    for(int k=0; k<nz; k++) {
      int k1 = k * nyp1;
      int k2 = k * nx * ny;
      int k11 = k;
      double factor1 = 0.125 * dz[k];
      for(int j=1; j<ny; j++) {
        int j1 = j + k1;
        int j2 = j * nx + k2;
        int j11 = k11 + ( j - 1 ) * ntemp3;
        double factor21 = factor1 * dy[j-1];
        double factor22 = factor1 * dy[j];
        for(int i=0; i<nx; i++) {
          int indp = i + j2;
          int indvo = i * ntemp1 + j1;
          int indvi1 = i * nzp1 + j11;
          int indvi2 = indvi1 + 1;
          int indvi3 = indvi1 + ntemp3;
          int indvi4 = indvi3 + 1;
          double valm = accyz[indp-nx];
          double valp = accyz[indp];
          vyo[indvo] = vyo[indvo]
                     + ( valm * ( vz[indvi1] + vz[indvi2] ) * factor21
                     + valp * ( vz[indvi3] + vz[indvi4] ) * factor22 ) * dx[i];
        }
      }
    }
    return vyo;
  }

  private double [] vnprjctz(double [] vx, double [] vy) {
    double [] vzo = new double[nvelz];
    for(int k=0; k<nvelz; k++) vzo[k] = 0;

    // this does the x component
    int ntemp1 = nzp1 * nx;
    int ntemp2 = nx * ny;
    int ntemp3 = nxp1 * ny;
    for(int k=1; k<nz; k++) {
      int k1 = k;
      int k2 = k * nx * ny;
      int k11 = ( k - 1 ) * ntemp3;
      double factor11 = 0.125 * dz[k-1];
      double factor12 = 0.125 * dz[k];
      for(int j=0; j<ny; j++) {
        int j1 = k1 + j * ntemp1;
        int j2 = j * nx + k2;
        int j11 = j * nxp1 + k11;
        double factor21 = factor11 * dy[j];
        double factor22 = factor12 * dy[j];
        for(int i=0; i<nx; i++) {
          int indp = i + j2;
          int indvo = i * nzp1 + j1;
          int indvi1 = i + j11;
          int indvi2 = indvi1 + 1;
          int indvi3 = indvi1 + ntemp3;
          int indvi4 = indvi3 + 1;
          double valm = acczx[indp-nx*ny];
          double valp = acczx[indp];
          vzo[indvo] = ( valm * ( vx[indvi1] + vx[indvi2] ) * factor21
                     + valp * ( vx[indvi3] + vx[indvi4] ) * factor22 ) * dx[i];
        }
      }
    }
    ntemp3 = nyp1 * nz;
    for(int k=1; k<nz; k++) {
      int k1 = k;
      int k2 = k * nx * ny;
      int k11 = ( k - 1 ) * nyp1;
      double factor11 = 0.125 * dz[k-1];
      double factor12 = 0.125 * dz[k];
      for(int j=0; j<ny; j++) {
        int j1 = k1 + j * ntemp1;
        int j2 = j * nx + k2;
        int j11 = j + k11;
        double factor21 = factor11 * dy[j];
        double factor22 = factor12 * dy[j];
        for(int i=0; i<nx; i++) {
          int indp = i + j2;
          int indvo = i * nzp1 + j1;
          int indvi1 = i * ntemp3 + j11;
          int indvi2 = indvi1 + 1;
          int indvi3 = indvi1 + nyp1;
          int indvi4 = indvi2 + 1;
          double valm = acczx[indp-nx*ny];
          double valp = acczy[indp];
          vzo[indvo] = vzo[indvo]
                     + ( valm * ( vy[indvi1] + vy[indvi2] ) * factor21
                     + valp * ( vy[indvi3] + vy[indvi4] ) * factor22 ) * dx[i];
        }
      }
    }
    return vzo;
  }

  private double [] ntmlt(double [] vx, double [] vy, double [] vz) {
    int nt2 = nx * ny;
    // output vector
    double [] p = new double[nnp];
    // compute the contribution due to the x component
    int nt1 = nxp1 * ny;
    for(int k=0; k<nz; k++) {
      int k1 = k * nt1;
      int k2 = k * nt2;
      for(int j=0; j<ny; j++) {
        int indpt = k2 + j * nx;
        int indvt = k1 + j * nxp1;
        double factor = dy[j] * dz[k];
        for(int i=0; i<nx; i++) {
          int indp = i + indpt;
          int indv = i + indvt;
          p[indp] = ( vx[indv+1] - vx[indv] ) * factor;
        }
      }
    }
    // compute the contribution due to the y component
    nt1 = nyp1 * nz;
    for(int i=0; i<nx; i++) {
      int i1 = i * nt1;
      for(int k=0; k<nz; k++) {
        int indpt = i + k * nt2;
        int indvt = i1 + k * nyp1;
        double factor = dx[i] * dz[k];
        for(int j=0; j<ny; j++) {
          int indv = j + indvt;
          int indp = indpt + j * nx;
          p[indp] = p[indp] + ( vy[indv+1] - vy[indv] ) * factor;
        } 
      }
    }
    // compute the contribution due to the z component
    nt1 = nzp1 * nx;
    for(int j=0; j<ny; j++) {
      int j1 = j * nt1;
      int j2 = j * nx;
      for(int i=0; i<nx; i++) {
        int indpt = i + j2;
        int indvt = j1 + i * nzp1;
        double factor = dy[j] * dx[i];
        for(int k=0; k<nz; k++) {
          int indv = k + indvt;
          int indp = indpt + k * nt2;
          p[indp] = p[indp] + ( vz[indv+1] - vz[indv] ) * factor;
        }
      }
    }
    return p;
  }

  private double [] ntmltf(double [] vx, double [] vy, double [] vz) {
    int nt2 = nx * ny;
    // output vector
    double [] p = new double[nnp];
    // compute the contribution due to the x component
    int nt1 = nxp1 * ny;
    for(int k=0; k<nz; k++) {
      int k1 = k * nt1;
      int k2 = k * nt2;
      for(int j=0; j<ny; j++) {
        int indpt = k2 + j * nx;
        int indvt = k1 + j * nxp1;
        double factor = dy[j] * dz[k];
        for(int i=0; i<nx; i++) {
          int indp = i + indpt;
          int indv = i + indvt;
          p[indp] = ibith[indp] * ( vx[indv+1] - vx[indv] ) * factor;
        }
      }
    }
    // compute the contribution due to the y component
    nt1 = nyp1 * nz;
    for(int i=0; i<nx; i++) {
      int i1 = i * nt1;
      for(int k=0; k<nz; k++) {
        int indpt = i + k * nt2;
        int indvt = i1 + k * nyp1;
        double factor = dx[i] * dz[k];
        for(int j=0; j<ny; j++) {
          int indv = j + indvt;
          int indp = indpt + j * nx;
          p[indp] = p[indp] + ibith[indp] * ( vy[indv+1] - vy[indv] ) * factor;
        } 
      }
    }
    // compute the contribution due to the z component
    nt1 = nzp1 * nx;
    for(int j=0; j<ny; j++) {
      int j1 = j * nt1;
      int j2 = j * nx;
      for(int i=0; i<nx; i++) {
        int indpt = i + j2;
        int indvt = j1 + i * nzp1;
        double factor = dy[j] * dx[i];
        for(int k=0; k<nz; k++) {
          int indv = k + indvt;
          int indp = indpt + k * nt2;
          p[indp] = p[indp] + ibith[indp] * ( vz[indv+1] - vz[indv] ) * factor;
        }
      }
    }
    return p;
  }

  private double [] diagmult(double [] p) {
    double [] d = new double[nnp];
    for(int i=0; i<nnp; i++) d[i] = - p[i] + ibith[i] * diagh[i] * hold[i];
    return d;
  }

  public void buildMassMatrix() {
    // build the pieces of the mass matrix
    admssx = new double[nvelx];
    asmssx = new double[nvelx];
    almssx = new double[nvelx];
    admssy = new double[nvely];
    asmssy = new double[nvely];
    almssy = new double[nvely];
    admssz = new double[nvelz];
    asmssz = new double[nvelz];
    almssz = new double[nvelz];
    // build the x component
    int nt1 = nxp1 * ny;
    for(int k=0; k<nz; k++) {
      int k1 = k * nt1;
      double factor1 = dz[k];
      for(int j=0; j<ny; j++) {
        int j1 = j * nxp1 + k1;
        double factor2 = factor1 * dy[j];
        for(int i=1; i<nx; i++) {
          int indv = i + j1;
          admssx[indv] = factor2 * ( dx[i-1] + dx[i] ) / 3.0;
          if(i<nxm1) {
            asmssx[indv] = factor2 * dx[i-1] / 6.0;
            almssx[indv] = factor2 * dx[i-1] / 6.0;
          }
        }
      }
    }
    // build the y component
    nt1 = nyp1 * nz;
    for(int i=0; i<nx; i++) {
      int i1 = i * nt1;
      double factor1 = dx[i];
      for(int k=0; k<nz; k++) {
        int k1 = k * nyp1 + i1;
        double factor2 = factor1 * dz[k];
        for(int j=1; j<ny; j++) {
          int indv = j + k1;
          admssy[indv] = factor2 * ( dy[j-1] + dy[j] ) / 3.0;
          if(j<nym1) {
            asmssy[indv] = factor2 * dy[j-1] / 6.0;
            almssy[indv] = factor2 * dy[j-1] / 6.0;
          }
        }
      }
    }
    // build the z component
    nt1 = nzp1 * nx;
    for(int j=0; j<ny; j++) {
      int j1 = j * nt1;
      double factor1 = dy[j];
      for(int i=0; i<nx; i++) {
        int i1 = i * nzp1 + j1;
        double factor2 = factor1 * dx[i];
        for(int k=1; k<nz; k++) {
          int indv = k + i1;
          admssz[indv] = factor2 * ( dz[k-1] + dz[j] ) / 3.0;
          if(k<nzm1) {
            asmssz[indv] = factor2 * dz[k-1] / 6.0;
            almssz[indv] = factor2 * dz[k-1] / 6.0;
          }
        }
      }
    }
  }

  private double [] lusol(double [] ad, double [] as,
                          int nxl, int nyl, int nzl, double [] s) {
    int nxp1l = nxl + 1;
    int nt1 = nxp1l * nyl;
    int n = s.length;
    double [] wrk = new double[n];
    for(int i=0; i<n; i++) wrk[i] = 0.0;
    double [] ans = new double[n];
    for(int i=0; i<n; i++) ans[i] = 0.0;

    // forward elimination

    // all the first blocks in the matrix structure
    if(nxl > 1) {
      for(int k=0; k<nzl; k++) {
        int k1 = k * nt1;
        for(int j=0; j<nyl; j++) {
          int ind = j * nxp1l + k1 + 1;
          wrk[ind] = s[ind];
        }
      }
    }

    // the rest of the blocks
    for(int i=2; i<nxl; i++) {
      for(int k=0; k<nzl; k++) {
        int k1 = k * nt1;
        for(int j=0; j<nyl; j++) {
          int ind = i + j * nxp1l + k1;
          wrk[ind] = s[ind] - wrk[ind-1] * as[ind-1];
        }
      }
    }

    // back substitution

    // all the last blocks in the matrix structure
    if(nxl > 1) {
      for(int k=0; k<nzl; k++) {
        int k1 = k * nt1;
        for(int j=0; j<nyl; j++) {
          int ind = j * nxp1l + k1 + nxl - 1;
          ans[ind] = wrk[ind] / ad[ind];
        }
      }
    }

    // the rest of the blocks
    for(int i=nxl-2; i>0; i--) {
      for(int k=0; k<nzl; k++) {
        int k1 = k * nt1;
        for(int j=0; j<nyl; j++) {
          int ind = j * nxp1l + k1 + i;
          ans[ind] = wrk[ind] / ad[ind] - as[ind] * ans[ind+1]; 
        }
      }
    }

    return ans;

  }

  private void reduceVelMatrices() {
    // reduce the x component matrix
    int nt1 = nxp1 * ny;
    if(nx > 1) {
      for(int k=0; k<nz; k++) {
        int k1 = k * nt1;
        for(int j=0; j<ny; j++) {
          int ind = j * nxp1 + k1 + 1;
          asx[ind] = asx[ind] / adx[ind];
        }
      }
    }
    for(int i=2; i<nx; i++) {
      for(int k=0; k<nz; k++) {
        int k1 = k * nt1;
        for(int j=0; j<ny; j++) {
          int ind = j * nxp1 + k1;
          adx[ind+i] = adx[ind+i] - asx[ind+i-1] * adx[ind+i-1] * asx[ind+i-1];
        }
      }
      for(int k=0; k<nz; k++) {
        int k1 = k * nt1;
        for(int j=0; j<ny; j++) {
          int ind = j * nxp1 + k1;
          asx[ind+i] = asx[ind+i] / adx[ind+i];
        }
      }
    }
    // reduce the y component matrix
    nt1 = nyp1 * nz;
    if(ny > 1) {
      for(int i=0; i<nx; i++) {
        int i1 = i * nt1;
        for(int k=0; k<nz; k++) {
          int ind = k * nyp1 + i1 + 1;
          asy[ind] = asy[ind] / ady[ind];
        }
      }
    }
    for(int j=2; j<ny; j++) {
      for(int i=0; i<nx; i++) {
        int i1 = i * nt1;
        for(int k=0; k<nz; k++) {
          int ind = k * nyp1 + i1;
          ady[ind+j] = ady[ind+j] - asy[ind+j-1] * ady[ind+j-1] * asy[ind+j-1];
        }
      }
      for(int i=0; i<nx; i++) {
        int i1 = i * nt1;
        for(int k=0; k<nz; k++) {
          int ind = k * nyp1 + i1;
          asy[ind+j] = asy[ind+j] / ady[ind+j];
        }
      }
    }
    // reduce the z component matrix
    nt1 = nzp1 * nx;
    if(nz > 1) {
      for(int j=0; j<ny; j++) {
        int j1 = j * nt1;
        for(int i=0; i<nx; i++) {
          int ind = i * nzp1 + j1 + 1;
          asz[ind] = asz[ind] / adz[ind];
        }
      }
    }
    for(int k=2; k<nz; k++) {
      for(int j=0; j<ny; j++) {
        int j1 = j * nt1;
        for(int i=0; i<nx; i++) {
          int ind = i * nzp1 + j1;
          adz[ind+k] = adz[ind+k] - asz[ind+k-1] * adz[ind+k-1] * asz[ind+k-1];
        }
      }
      for(int j=0; j<ny; j++) {
        int j1 = j * nt1;
        for(int i=0; i<nx; i++) {
          int ind = i * nzp1 + j1;
          asz[ind+k] = asz[ind+k] / adz[ind+k];
        }
      }
    }
  }

  private void reduceMassMatrices() {
    // reduce the x component matrix
    int nt1 = nxp1 * ny;
    if(nx > 1) {
      for(int k=0; k<nz; k++) {
        int k1 = k * nt1;
        for(int j=0; j<ny; j++) {
          int ind = j * nxp1 + k1 + 1;
          asmssx[ind] = asmssx[ind] / admssx[ind];
        }
      }
    }
    for(int i=2; i<nx; i++) {
      for(int k=0; k<nz; k++) {
        int k1 = k * nt1;
        for(int j=0; j<ny; j++) {
          int ind = j * nxp1 + k1;
          admssx[ind+i] = admssx[ind+i]
              - asmssx[ind+i-1] * admssx[ind+i-1] * asmssx[ind+i-1];
        }
      }
      for(int k=0; k<nz; k++) {
        int k1 = k * nt1;
        for(int j=0; j<ny; j++) {
          int ind = j * nxp1 + k1;
          asmssx[ind+i] = asmssx[ind+i] / admssx[ind+i];
        }
      }
    }
    // reduce the y component matrix
    nt1 = nyp1 * nz;
    if(ny > 1) {
      for(int i=0; i<nx; i++) {
        int i1 = i * nt1;
        for(int k=0; k<nz; k++) {
          int ind = k * nyp1 + i1 + 1;
          asmssy[ind] = asmssy[ind] / admssy[ind];
        }
      }
    }
    for(int j=2; j<ny; j++) {
      for(int i=0; i<nx; i++) {
        int i1 = i * nt1;
        for(int k=0; k<nz; k++) {
          int ind = k * nyp1 + i1;
          admssy[ind+j] = admssy[ind+j]
              - asmssy[ind+j-1] * admssy[ind+j-1] * asmssy[ind+j-1];
        }
      }
      for(int i=0; i<nx; i++) {
        int i1 = i * nt1;
        for(int k=0; k<nz; k++) {
          int ind = k * nyp1 + i1;
          asmssy[ind+j] = asmssy[ind+j] / admssy[ind+j];
        }
      }
    }
    // reduce the z component matrix
    nt1 = nzp1 * nx;
    if(nz > 1) {
      for(int j=0; j<ny; j++) {
        int j1 = j * nt1;
        for(int i=0; i<nx; i++) {
          int ind = i * nzp1 + j1 + 1;
          asmssz[ind] = asmssz[ind] / admssz[ind];
        }
      }
    }
    for(int k=2; k<nz; k++) {
      for(int j=0; j<ny; j++) {
        int j1 = j * nt1;
        for(int i=0; i<nx; i++) {
          int ind = i * nzp1 + j1;
          admssz[ind+k] = admssz[ind+k]
              - asmssz[ind+k-1] * admssz[ind+k-1] * asmssz[ind+k-1];
        }
      }
      for(int j=0; j<ny; j++) {
        int j1 = j * nt1;
        for(int i=0; i<nx; i++) {
          int ind = i * nzp1 + j1;
          asmssz[ind+k] = asmssz[ind+k] / admssz[ind+k];
        }
      }
    }
  }

  private void buildCellCenteredCoefficientValues()
                                 throws WellPosedProblemException {
    String s;
    accxy = new double [nnp];
    accxz = new double [nnp];
    accyx = new double [nnp];
    accyz = new double [nnp];
    acczx = new double [nnp];
    acczy = new double [nnp];

    // check for a symbolic evaluation
    if(coefType.equals(SYMBOLIC_FUNCTION_DEFINITION)) {
      for(int k=0; k<nz; k++) {
        int k2 = k * nx * ny;
        double zval = z[k];
        for(int j=0; j<ny; j++) {
          double yval = y[j];
          int j2 = j * nx + k2;
          for(int i=0; i<nx; i++) {
            double xval = x[i];
            int index = i + j2;
            String cs = a[0][1] + "/" + a[1][1];
            accxy[index] = getValueFromString(cs, xval, yval, zval);
            if(accxy[index] == garbage) {
              s = "The modified coefficient symbolic eval returns garbage";
              throw new WellPosedProblemException(s);
            }
            cs = a[0][2] + "/" + a[2][2];
            accxz[index] = getValueFromString(cs, xval, yval, zval);
            if(accxz[index] == garbage) {
              s = "The modified coefficient symbolic eval returns garbage";
              throw new WellPosedProblemException(s);
            }
            cs = a[1][0] + "/" + a[0][0];
            accyx[index] = getValueFromString(cs, xval, yval, zval);
            if(accyx[index] == garbage) {
              s = "The modified coefficient symbolic eval returns garbage";
              throw new WellPosedProblemException(s);
            }
            cs = a[1][2] + "/" + a[2][2];
            accyz[index] = getValueFromString(cs, xval, yval, zval);
            if(accyz[index] == garbage) {
              s = "The modified coefficient symbolic eval returns garbage";
              throw new WellPosedProblemException(s);
            }
            cs = a[2][0] + "/" + a[0][0];
            acczx[index] = getValueFromString(cs, xval, yval, zval);
            if(acczx[index] == garbage) {
              s = "The modified coefficient symbolic eval returns garbage";
              throw new WellPosedProblemException(s);
            }
            cs = a[2][1] + "/" + a[1][1];
            acczy[index] = getValueFromString(cs, xval, yval, zval);
            if(acczy[index] == garbage) {
              s = "The modified coefficient symbolic eval returns garbage";
              throw new WellPosedProblemException(s);
            }
          }
        }
      }
      return;
    }

    // check for a symbolic evaluation
    if(coefType.equals(NUMERIC_FUNCTION_DEFINITION)) {
      for(int k=0; k<nz; k++) {
        int k2 = k * nx * ny;
        double zval = z[k];
        for(int j=0; j<ny; j++) {
          double yval = y[j];
          int j2 = j * nx + k2;
          for(int i=0; i<nx; i++) {
            double xval = x[i];
            int index = i + j2;
            if(asampxx[index] == 0.0) {
              s = "The modified coefficient sample divisor is zero - xx";
              outputMessage("  " + s);
              throw new WellPosedProblemException(s);
            } else {
              accxy[index] = asampxy[index] / asampyy[index]; 
              accxz[index] = asampxz[index] / asampzz[index]; 
            }
            if(asampyy[index] == 0.0) {
              s = "The modified coefficient sample divisor is zero - yy";
              outputMessage("  " + s);
              throw new WellPosedProblemException(s);
            } else {
              accyx[index] = asampyx[index] / asampxx[index]; 
              accyz[index] = asampyz[index] / asampzz[index]; 
            }
            if(asampzz[index] == 0.0) {
              s = "The modified coefficient sample divisor is zero - zz";
              outputMessage("  " + s);
              throw new WellPosedProblemException(s);
            } else {
              acczx[index] = asampzx[index] / asampxx[index]; 
              acczy[index] = asampzy[index] / asampyy[index]; 
            }
          }
        }
      }
      return;
    }

    s = "The coefficient evaluation type is incorrect";
    throw new WellPosedProblemException(s);

  }

  private double [] femmm(double [] t) {
    // set up some temporary storage
    double [] sx = new double[nvelx];
    double [] sy = new double[nvely];
    double [] sz = new double[nvelz];
    double [] q = new double[nnp];
    // compute the differencing operation
    sx = nmltx(t);
    sy = nmlty(t);
    sz = nmltz(t);
    // invert the velocity matrices into the result
    sx = lusol(adx, asx, nx, ny, nz, sx);
    sy = lusol(ady, asy, ny, nz, nx, sy);
    sz = lusol(adz, asz, nz, nx, ny, sz);
    // do the modification of the multiplication for the off-diagonal terms
    // in the tensor
    velx = ndmmodx(sx, sy, sz);
    vely = ndmmody(sx, sy, sz);
    velz = ndmmodz(sx, sy, sz);
    zeroFlowX(velx);
    zeroFlowY(vely);
    zeroFlowZ(velz);
    // compute the inverse of the mass matrx
    // compute the gradiaent of the output from the calculation
    q = ntmlt(velx, vely, velz);
    // add on the diagonal term from a time dependent piece - not necessary for
    // the elliptic equation
    for(int i=0; i<nnp; i++) q[i] = q[i] + ibith[i] * diagh[i] * t[i];
    return q;
  }

  public double getValueFromString(String s, double x, double y, double z) {
    ComplexNumber value = null;
    // Create an input stream to be parsed
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes());
    if(stream == null) return garbage;
    // Create the parser with the stream
    formulaParser = new FormulaParser(stream);
    formulaParser.ReInit(stream);
    String [] indepVars = new String[3];
    indepVars[0] = "x";
    indepVars[1] = "y";
    indepVars[2] = "z";
    ComplexNumber [] inpVals = new ComplexNumber[3];
    inpVals[0] = new ComplexNumber(x, 0.0);
    inpVals[1] = new ComplexNumber(y, 0.0);
    inpVals[2] = new ComplexNumber(z, 0.0);
    try {
      value = formulaParser.evaluateC(s,indepVars,inpVals);
    } catch (ParseException except) {
      value = new ComplexNumber(garbage, 0.0);
    }
    return value.Re();
  }

  public double getValueFromString(String s, double x, String var) {
    ComplexNumber value = null;
    // Create an input stream to be parsed
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes());
    if(stream == null) return garbage;
    // Create the parser with the stream
    formulaParser = new FormulaParser(stream);
    formulaParser.ReInit(stream);
    String [] indepVars = new String[1];
    indepVars[0] = var;
    ComplexNumber [] inpVals = new ComplexNumber[1];
    inpVals[0] = new ComplexNumber(x, 0.0);
    try {
      value = formulaParser.evaluateC(s,indepVars,inpVals);
    } catch (ParseException except) {
      value = new ComplexNumber(garbage, 0.0);
    }
    return value.Re();
  }

  public double [] getSolution() {
    if(h == null) return null;
    double [] tmp = new double[nnp];
    for(int i=0; i<nnp; i++) tmp[i] = h[i];
    return tmp;
  }
  public double [] getXFlux() {
    if(velx == null) return null;
    double [] tmp = new double[nvelx];
    for(int i=0; i<nvelx; i++) tmp[i] = velx[i];
    return tmp;
  }
  public double [] getYFlux() {
    if(vely == null) return null;
    double [] tmp = new double[nvely];
    for(int j=0; j<nvely; j++) tmp[j] = vely[j];
    return tmp;
  }
  public double [] getZFlux() {
    if(velz == null) return null;
    double [] tmp = new double[nvelz];
    for(int k=0; k<nvelz; k++) tmp[k] = velz[k];
    return tmp;
  }
  public int getNumberOfXBlocks() {
    return nx;
  }
  public int getNumberOfYBlocks() {
    return ny;
  }
  public int getNumberOfZBlocks() {
    return nz;
  }
  public double [] getXNodes() {
    if(x == null) return null;
    double [] tmp = new double[nx];
    for(int i=0; i<nx; i++) tmp[i] = x[i];
    return tmp;
  }
  public double [] getYNodes() {
    if(y == null) return null;
    double [] tmp = new double[ny];
    for(int j=0; j<ny; j++) tmp[j] = y[j];
    return tmp;
  }
  public double [] getZNodes() {
    if(z == null) return null;
    double [] tmp = new double[nz];
    for(int k=0; k<nz; k++) tmp[k] = z[k];
    return tmp;
  }
  public double [] getXGrid() {
    if(xgr == null) return null;
    double [] tmp = new double[nx+1];
    for(int i=0; i<=nx; i++) tmp[i] = xgr[i];
    return tmp;
  }
  public double [] getYGrid() {
    if(ygr == null) return null;
    double [] tmp = new double[ny+1];
    for(int j=0; j<=ny; j++) tmp[j] = ygr[j];
    return tmp;
  }
  public double [] getZGrid() {
    if(zgr == null) return null;
    double [] tmp = new double[nz+1];
    for(int k=0; k<=nz; k++) tmp[k] = zgr[k];
    return tmp;
  }

  public void setSolution(double [] d) {
    if(d == null) return;
    if(d.length != nnp) return;
    if(h == null) h = new double[nnp];
    for(int i=0; i<nnp; i++) h[i] = d[i];
  }
  public void setXFlux(double [] d) {
    if(d == null) return;
    if(d.length != nvelx) return;
    if(velx == null) velx = new double[nvelx];
    for(int i=0; i<nvelx; i++) velx[i] = d[i];
  }
  public void setYFlux(double [] d) {
    if(d == null) return;
    if(d.length != nvely) return;
    if(vely == null) vely = new double[nvely];
    for(int j=0; j<nvely; j++) vely[j] = d[j];
  }
  public void setZFlux(double [] d) {
    if(d == null) return;
    if(d.length != nvelz) return;
    if(velz == null) velz = new double[nvelz];
    for(int k=0; k<nvelz; k++) velz[k] = d[k];
  }

  public void setImageDimensions(int w, int h) {
    if(w<1) return;
    if(h<1) return;
    imageWidth = w;
    imageHeight = h;
  }

  public Dimension getImageDimensions() {
    Dimension d = new Dimension(imageWidth, imageHeight);
    return d;
  }

  public Image getSolPPMImageXY(Frame f, int kval) {
    if(kval < 0 || kval >= nz) return null;
    return getSolPPMImageXY(f, z[kval]);
  }

  public Image getSolPPMImageXY(Frame f, double zval) {
    if(h == null) return null;
    if(hImage != null) return hImage;
    if(zgr == null) return null;
    int k = 0;
    if(zval < zgr[0]) k = 0;
    for(int i=0; i<nz; i++) {
      if(zval >= zgr[i] && zval < zgr[i+1]) k = i;
    }
    if(zval >= zgr[nz]) k = nz - 1;
    double [] data = new double[nxny];
    int k2 = k * nxny;
    for(int j=0; j<ny; j++) {
      int j2 = j * nx;
      for(int i=0; i<nx; i++) {
        data[i+j2] = h[i+j2+k2];
      }
    }
    PPMGraphics ppmGraphics = new PPMGraphics(x, y, data,
                                              PPMGraphics.COLOR_MAP);
    ppmGraphics.setLinearFilter();
    ppmGraphics.setBorder(true);
    ppmGraphics.setBorderValue(10);
    hImage = ppmGraphics.buildImage(f, imageWidth);
    return hImage;
  }

  public boolean writeSolPPMImageXY(Frame f, int kval, File file) {
    if(kval < 0 || kval >= nz) return false;
    return writeSolPPMImageXY(f, z[kval], file);
  }

  public boolean writeSolPPMImageXY(Frame f, double zval, File file) {
    if(h == null) return false;
    if(zgr == null) return false;
    int k = 0;
    if(zval < zgr[0]) k = 0;
    for(int i=0; i<nz; i++) {
      if(zval >= zgr[i] && zval < zgr[i+1]) k = i;
    }
    if(zval >= zgr[nz]) k = nz - 1;
    double [] data = new double[nxny];
    int k2 = k * nxny;
    for(int j=0; j<ny; j++) {
      int j2 = j * nx;
      for(int i=0; i<nx; i++) {
        data[i+j2] = h[i+j2+k2];
      }
    }
    PPMGraphics ppmGraphics = new PPMGraphics(x, y, data,
                                              PPMGraphics.COLOR_MAP);
    ppmGraphics.setLinearFilter();
    ppmGraphics.setBorder(true);
    ppmGraphics.setBorderValue(10);
    ppmGraphics.writePPM(file, imageWidth, imageHeight);
    return true;
  }

  public Image getFluxPPMImageXY(Frame f, int kval) {
    if(kval < 0 || kval >= nz) return null;
    return getFluxPPMImageXY(f, z[kval]);
  }

  public Image getFluxPPMImageXY(Frame f, double zval) {
    if(h == null) return null;
    if(fluxImage != null) return fluxImage;
    if(zgr == null) return null;
    int k = 0;
    if(zval < zgr[0]) k = 0;
    for(int i=0; i<nz; i++) {
      if(zval >= zgr[i] && zval < zgr[i+1]) k = i;
    }
    if(zval >= zgr[nz]) k = nz - 1;
    double [] data = new double[nxny];
    int k1a = k * nxp1 * ny;
    int k1b = k * nyp1;
    int k2 = k * nxny;
    for(int j=0; j<ny; j++) {
      int j1a = j * nx * nzp1;
      int j1b = j * nxp1;
      int j2 = j * nx;
      for(int i=0; i<nx; i++) {
        int i1a = i * nyp1 * nz;
        int i1b = i * nzp1;
        double valx = 0.5 * ( velx[i+j1b+k1a] + velx[i+1+j1b+k1a] );
        double valy = 0.5 * ( vely[j+k1b+i1a] + vely[j+1+k1b+i1a] );
        double valz = 0.5 * ( velz[k+i1b+j1a] + velz[k+1+i1b+j1a] );
        data[i+j2] = Math.sqrt( valx * valx + valy * valy + valz * valz );
      }
    }
    PPMGraphics ppmGraphics = new PPMGraphics(x, y, data,
                                              PPMGraphics.COLOR_MAP);
    ppmGraphics.setLinearFilter();
    ppmGraphics.setBorder(true);
    ppmGraphics.setBorderValue(10);
    fluxImage = ppmGraphics.buildImage(f, imageWidth);
    return fluxImage;
  }

  public boolean writeFluxPPMImageXY(Frame f, int kval, File file) {
    if(kval < 0 || kval >= nz) return false;
    return writeFluxPPMImageXY(f, z[kval], file);
  }

  public boolean writeFluxPPMImageXY(Frame f, double zval, File file) {
    if(h == null) return false;
    if(zgr == null) return false;
    int k = 0;
    if(zval < zgr[0]) k = 0;
    for(int i=0; i<nz; i++) {
      if(zval >= zgr[i] && zval < zgr[i+1]) k = i;
    }
    if(zval >= zgr[nz]) k = nz - 1;
    double [] data = new double[nxny];
    int k1a = k * nxp1 * ny;
    int k1b = k * nyp1;
    for(int j=0; j<ny; j++) {
      int j1a = j * nx * nzp1;
      int j1b = j * nxp1;
      int j2 = j * nx;
      for(int i=0; i<nx; i++) {
        int i1a = i * nyp1 * nz;
        int i1b = i * nzp1;
        double valx = 0.5 * ( velx[i+j1b+k1a] + velx[i+1+j1b+k1a] );
        double valy = 0.5 * ( vely[j+k1b+i1a] + vely[j+1+k1b+i1a] );
        double valz = 0.5 * ( velz[k+i1b+j1a] + velz[k+1+i1b+j1a] );
        data[i+j2] = Math.sqrt( valx * valx + valy * valy + valz * valz );
      }
    }
    PPMGraphics ppmGraphics = new PPMGraphics(x, y, data,
                                              PPMGraphics.COLOR_MAP);
    ppmGraphics.setLinearFilter();
    ppmGraphics.setBorder(true);
    ppmGraphics.setBorderValue(10);
    ppmGraphics.writePPM(file, imageWidth, imageHeight);
    return true;
  }

  public String toString() {
    String s = "";
    s = "\n EllipticRectMFEObject Object Information\n";
    s = s + "\n\n x-coordinate Mesh Information\n";
    if(gridtypex == 0) s = s + "\n Grid Type String:  " + "UNIFORM GRID\n";
    if(gridtypex == 1) s = s + "\n Grid Type String:  " + "INPUT ARRAY NODE\n";
    if(gridtypex == 3) s = s + "\n Grid Type String:  " + "INPUT ARRAY GRID\n";
    if(gridtypex == 4) s = s + "\n Grid Type String:  " + "FUNCTION GRID\n" ;
    s = s + meshx.toString();
    s = s + "\n\n y-coordinate Mesh Information\n";
    if(gridtypey == 0) s = s + "\n Grid Type String:  " + "UNIFORM GRID\n";
    if(gridtypey == 1) s = s + "\n Grid Type String:  " + "INPUT ARRAY NODE\n";
    if(gridtypey == 3) s = s + "\n Grid Type String:  " + "INPUT ARRAY GRID\n";
    if(gridtypey == 4) s = s + "\n Grid Type String:  " + "FUNCTION GRID\n" ;
    s = s + meshy.toString();
    s = s + "\n\n z-coordinate Mesh Information\n";
    if(gridtypez == 0) s = s + "\n Grid Type String:  " + "UNIFORM GRID\n";
    if(gridtypez == 1) s = s + "\n Grid Type String:  " + "INPUT ARRAY NODE\n";
    if(gridtypez == 3) s = s + "\n Grid Type String:  " + "INPUT ARRAY GRID\n";
    if(gridtypez == 4) s = s + "\n Grid Type String:  " + "FUNCTION GRID\n" ;
    s = s + meshz.toString();
    s = s + "\n\n Method Order:  " + order + "\n";
    s = s + "\n h-values";
    s = s + printArray(h);
    s = s + "\n velx-values";
    s = s + printArray(velx);
    s = s + "\n vely-values";
    s = s + printArray(vely);
    s = s + "\n velz-values";
    s = s + printArray(velz);
    return s;
  }

  private String digitFilter(double value, int n) {
    String integerPart = "0";
    String dot = ".";
    String fractionPart = "0";
    if(Math.abs(value) < .000000000001) value = 0.0;
    String s = Double.toString(value);
    StringTokenizer st = new StringTokenizer(s, " .");
    if(st.hasMoreTokens()) integerPart = st.nextToken();
    if(st.hasMoreTokens()) {
      String t = st.nextToken();
      if(t.length() <= n) fractionPart = t;
      if(t.length() > n) fractionPart = t.substring(0,n);
    }
    return integerPart + dot + fractionPart;
  }

  public double [] sort(double[] input) {
    int n = input.length;
    int i;
    for(int j=1; j<n; j++) {
      double s = input[j];
      for(i=j-1; i>=0; i--) {
        if(s > input[i]) break;
        input[i+1] = input[i];
      }
      input[i+1] = s;
    }
    return input;
  }

  private String printArray(double [] a) {
    String s = "\n array output:";
    if(a == null) return null;
    int n = a.length;
    for(int i=0; i<n; i++) {
      s = s + "\n  index:  " + i + "  value:  " + a[i];
    } 
    s = s + "\n";
    return s;
  }

  private double [][] testMatrix() {
    Random r = new Random();
    double [][] a = new double[nnp][nnp];
    for(int i=0; i<nnp; i++) {
      for(int j=i+1; j<nnp; j++) {
        a[i][j] = 0.0;
        if(i>j) a[i][j] = r.nextDouble();
        if(j==i+1) a[i][j] = 1.0;
        a[j][i] = a[i][j];
      }
      a[i][i] = -4.0;
    }
    return a;
  }

  private double [] testMatMult(double [][] a, double [] input) {
    double [] output = new double [nnp];
    for(int i=0; i<nnp; i++) {
      output[i] = 0.0;
      for(int j=0; j<nnp; j++) {
        output[i] = output[i] + a[i][j] * input[j];
      }
    }
    return output;
  }

  private void stopHere() {
    System.out.println("Stop Here");
    try{
      System.in.read();
      System.in.read();
    } catch(IOException ioe) {
    }
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   * main() - called when the application is run.                            *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  public static void main(String args[]) {

    // coefficient data for testing the object
    String [][] a = {{"1.0", "0.0", "0.0"},
                     {"0.0", "1.0", "0.0"},
                     {"0.0", "0.0", "1.0"}};
    double [] asamp;

    // forcing function data for testing the object
    String f = "0.0";
    double [] fsamp;
    PointSources ps = new PointSources();

    // boundary condition for testing the object
    String bct = NEUMANN_BOUNDARY;
    // general function definition
    String fbfxl = "0.0";
    String fbfxr = "0.0";
    String fbfyl = "0.0";
    String fbfyr = "0.0";
    String fbfzl = "0.0";
    String fbfzr = "0.0";
    String pbfxl = "0.0";
    String pbfxr = "0.0";
    String pbfyl = "0.0";
    String pbfyr = "0.0";
    String pbfzl = "0.0";
    String pbfzr = "0.0";

    // the case nx=5, ny=5, nz=1 is the old test case
    int nx = 5;
    int ny = 5;
    int nz = 1;
    int nxr = 2;
    int nyr = 2;
    int nzr = 1;

    String grtypx = "UNIFORM GRID";
    String grtypy = "UNIFORM GRID";
    String grtypz = "UNIFORM GRID";

    String fnodex = "NOTHING";
    String fnodey = "NOTHING";
    String fnodez = "NOTHING";

    String fgridx = "NOTHING";
    String fgridy = "NOTHING";
    String fgridz = "NOTHING";

    double xl = 0.0;
    double xr = 1.0;
    double yl = 0.0;
    double yr = 1.0;
    double zl = 0.0;
    double zr = 1.0;

    double [] x = new double [nx];
    double [] xgr = new double [nx+1];
    double dx = ( xr - xl ) / ((double) nx);
    x[0] = dx / 2.0;
    for(int i=1; i<nx; i++) x[i] = x[i-1] + dx;
    xgr[0] = xl;
    for(int i=1; i<nx+1; i++) xgr[i] = xgr[i-1] + dx;

    double [] y = new double [ny];
    double [] ygr = new double [ny+1];
    double dy = ( yr - yl ) / ((double) ny);
    y[0] = dy / 2.0;
    for(int i=1; i<ny; i++) y[i] = y[i-1] + dy;
    ygr[0] = yl;
    for(int i=1; i<ny+1; i++) ygr[i] = ygr[i-1] + dy;

    double [] z = new double [nz];
    double [] zgr = new double [nz+1];
    double dz = ( zr - zl ) / ((double) nz);
    z[0] = dz / 2.0;
    for(int i=1; i<nz; i++) z[i] = z[i-1] + dz;
    zgr[0] = zl;
    for(int i=1; i<nz+1; i++) zgr[i] = zgr[i-1] + dz;

    int nnp = nx * ny * nz;
    int nxny = nx * ny;
    asamp = new double[9*nnp];
    for(int i=0; i<nnp; i++) asamp[i] = 1.0;
    int shft = nnp;
    for(int i=0; i<nnp; i++) asamp[i+shft] = 0.0;
    shft = shft + nnp;
    for(int i=0; i<nnp; i++) asamp[i+shft] = 0.0;
    shft = shft + nnp;
    for(int i=0; i<nnp; i++) asamp[i+shft] = 0.0;
    shft = shft + nnp;
    for(int i=0; i<nnp; i++) asamp[i+shft] = 1.0;
    shft = shft + nnp;
    for(int i=0; i<nnp; i++) asamp[i+shft] = 0.0;
    shft = shft + nnp;
    for(int i=0; i<nnp; i++) asamp[i+shft] = 0.0;
    shft = shft + nnp;
    for(int i=0; i<nnp; i++) asamp[i+shft] = 0.0;
    shft = shft + nnp;
    for(int i=0; i<nnp; i++) asamp[i+shft] = 1.0;

    fsamp = new double[nnp];
    for(int i=1; i<nnp; i++) fsamp[i] = 0.0;
    fsamp[0] = 0.25;
    fsamp[nnp-1] = -0.25;

    ps.addSource(0.0, 0.0, 0.5, 1.0);
    ps.addSource(1.0, 1.0, 0.5, -1.0);

    double [][] fbcx = new double[ny*nz][2];
    double [][] pbcx = new double[ny*nz][2];
    int [][] mbcx = new int[ny*nz][2];
    for(int k=0; k<nz; k++) {
      int k2 = k * ny;
      for(int j=0; j<ny; j++) {
        fbcx[j+k2][0] = 1.0;
        fbcx[j+k2][1] = 1.0;
        pbcx[j+k2][0] = 0.0;
        pbcx[j+k2][1] = 1.0;
      }
    }
    double [][] fbcy = new double[nx*nz][2];
    double [][] pbcy = new double[nx*nz][2];
    int [][] mbcy = new int[nx*nz][2];
    for(int k=0; k<nz; k++) {
      int k2 = k * nx;
      for(int i=0; i<nx; i++) {
        fbcy[i+k2][0] = 1.0;
        fbcy[i+k2][1] = 1.0;
        pbcy[i+k2][0] = 0.0;
        pbcy[i+k2][1] = 1.0;
      }
    }
    double [][] fbcz = new double[nx*ny][2];
    double [][] pbcz = new double[nx*ny][2];
    int [][] mbcz = new int[nx*ny][2];
    for(int j=0; j<ny; j++) {
      int j2 = j * nx;
      for(int i=0; i<nx; i++) {
        fbcz[i+j2][0] = 1.0;
        fbcz[i+j2][1] = 1.0;
        pbcz[i+j2][0] = 0.0;
        pbcz[i+j2][1] = 1.0;
      }
    }

    double tc = 1.0;
    double dt = 0.01;

    int order = 1;

    System.out.println("********************* Test #0 *********************");
    System.out.println("*************** 5x5 Analytic Test Case ************");

    try {
      EllipticRectMFEObject ermfeo =
         new EllipticRectMFEObject(
                "SYMBOLIC", a, asamp,               // elliptic coefficent info
                "POINT_SOURCE", f, fsamp, ps,       // source function info
                "SYMBOLIC", bct,                    // boundary cond type
                fbfxl, fbfxr,
                fbfyl, fbfyr,
                fbfzl, fbfzr,
                pbfxl, pbfxr,
                pbfyl, pbfyr,
                pbfzl, pbfzr,
                fbcx, fbcy, fbcz,
                pbcx, pbcy, pbcz,
                mbcx, mbcy, mbcz,
                xl, xr, yl, yr, zl, zr,
                nx, ny, nz,
                nxr, nyr, nzr,
                grtypx, grtypy, grtypz,
                x, y, z,
                fnodex, fnodey, fnodez,
                xgr, ygr, zgr,
                fgridx, fgridy, fgridz,
                // the next line sets time dependence to 0
                0.0, dt,
                order, null, true);
      System.out.println("\n\nEllipticRectMFEObject Info:\n\n");
      System.out.println(ermfeo.toString());

      Frame f1 = new Frame("EllipticRectMFEObject Graphics Test");
      Dimension d = ermfeo.getImageDimensions();
      int w = d.width;
      int h = d.height;
      if(w<320) w = 320;
      if(h<320) h = 320;
      ImageCanvas imgCanvas = new ImageCanvas(w, h);
      f1.add(imgCanvas);
      f1.setSize(w+40, h+40);
      f1.show();

      Image img = ermfeo.getSolPPMImageXY(f1, 0);
      imgCanvas.redraw(0, 0, img);

    } catch(WellPosedProblemException wppe) {
      System.out.println("Well Posed Problem Exception Thrown");
    }

    System.out.println("********************* End Test ********************");

  }

  // local variables
  private String coefType;    // identifies how the coefficient is specified
  private String [][] a;      // coefficient in the elliptic operator
  private double [] asampxx;  // sample version of the coefficient
  private double [] asampxy;  // sample version of the coefficient
  private double [] asampxz;  // sample version of the coefficient
  private double [] asampyx;  // sample version of the coefficient
  private double [] asampyy;  // sample version of the coefficient
  private double [] asampyz;  // sample version of the coefficient
  private double [] asampzx;  // sample version of the coefficient
  private double [] asampzy;  // sample version of the coefficient
  private double [] asampzz;  // sample version of the coefficient
  private double [] accxy;    // modified coefficient values at the cell centers
  private double [] accxz;    // modified coefficient values at the cell centers
  private double [] accyx;    // modified coefficient values at the cell centers
  private double [] accyz;    // modified coefficient values at the cell centers
  private double [] acczx;    // modified coefficient values at the cell centers
  private double [] acczy;    // modified coefficient values at the cell centers
  private String forcingType; // forcing function in the pde
  private PointSources pointSources; // point source object
  private String f;           // forcing function in the pde
  private double [] fsamp;    // forcing function in the pde
  private String boundaryConditionType; // boundary condition type
  private String boundaryEvaluationType;   // bndry val evaluation type
  private String fluxBoundaryFunctionXL = null;  // function to define bndry val
  private String fluxBoundaryFunctionXR = null;  // function to define bndry val
  private String fluxBoundaryFunctionYL = null;  // function to define bndry val
  private String fluxBoundaryFunctionYR = null;  // function to define bndry val
  private String fluxBoundaryFunctionZL = null;  // function to define bndry val
  private String fluxBoundaryFunctionZR = null;  // function to define bndry val
  private String prsBoundaryFunctionXL = null;  // function to define bndry val
  private String prsBoundaryFunctionXR = null;  // function to define bndry val
  private String prsBoundaryFunctionYL = null;  // function to define bndry val
  private String prsBoundaryFunctionYR = null;  // function to define bndry val
  private String prsBoundaryFunctionZL = null;  // function to define bndry val
  private String prsBoundaryFunctionZR = null;  // function to define bndry val
  private double [][] xBoundaryPrs;// function to define bndry val
  private double [][] yBoundaryPrs;// function to define bndry val
  private double [][] zBoundaryPrs;// function to define bndry val
  private double [][] xBoundaryFlux;// function to define bndry val
  private double [][] yBoundaryFlux;// function to define bndry val
  private double [][] zBoundaryFlux;// function to define bndry val
  private int [][] xBoundaryMask;// boundary value mask
  private int [][] yBoundaryMask;// boundary value mask
  private int [][] zBoundaryMask;// boundary value mask
  private double xl;          // left boundary in x
  private double xr;          // right boundary in x
  private double yl;          // left boundary in y
  private double yr;          // right boundary in y
  private double zl;          // left boundary in z
  private double zr;          // right boundary in z
  private int nx;             // number of nodes in x
  private int nxm1;           // extra grid parameter for convenience
  private int nxp1;           // number of velocity nodes in x
  private int nxref;          // number of refinement nodes in x
  private int ny;             // number of nodes in y
  private int nym1;           // extra grid parameter for convenience
  private int nyp1;           // number of velocity nodes in y
  private int nyref;          // number of refinement nodes in y
  private int nz;             // number of nodes in z
  private int nzm1;           // extra grid parameter for convenience
  private int nzp1;           // number of velocity nodes in z
  private int nzref;          // number of refinement nodes in z
  private double [] x;        // explicit node values in x
  private double [] y;        // explicit node values in y
  private double [] z;      // explicit node values in z
  private String fnodex;    // function to define grid in x
  private String fnodey;    // function to define grid in y
  private String fnodez;    // function to define grid in z
  private double [] xgr;    // explicit grid line values in x
  private double [] ygr;    // explicit grid line values in y
  private double [] zgr;    // explicit grid line values in z
  private String fgridx;    // function to define grid in x
  private String fgridy;    // function to define grid in y
  private String fgridz;    // function to define grid in z
  private int order;        // order of the method

  private boolean problem = false;

  // default coefficient in the operator
  String [][] defaultCoeff = {{"1.0", "0.0", "0.0"},
                              {"0.0", "1.0", "0.0"},
                              {"0.0", "0.0", "1.0"}};

  // some integer parameters
  private int nvelx;
  private int nvely;
  private int nvelz;
  private int nnp;
  private int nxny;

  private double [] dx;
  private double [] dy;
  private double [] dz;

  // mesh definitions
  private int gridtypex;
  private int gridtypey;
  private int gridtypez;
  private Mesh1D meshx;
  private Mesh1D meshy;
  private Mesh1D meshz;

  // basis function definitions
  private TensorProduct3D [] pressureFunction;
  private int noOfPresFunc;
  private TensorProduct3D [] xVelocityFunction;
  private int noOfXVelFunc;
  private TensorProduct3D [] yVelocityFunction;
  private int noOfYVelFunc;
  private TensorProduct3D [] zVelocityFunction;
  private int noOfZVelFunc;

  // mfe matrix stuff
  private double [] adx;
  private double [] asx;
  private double [] ady;
  private double [] asy;
  private double [] adz;
  private double [] asz;
  private double [] rhsvx;
  private double [] rhsvy;
  private double [] rhsvz;
  private double [] rhsh;
  private double [] an;

  // time dependent piece
  private double [] diagh;
  private double tstep = 0.01;
  private double tcoeff = 0.0;

  // mass matrices
  private double [] admssx;
  private double [] asmssx;
  private double [] almssx;
  private double [] admssy;
  private double [] asmssy;
  private double [] almssy;
  private double [] admssz;
  private double [] asmssz;
  private double [] almssz;

  private double [] ibith;
  private double [] ibitvx;
  private double [] ibitvy;
  private double [] ibitvz;

  // variable stuff
  private double [] h;
  private double [] hold;
  private double [] velx;
  private double [] vely;
  private double [] velz;

  private Image hImage = null;
  private Image fluxImage = null;
  private int imageWidth = 256;
  private int imageHeight = 256;

  // constants
  private double [] sig = {0.2113248654052, 0.7886751345948};

  // some extra stuff
  private FormulaParser formulaParser;
  private double garbage = -100000000.0;

  public static final String NEUMANN_BOUNDARY = "NEUMANN";
  public static final String DIRICHLET_BOUNDARY = "DIRICHLET";
  public static final String MIXED_BOUNDARY = "MIXED";

  public static final String SYMBOLIC_FUNCTION_DEFINITION = "SYMBOLIC";
  public static final String NUMERIC_FUNCTION_DEFINITION = "NUMERIC";

  public static final String POINT_SOURCE_FUNCTION_DEFINITION = "POINT_SOURCE";

  private TextArea outputArea = null; 
  public boolean debug = false;

}
